/*
 * Copyright 2008  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "asteroidgen.h"

#include "mymath.h"

#include <QList>
#include <QFile>
#include <QTextStream>
#include <QtDebug>

#include <KConfig>
#include <KConfigGroup>

#include <Eigen/Geometry>

using namespace Eigen;

#include <stdlib.h>
using namespace std;


QDataStream& operator<<(QDataStream& stream, const Vector2f& v)
{
    return stream << v.x() << v.y();
}
QDataStream& operator<<(QDataStream& stream, const Vector3f& v)
{
    return stream << v.x() << v.y() << v.z();
}

QDebug& operator<<(QDebug& debug, const Vector3f& v)
{
    return debug << "(" << v.x() << "," << v.y() << "," << v.z() << ")";
}


AsteroidGen::AsteroidGen()
{
    seed = time(0);
    lodCount = 4;
    randomInitial = 0.4;
    randomIterative = 0.2;
    randomDamping = 1.8;
    randomPost = 0.0;

    faces = 0;
}


AsteroidGen::~AsteroidGen()
{
    delete[] faces;
}

QString AsteroidGen::modelParameters() const
{
    QString params;
    QTextStream stream(&params);
    const int version = 1;
    stream << version << ' ' << seed << ' ' << randomInitial << ' ' << randomIterative << ' ' <<
            randomDamping << ' ' << randomPost;
    return params;
}

void AsteroidGen::setModelParameters(const QString& params)
{
    QString _params(params);
    QTextStream stream(&_params);
    int version;
    stream >> version;
    qDebug() << "params:" << params << ", version:" << version;
    stream >> seed >> randomInitial >> randomIterative >> randomDamping >> randomPost;
    qDebug() << "seed:" << seed << ", randomInitial:" << randomInitial;
}

int AsteroidGen::getVertexIndex(int parenta, int parentb)
{
    // Two possibilities
    QPair<int, int> edge1 = qMakePair(parenta, parentb);
    QPair<int, int> edge2 = qMakePair(parentb, parenta);

    if (edgeMapper.contains(edge1)) {
        return edgeMapper[edge1];
    } else if (edgeMapper.contains(edge2)) {
        return edgeMapper[edge2];
    } else {
        vertices.append(Vertex(vertices[parenta], vertices[parentb]));
        edgeMapper[edge1] = vertices.count()-1;
        return edgeMapper[edge1];
    }
}

bool AsteroidGen::generate()
{
    // Clean old data
    delete[] faces;
    vertices.clear();
    edgeMapper.clear();

    currentLodCount = lodCount;
    faces = new QList<Triangle>[currentLodCount];

    // Random seed
    srand(seed);

    // Add initial vertices
    vertices.append(Vertex(Vector3f( 1,  0,  0), Vector2f(0.0 , 0.5)));
    vertices.append(Vertex(Vector3f(-1,  0,  0), Vector2f(0.5 , 0.5)));
    vertices.append(Vertex(Vector3f( 0,  1,  0), Vector2f(0.25, 0.5)));
    vertices.append(Vertex(Vector3f( 0, -1,  0), Vector2f(0.75, 0.5)));
    vertices.append(Vertex(Vector3f( 1,  0,  0), Vector2f(1.0 , 0.5)));
    vertices.append(Vertex(Vector3f( 0,  0,  1), Vector2f(0.5 , 1.0)));
    vertices.append(Vertex(Vector3f( 0,  0, -1), Vector2f(0.5 , 0.0)));
    // Randomize them
    for (int i = 0; i < 7; i++) {
        vertices[i].pos *= randFloat(1.0 - randomInitial, 1.0 + randomInitial);
    }

    // Add initial faces
    faces[0].append(Triangle(1, 6, 3));
    faces[0].append(Triangle(3, 6, 0 /*4*/));
    faces[0].append(Triangle(0, 6, 2));
    faces[0].append(Triangle(2, 6, 1));
    faces[0].append(Triangle(1, 3, 5));
    faces[0].append(Triangle(3, 0 /*4*/, 5));
    faces[0].append(Triangle(0, 2, 5));
    faces[0].append(Triangle(2, 1, 5));

    edgeMapper.clear();

    // Subdivide the model
    float delta = randomIterative;
    for (int level = 1; level < currentLodCount; level++) {
        // Subdivide the model
        QList<Triangle> newfaces;
        for (int i = 0; i < faces[level-1].count(); i++) {
            Triangle old = faces[level-1][i];
            // Subdivide this triangle
            int aidx = getVertexIndex(old.a, old.c);
            int bidx = getVertexIndex(old.a, old.b);
            int cidx = getVertexIndex(old.b, old.c);


            newfaces.append(Triangle(old.a, bidx, aidx));
            newfaces.append(Triangle(bidx, old.b, cidx));
            newfaces.append(Triangle(aidx, bidx, cidx));
            newfaces.append(Triangle(aidx, cidx, old.c));
        }

        faces[level] = newfaces;
        // Randomize the new vertices
        for (int i = 0; i < vertices.count(); i++) {
            vertices[i].pos *= randFloat(1.0 - delta, 1.0 + delta);
        }

        delta /= randomDamping;
    }

    // Randomize
    for (int i = 0; i < vertices.count(); i++) {
        vertices[i].pos *= randFloat(1.0 - randomPost, 1.0 + randomPost);
    }

    // Fix texcoords
    for (int i = 0; i < vertices.count(); i++) {
        if (vertices[i].tex.y() > 0.01 && vertices[i].tex.y() < 0.99) {
            vertices[i].tex.x() = (vertices[i].tex.x() - 0.5) / (1.0 - 2*qAbs(0.5 - vertices[i].tex.y())) + 0.5;
        }
    }

    // Calc normals
    for (int i = 0; i < vertices.count(); i++) {
        vertices[i].normal = Vector3f(0, 0, 0);
    }
    // Calculate face normals
    for (int i = 0; i < faces[currentLodCount-1].count(); i++) {
        Triangle& t = faces[currentLodCount-1][i];
        t.normal = (vertices[t.c].pos - vertices[t.b].pos).cross(vertices[t.a].pos - vertices[t.b].pos);
        t.normal.normalize();
        vertices[t.a].normal += t.normal;
        vertices[t.b].normal += t.normal;
        vertices[t.c].normal += t.normal;
    }
    for (int i = 0; i < vertices.count(); i++) {
        vertices[i].normal.normalize();
    }

    // Center the asteroid and make it's bounding sphere's radius 1
    // Calculate bounding box
    Vector3f minpos, maxpos;
    minpos = maxpos = vertices[0].pos;
    for (int i = 0; i < vertices.count(); i++) {
        minpos.x() = qMin(minpos.x(), vertices[i].pos.x());
        maxpos.x() = qMax(maxpos.x(), vertices[i].pos.x());
        minpos.y() = qMin(minpos.y(), vertices[i].pos.y());
        maxpos.y() = qMax(maxpos.y(), vertices[i].pos.y());
        minpos.z() = qMin(minpos.z(), vertices[i].pos.z());
        maxpos.z() = qMax(maxpos.z(), vertices[i].pos.z());
    }
    // Center the asteroid and calculate current radius
    Vector3f translate = -(maxpos + minpos) /2;
    float radiusSq = 0;
    for (int i = 0; i < vertices.count(); i++) {
        vertices[i].pos += translate;
        radiusSq = qMax(radiusSq, vertices[i].pos.squaredNorm());
    }
    // Make asteroid's radius 1
    float radius = sqrt(radiusSq);
    for (int i = 0; i < vertices.count(); i++) {
        vertices[i].pos /= radius;
    }

    return true;
}

bool AsteroidGen::saveLod(int lod, const QString& filename)
{
    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly)) {
        qCritical() << "Can't open output file for writing";
        return false;
    }

    QTextStream out(&file);

    // We only want to store the vertices which are actually used by this lod.
    // To do that, we create vertex mapping that maps our vertex ids to saved ids.
    QVector<int> internalToSaved(vertices.count(), -1);
    QVector<int> savedVertices;

    // First figure out which vertices to use
    for (int i = 0; i < faces[lod].count(); i++) {
        int a = faces[lod][i].a;
        int b = faces[lod][i].b;
        int c = faces[lod][i].c;
        if (internalToSaved[a] == -1) {
            savedVertices.append(a);
            internalToSaved[a] = savedVertices.count();
        }
        if (internalToSaved[b] == -1) {
            savedVertices.append(b);
            internalToSaved[b] = savedVertices.count();
        }
        if (internalToSaved[c] == -1) {
            savedVertices.append(c);
            internalToSaved[c] = savedVertices.count();
        }
    }

    out << "# Random seed " << seed << "\n";
    out << "# Model params '" << modelParameters() << "'\n";
    out << "# LOD " << lod << "\n";
    out << "# " << savedVertices.count() << " vertices, " << faces[lod].count() << " faces\n";

    for (int i = 0; i < savedVertices.count(); i++) {
        const Vertex& v = vertices.at(savedVertices[i]);
        out << "v "  << v.pos.x() << " " << v.pos.y() << " " << v.pos.z() << "\n";
    }
    for (int i = 0; i < savedVertices.count(); i++) {
        const Vertex& v = vertices.at(savedVertices[i]);
        out << "vt " << v.tex.x() << " " << v.tex.y() << "\n";
    }
    for (int i = 0; i < savedVertices.count(); i++) {
        const Vertex& v = vertices.at(savedVertices[i]);
        out << "vn " << v.normal.x() << " " << v.normal.y() << " " << v.normal.z() << "\n";
    }

    // Write indices. Note that indices are 1-based
    for (int i = 0; i < faces[lod].count(); i++) {
        int a = internalToSaved[faces[lod][i].a];
        int b = internalToSaved[faces[lod][i].c];
        int c = internalToSaved[faces[lod][i].b];
        out << QString("f %1/%1/%1 %2/%2/%2 %3/%3/%3").arg(c).arg(b).arg(a) << "\n";
    }

    return true;
}

bool AsteroidGen::saveModel(const QString& filenamePrefix)
{
    return saveModel(filenamePrefix + "-%1.obj", filenamePrefix + ".index");
}

bool AsteroidGen::saveModel(const QString& filenameTemplate, const QString& indexfile)
{
    KConfig conf(indexfile);
    KConfigGroup modelconf(&conf, "Model");
    modelconf.writeEntry("LODs", currentLodCount);

    for (int lod = 0; lod < currentLodCount; lod++) {
        QString modelfile = QString(filenameTemplate).arg(lod);
        if (!saveLod(lod, modelfile)) {
            return false;
        }

        KConfigGroup lodgroup(&conf, QString("LOD-%1").arg(lod));
        if (modelfile.lastIndexOf('/') != -1) {
            lodgroup.writeEntry("Model", modelfile.mid(modelfile.lastIndexOf('/')+1));
        } else {
            lodgroup.writeEntry("Model", modelfile);
        }
    }

    return true;
}

