/*
 * Copyright 2009  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "datastore.h"

#include "model.h"
#include "shadermanager.h"

#include <kgllib/program.h>
#include <kgllib/texture.h>

#include <KDebug>
#include <QDir>
#include <KConfigGroup>
#include <KConfig>
#include <kgllib/geometrybuffer.h>


DataStore* DataStore::mInstance = 0;


void DataStore::initInstance(const QString& datadir)
{
    if (mInstance) {
        kFatal() << "Instance already exists!";
        return;
    }

    mInstance = new DataStore(datadir);
}


DataStore::DataStore(const QString& datadir)
{
    mDataDir = datadir;
    kDebug() << "Data dir is" << mDataDir;

    mShaderManager = new ShaderManager(datadir + "shaders/");
}

DataStore::~DataStore()
{
}

Model* DataStore::model(const QString& name)
{
    return mModels[name];
}

KGLLib::Texture* DataStore::texture(const QString& name)
{
    if (!mTextures.contains(name)) {
        kDebug() << name;
        Texture* tex = new Texture(mDataDir + "textures/" + name);
        if (!tex->isValid()) {
            kError() << "Couldn't load texture" << name;
            delete tex;
            tex = 0;
        }
        mTextures[name] = tex;
    }

    return mTextures[name];
}

KGLLib::Program* DataStore::program(const QString& name)
{
    if (!mPrograms.contains(name)) {
        kDebug() << name;
        const QString shaderdir = mDataDir + "shaders/";
        Program* prog = new Program(shaderdir + name + ".vert", shaderdir + name + ".frag");
        if (!prog->isValid()) {
            kError() << "Couldn't load program" << name;
            delete prog;
            return 0;
        }

        prog->bind();
        prog->setUniform("diffuseTexture", 0);
        prog->setUniform("glossTexture", 1);
        prog->setUniform("emissionTexture", 2);
        prog->setUniform("normalTexture", 3);
        prog->unbind();
        mPrograms[name] = prog;
    }

    return mPrograms[name];
}

bool DataStore::loadAsteroidModels()
{
    QDir modelsDir(mDataDir + "asteroids/");
    QStringList modelFiles = modelsDir.entryList(QStringList("*.index"));
    kDebug() << "Found" << modelFiles.count() << "asteroid models" << endl;

    for (int i = 0; i < modelFiles.count(); i++) {
        QString file = modelFiles[i];
        QString modelName = file.left(file.length() - 6);
        emit loadingAsteroid(modelName, i, modelFiles.count());
        kDebug() << "Loading model" << modelName;

        Model* model = new Model(mDataDir + "asteroids/" + file);
        if (!model->isValid()) {
            return false;
        }
        model->setTexture(texture("asteroid.png"));
        model->setMaterial("asteroid");
        mNonFinalizedModels[modelName] = model;
        mAsteroidModels.append(model);
    }

    return true;
}

Model* DataStore::randomAsteroidModel()
{
    return mAsteroidModels.at(rand() % mAsteroidModels.count());
}

bool DataStore::loadModel(const QString& name)
{
    if (mModels.contains(name) || mNonFinalizedModels.contains(name)) {
        return true;
    }

    KConfig objectconf(mDataDir + "objects/" + name + ".index");
    KConfigGroup modelconf(&objectconf, "Model");
    Model* model = new Model(mDataDir + "objects/" + name + ".index", mDataDir + "models/");

    if (!model->isValid()) {
        kError() << "Couldn't load model" << name;
        return false;
    }

    foreach (const QString& texkey, modelconf.keyList().filter("Texture-")) {
        bool isnum = false;
        int texnum = texkey.mid(8).toInt(&isnum);
        if (!isnum) {
            kError() << "Invalid texture key" << texkey;
            continue;
        }
        Texture* tex = texture(modelconf.readEntry(texkey));
        if (!tex) {
            return false;
        }
        model->setTexture(texnum, tex);
    }
    mNonFinalizedModels[name] = model;

    return true;
}

bool DataStore::finalizeModels()
{
    int totalvertices = 0;
    int totalindices = 0;
    foreach (Model* model, mNonFinalizedModels) {
        totalvertices += model->vertexCount();
        totalindices += model->indexCount();
    }
    kDebug() << "Have total of" << totalvertices << "vertices and" << totalindices << "indices for" << mNonFinalizedModels.count() << "models";

    // Create GeometryBuffer
    GeometryBufferFormat format(totalvertices, totalindices);
    format.addVertices(3);
    format.addNormals();
    format.addTexCoords(2);
    GeometryBuffer* buffer = GeometryBuffer::createBuffer(format);

    int vertexoffset = 0;
    int indexoffset = 0;
    QHashIterator<QString, Model*> it(mNonFinalizedModels);
    while (it.hasNext()) {
        it.next();
        it.value()->load(buffer, &vertexoffset, &indexoffset);
        mModels[it.key()] = it.value();
    }

    mNonFinalizedModels.clear();
    mGeometryBuffers.append(buffer);

    return true;
}
