/*
    Copyright 2009  Rivo Laks <rivolaks@hot.ee>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "gameguiview.h"

#include "gameworld.h"
#include "ship.h"
#include "renderinfo.h"
#include "displaywidget.h"

#include <kgllib/glwidget.h>
#include <kgllib/textrenderer.h>
#include <kgllib/texture.h>

#include <QFont>
#include <KDebug>

using namespace KGLLib;


class GLMeter
{
    public:
        GLMeter(const QString& datapath, float max, float value = 0.0f);
        ~GLMeter();

        void setValue(float value);
        void setRect(const QRectF& rect);
        void setLabel(const QString& text);

        float displayedValue() const  { return mValue; }

        void render();
        void renderTexts(TextRenderer* textrenderer);

        bool isValid() const  { return mValid; }

    protected:
        void renderCenteredQuad(float width, float height);

    private:
        KGLLib::Texture* mBackgroundTex;
        KGLLib::Texture* mForegroundTex;
        bool mValid;

        QRectF mRect;
        QRect mTextRect;
        QFont mFont;
        QString mLabel;
        float mMaxValue;

        float mValue;
        float mRealValue;
};


GLMeter::GLMeter(const QString& datapath, float max, float value)
{
    mMaxValue = max;
    mRealValue = mValue = value;
    mValid = true;

    mBackgroundTex = new KGLLib::Texture(datapath + "gui/meter-bg.png");
    if (!mBackgroundTex->isValid()) {
        kError() << "Couldn't load background texture";
        mValid = false;
    }
    mForegroundTex = new KGLLib::Texture(datapath + "gui/meter-fg.png");
    if (!mForegroundTex->isValid()) {
        kError() << "Couldn't load foreground texture";
        mValid = false;
    }
}

GLMeter::~GLMeter()
{
    delete mForegroundTex;
    delete mBackgroundTex;
}

void GLMeter::setValue(float value)
{
    mRealValue = value;
}

void GLMeter::setLabel(const QString& text)
{
    mLabel = text;
}

void GLMeter::setRect(const QRectF& rect)
{
    mRect = rect;

    // Recalculate text rectangle
    mTextRect = QRect(mRect.left() + mRect.width()/4, mRect.top() + mRect.height() * 0.65, mRect.width()/2, mRect.height()/4);
    // Find a good font size
    mFont = QFont("Arial");
    mFont.setPixelSize(qMax(8, int(mRect.height() * 0.12)));
}

void GLMeter::renderCenteredQuad(float width, float height)
{
    float halfWidth = width/2;
    float halfHeight = height/2;

    glBegin(GL_QUADS);
        glTexCoord2f(0.0, 0.0);
        glVertex2f(-halfWidth, -halfHeight);
        glTexCoord2f(1.0f, 0.0);
        glVertex2f( halfWidth, -halfHeight);
        glTexCoord2f(1.0f, 1.0f);
        glVertex2f( halfWidth,  halfHeight);
        glTexCoord2f(0.0, 1.0f);
        glVertex2f(-halfWidth,  halfHeight);
    glEnd();
}

void GLMeter::render()
{
    if (!mValid) {
        kWarning() << "Meter not valid!";
        return;
    }

    // Controls how fast the indicator reaches the actual value (the bigger the faster)
    float a = 0.05;
    mValue = mValue * (1-a) + mRealValue*a;

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(1.0f, 1.0f, 1.0f, 0.9f);

    float width = mRect.width();
    float height = mRect.height();

    // Translate origin to the center of the given rect
    glPushMatrix();
    glTranslatef(mRect.left() + mRect.width()/2, mRect.top() + mRect.height()/2, 0);

    // Draw background
    mBackgroundTex->enable();
    renderCenteredQuad(width, height);

    // Draw rotated foreground
    glRotatef(45 + 270 * (mValue/mMaxValue), 0, 0, 1);
    mForegroundTex->enable();
    renderCenteredQuad(width, height);
    mForegroundTex->disable();
    glPopMatrix();
}

void GLMeter::renderTexts(TextRenderer* textRenderer)
{
    if (!mLabel.isEmpty()) {
        QRect labelRect(mRect.left(), mRect.top() - 53, mRect.width(), 50);
        textRenderer->draw(labelRect, mLabel, Qt::AlignHCenter | Qt::AlignBottom, mFont);
    }
    textRenderer->draw(mTextRect, QString("%1").arg(displayedValue(), 0, 'f', 1), Qt::AlignCenter, mFont);
}



GameGUIView::GameGUIView(KGLLib::GLWidget* parent, GameWorld* world) : GLView(parent)
{
    mWorld = world;
    mThrustoMeter = 0;
    mSpeedoMeter = 0;
    mShowRenderInfo = false;

    setViewType(View_GUI);
}

void GameGUIView::setShowRenderInfo(bool show)
{
    mShowRenderInfo = show;
}

void GameGUIView::init()
{
    kDebug();
    mThrustoMeter = new GLMeter(mWorld->dataPath(), 20.0f);
    mThrustoMeter->setLabel("Thrust:");
    mSpeedoMeter = new GLMeter(mWorld->dataPath(), 200.0f);
    mSpeedoMeter->setLabel("Speed:");
}

void GameGUIView::resize(int w, int h)
{
    float thrustDialSize = 120;
    QRectF thrustRect(20, h - thrustDialSize - 20, thrustDialSize, thrustDialSize);
    mThrustoMeter->setRect(thrustRect);

    float speedDialSize = 160;
    QRectF speedRect(20 + thrustDialSize + 15, h - speedDialSize - 20, speedDialSize, speedDialSize);
    mSpeedoMeter->setRect(speedRect);
}

void GameGUIView::render()
{
    // The meters cannot be rendered while the textrender is active. So render them first and then texts.
    setupOrthoView();

    mThrustoMeter->setValue(mWorld->ship()->thrust);
    mThrustoMeter->render();

    mSpeedoMeter->setValue(mWorld->ship()->velocity().norm());
    mSpeedoMeter->render();

    textRenderer()->begin(glWidget());

    // Render health
    QFont healthfont("Arial", 14);
    QRect healthrect = glWidget()->rect().adjusted(10, 10, -10, -10);
    if (mWorld->ship()->health > 0) {
        glColor4f(0.6, 0.6, 1.0, 0.7);
        textRenderer()->draw(healthrect, QString("Health: %1").arg((int)mWorld->ship()->health),
                              Qt::AlignRight | Qt::AlignBottom, healthfont);
    } else {
        glColor4f(1.0, 0.5, 0.3, 0.7);
        textRenderer()->draw(healthrect, QString("Ship destroyed"),
                              Qt::AlignRight | Qt::AlignBottom, healthfont);
    }

    // Thrust
    glColor4f(1.0f, 0.0f, 0.0f, 0.9f);
    mThrustoMeter->renderTexts(textRenderer());
    mSpeedoMeter->renderTexts(textRenderer());

    // Paused label
    if (mWorld->gamePaused()) {
        glColor4f(0.4, 0.4, 1.0, 1.0);
        textRenderer()->draw(glWidget()->rect(), QString("Game Paused"),
                              Qt::AlignCenter | Qt::AlignVCenter, QFont("Arial", 24));
    }

    // Time
    glColor4f(0.4, 0.4, 1.0, 1.0);
    QRect timerect = glWidget()->rect().adjusted(10, 10, -10, -30);
    QFont timefont("Arial", 14);
    // Show the time in "MM:SS.xx" format
    QString timestring("--:--.--");
    float trackTime = mWorld->trackTime();
    if (trackTime > 0.0f) {
        timestring.sprintf("%02d:%02d.%02d", int(trackTime / 60), int(trackTime) % 60,
                            int((trackTime - int(trackTime)) * 100));
    }
    textRenderer()->draw(timerect, timestring,
                          Qt::AlignRight | Qt::AlignBottom, timefont);

    // Dodges available
    QRect dodgerect = glWidget()->rect().adjusted(10, 10, -10, -50);
    textRenderer()->draw(dodgerect, QString("Dodges: %1").arg(mWorld->ship()->dodgesAvailable, 0, 'f', 1),
                          Qt::AlignRight | Qt::AlignBottom, timefont);

    if (mShowRenderInfo) {
        QString statistics;
        Vector3f shipPos = mWorld->ship()->position();
        statistics += QString("Ship pos: (%1; %2; %3)\n")
                .arg(shipPos.x(), 7, 'f', 1).arg(shipPos.y(), 7, 'f', 1).arg(shipPos.z(), 7, 'f', 1);
        Vector3f shipDir = mWorld->ship()->direction();
        statistics += QString("Ship dir: (%1; %2; %3)\n")
                .arg(shipDir.x(), 5, 'f', 2).arg(shipDir.y(), 5, 'f', 2).arg(shipDir.z(), 5, 'f', 2);
        Vector3f shipVeloN = mWorld->ship()->velocity().normalized();
        statistics += QString("Ship dir: (%1; %2; %3) * %4\n")
                .arg(shipVeloN.x(), 5, 'f', 2).arg(shipVeloN.y(), 5, 'f', 1).arg(shipVeloN.z(), 5, 'f', 1)
                .arg(mWorld->ship()->velocity().norm(), 5, 'f', 1);

        statistics += QString("Objects: %1\n").arg(mWorld->objects()->count());
        statistics += QString("Lights: %1  batches: %2  faces: %3\n")
                .arg(renderInfo->lights()).arg(renderInfo->batches()).arg(renderInfo->faces());
        statistics += QString("State changes:\n");
        statistics += QString("buf: %1  mat: %2  tex: %3\n")
                .arg(renderInfo->bufferChanges()).arg(renderInfo->materialChanges()).arg(renderInfo->textureChanges());

        DisplayWidget* w = static_cast<DisplayWidget*>(glWidget());
        statistics += QString("HDR: %1\n").arg(w->hdrRenderingActive());
        if (w->hdrRenderingActive()) {
            statistics += QString("Exposure: %1\n").arg(w->exposure(), 0, 'f', 2);
        }

        glColor3f(0.8, 0.8, 0.8);
        QRect statsRect(10, 30, 300, 100);
        textRenderer()->draw(statsRect, statistics, Qt::AlignTop | Qt::AlignLeft, QFont("Courier New", 10));
    }

    textRenderer()->end();
}

