/*
 * Copyright 2008  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gameobject.h"

#include "gameworld.h"
#include "model.h"

#include <Eigen/Geometry>

#include <kgllib/batch.h>

#include <KDebug>

GameObject::GameObject(float r)
{
    id = -1;
    radius = r;

    mModel = 0;
    mModelScale = 1.0f;

    // ODE stuff
    body = dBodyCreate(GameWorld::instance()->odeWorld());
    dBodySetData(body, this);
    setMass(radius*radius);

    geom = 0;
    mTriMeshGeom = false;
}

GameObject::~GameObject()
{
}

void GameObject::init()
{
}

void GameObject::setMass(float massTotal)
{
    // Create a mass with the distribution of a sphere
    dMass mass;
    dMassSetSphereTotal(&mass, massTotal, radius);
    // Apply the mass to the cube
    dBodySetMass(body, &mass);
}

void GameObject::setVelocity(const Vector3f& v)
{
    dBodySetLinearVel(body, v.x(), v.y(), v.z());
}

void GameObject::addForce(const Vector3f& force)
{
    dBodyAddForce(body, force.x(), force.y(), force.z());
}

Matrix3f GameObject::rotationMatrix() const
{
    const dReal* rot = dBodyGetRotation(body);
    // TODO: optimize: there shouldn't be any need for the temporary matrix
    GLfloat matrix[9];
    matrix[0] = rot[0];
    matrix[1] = rot[4];
    matrix[2] = rot[8];
    matrix[3] = rot[1];
    matrix[4] = rot[5];
    matrix[5] = rot[9];
    matrix[6] = rot[2];
    matrix[7] = rot[6];
    matrix[8] = rot[10];
    return Matrix3f(matrix);
}

void GameObject::setRotation(const Vector3f& rot)
{
    dMatrix3 matrix;
    dRFromEulerAngles(matrix, rot.x(), rot.y(), rot.z());

    dBodySetRotation(body, matrix);
}

void GameObject::setDirection(const Vector3f& dir)
{
    Vector3f currentDir = direction();
    Vector3f cross = currentDir.cross(dir.normalized());

    dMatrix3 matrix;
    float angle = asin(cross.norm());
    dRFromAxisAndAngle(matrix, cross.x(), cross.y(), cross.z(), angle);

    dBodySetRotation(body, matrix);
}

Vector3f GameObject::direction() const
{
    return (rotationMatrix() * Vector3f(0, 0, -1)).normalized();
}

Vector3f GameObject::up() const
{
    return (rotationMatrix() * Vector3f(0, 1, 0)).normalized();
}

void GameObject::advance(float elapsed)
{
    /*const dReal* vel = dBodyGetLinearVel(body);
    const float factor = -0.25;
    dBodyAddForce(body, factor*vel[0]*vel[0], factor*vel[1]*vel[1], factor*vel[2]*vel[2]);*/
    if (mTriMeshGeom) {
        const dReal* Pos = dGeomGetPosition(geom);
        const dReal* Rot = dGeomGetRotation(geom);

        dReal Transform[16] =
        {
            Rot[0], Rot[1], Rot[2],  0,
            Rot[4], Rot[5], Rot[6],  0,
            Rot[8], Rot[9], Rot[10], 0,
            Pos[0], Pos[1], Pos[2],  1
        };
        dGeomTriMeshSetLastTransform( geom, *(dMatrix4*)(&Transform) );
    }
}

void GameObject::openglTransform() const
{
    const dReal* pos = dBodyGetPosition(body);
    const dReal* rot = dBodyGetRotation(body);
    GLfloat matrix[16];
    matrix[0] = rot[0];
    matrix[1] = rot[4];
    matrix[2] = rot[8];
    matrix[3] = 0;
    matrix[4] = rot[1];
    matrix[5] = rot[5];
    matrix[6] = rot[9];
    matrix[7] = 0;
    matrix[8] = rot[2];
    matrix[9] = rot[6];
    matrix[10] = rot[10];
    matrix[11] = 0;
    matrix[12] = pos[0];
    matrix[13] = pos[1];
    matrix[14] = pos[2];
    matrix[15] = 1;
    glMultMatrixf(matrix);
}

void GameObject::setModel(Model* m)
{
    mModel = m;
    // Set ODE's geometry
    KGLLib::Batch* cm = mModel->collisionMesh();
    createGeomFromMesh((const Vector3f*)cm->verticesArray(), cm->vertexCount(),
                       (const unsigned int*)cm->indicesArray(), cm->indicesCount());
}

void GameObject::createSphereGeometry()
{
    geom = dCreateSphere(GameWorld::instance()->odeSpace(), radius);
    dGeomSetBody(geom, body);
}

void GameObject::createGeomFromMesh(const Vector3f* vertices, int vertexCount,
                                    const unsigned int* indices, int indexCount)
{
    dTriMeshDataID trimeshData = dGeomTriMeshDataCreate();

    dGeomTriMeshDataBuildSingle(trimeshData,
                                 vertices, sizeof(Vector3f), vertexCount,
                                 indices, indexCount, 3*sizeof(unsigned int));

    geom = dCreateTriMesh(GameWorld::instance()->odeSpace(), trimeshData, 0, 0, 0);

    dMatrix3 R;
    dRSetIdentity(R);
    dGeomSetRotation(geom, R);

    dGeomSetBody(geom, body);
    mTriMeshGeom = true;
}
