/*
 * Copyright 2008  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GAMEWORLD_H
#define GAMEWORLD_H

#include <QObject>
#include <QList>

#include "gameobject.h"

#include <ode/ode.h>

class Ship;
class Asteroid;
class Bullet;
class Model;
class Planet;
class DataStore;


/**
 * @author Rivo Laks <rivolaks@hot.ee>
 */
class GameWorld : public QObject
{
Q_OBJECT
public:
    GameWorld(QObject *parent);
    ~GameWorld();

    static GameWorld* instance()  { return mInstance; }

    bool init();
    void advance(float elapsed);

    Ship* ship() const  { return mShip; }
    const QList<GameObject*>* objects() const  { return &mObjects; }
    const QList<Bullet*>* bullets() const  { return &mBullets; }
    bool gamePaused() const  { return mGamePaused; }
    QString dataPath() const  { return mDataDir; }
    DataStore* data() const  { return mData; }

    void addObject(GameObject* obj);
    void setShip(Ship* s);

    Planet* addPlanet(const QString& type);
    const QList<Planet*>* planets() const  { return &mPlanets; }

    void handleCollisionBetween(dGeomID o0, dGeomID o1);

    void createAsteroids(int count, float mean, float variance, float min, float max,
                         const Vector3f& mincoord, const Vector3f& maxcoord);
    void addAsteroids(int count);

    GameObject* addTrackPoint(const Vector3f& pos);
    const QList<Vector3f>& trackPoints() const  { return mTrack; }
    float trackTime() const;

    dWorldID odeWorld() const  { return mOdeWorld; }
    dSpaceID odeSpace() const  { return mOdeSpace; }

    GameObject* createObject(const QString& name);


public slots:
    void slotRotateShip(float dx, float dy, float dz);
    void slotRotateShip(int dx, int dy)  { slotRotateShip(dx, -dy, 0.0f); }
    void slotToggleGamePaused();

    void slotShootBullet();
    void setShipLight(bool on);

protected:
    void applyShipRotation(float elapsed);

    void collision(GameObject* a, GameObject* b);

    void processContactFeedbacks();
    void processTrackPoints();


private:
    class OdeContactFeedback;

    static GameWorld* mInstance;

    Ship* mShip;
    QList<GameObject*> mObjects;
    QList<Bullet*> mBullets;
    QList<Planet*> mPlanets;
    bool mGamePaused;
    float mTime;
    int mNextObjectId;
    QList<OdeContactFeedback*> mContactFeedbacks;

    QList<Vector3f> mTrack;
    float mTrackTime;
    bool mTrackActive;
    float mAdvanceTime;

    dWorldID mOdeWorld;
    dSpaceID mOdeSpace;
    dJointGroupID mOdeContactGroup;

    DataStore* mData;
    QString mDataDir;

public:
    static int mCollisionCandidatesWithDups;
    static int mCollisionCandidates;
    static int mCollisions;
};

#endif
