/*
    Copyright 2008-2009  Rivo Laks <rivolaks@hot.ee>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "shadermanager.h"

#include <kgllib/program.h>
#include <kgllib/shader.h>

#include <KDebug>
#include <QTime>

using namespace KGLLib;


uint qHash(const ShaderManager::ShaderInfo& info)
{
    return qHash(info.lightType) xor qHash(info.lightType) xor qHash(info.materialType);
}


ShaderManager::ShaderManager(const QString& shaderDir)
{
    mShaderDir = shaderDir;
    if (!mShaderDir.endsWith('/')) {
        mShaderDir.append('/');
    }

    mActiveProgram = 0;
}

ShaderManager::~ShaderManager()
{
}

KGLLib::Program* ShaderManager::activateShader()
{
    if (!mPrograms.contains(mActiveInfo)) {
        mPrograms.insert(mActiveInfo, compileProgram(mActiveInfo));
    }

    Program* prog = mPrograms[mActiveInfo];
    if (prog) {
        prog->bind();

        // Update variable values
        prog->setUniform("lightColor", mActiveInfo.lightColor);
        prog->setUniform("lightPosition", mActiveInfo.lightPosition);
        prog->setUniform("lightDirection", mActiveInfo.lightDirection);
        prog->setUniform("lightRadiusInverse", 1.0f / mActiveInfo.lightRadius);

        mActiveProgram = prog;
    }

    return prog;
}

bool ShaderManager::precompileProgram(const QString& renderMode, const QString& lightType, const QString& material)
{
    ShaderInfo info;
    info.renderMode = renderMode;
    info.lightType = lightType;
    info.materialType = material;

    Program* prog = compileProgram(info);
    if (!prog) {
        return false;
    }
    mPrograms.insert(info, prog);
    return true;
}

void ShaderManager::disableShader()
{
    if (mActiveProgram) {
        mActiveProgram->unbind();
        mActiveProgram = 0;
    }
}

void ShaderManager::setLight(const QString& type)
{
    mActiveInfo.lightType = type;
}

void ShaderManager::setNoLight()
{
    mActiveInfo.lightType = "none";
}

void ShaderManager::setLightColor(const Eigen::Vector3f& color)
{
    mActiveInfo.lightColor = color;
}

void ShaderManager::setLightPosition(const Eigen::Vector3f& position)
{
    mActiveInfo.lightPosition = position;
}

void ShaderManager::setLightDirection(const Eigen::Vector3f& dir)
{
    mActiveInfo.lightDirection = dir.normalized();
}

void ShaderManager::setLightRadius(float r)
{
    mActiveInfo.lightRadius = r;
}

void ShaderManager::setMaterial(const QString& type)
{
    mActiveInfo.materialType = type;
}

void ShaderManager::setRenderMode(const QString& mode)
{
    mActiveInfo.renderMode = mode;
}

Program* ShaderManager::compileProgram(const ShaderManager::ShaderInfo& info)
{
    QTime prof;
    prof.start();
    VertexShader lightVertexShader(mShaderDir + "light-" + info.lightType + ".vert");
    if (!lightVertexShader.isValid()) {
        kError() << "Failed to load light vertex shader";
        return 0;
    }
    VertexShader materialVertexShader(mShaderDir + "material-" + info.materialType + ".vert");
    if (!materialVertexShader.isValid()) {
        kError() << "Failed to load material vertex shader";
        return 0;
    }
    VertexShader renderVertexShader(mShaderDir + "render-" + info.renderMode + ".vert");
    if (!renderVertexShader.isValid()) {
        kError() << "Failed to load rendering vertex shader";
        return 0;
    }
    FragmentShader lightFragmentShader(mShaderDir + "light-" + info.lightType + ".frag");
    if (!lightFragmentShader.isValid()) {
        kError() << "Failed to load light fragment shader";
        return 0;
    }
    FragmentShader materialFragmentShader(mShaderDir + "material-" + info.materialType + ".frag");
    if (!materialFragmentShader.isValid()) {
        kError() << "Failed to load material fragment shader";
        return 0;
    }
    FragmentShader renderFragmentShader(mShaderDir + "render-" + info.renderMode + ".frag");
    if (!renderFragmentShader.isValid()) {
        kError() << "Failed to load rendering fragment shader";
        return 0;
    }

    Program* prog = new Program();
    prog->addShader(&renderVertexShader);
    prog->addShader(&lightVertexShader);
    prog->addShader(&materialVertexShader);
    prog->addShader(&renderFragmentShader);
    prog->addShader(&lightFragmentShader);
    prog->addShader(&materialFragmentShader);

    if (!prog->link()) {
        kError() << "Couldn't link program";
        delete prog;
        return 0;
    }

    // Set common uniforms
    prog->bind();
    prog->setUniform("diffuseTexture", 0);
    prog->setUniform("glossTexture", 1);
    prog->setUniform("emissionTexture", 2);
    prog->setUniform("normalTexture", 3);
    prog->unbind();

    kDebug() << "Dynamically compiled program for" << info.renderMode << "/" << info.lightType << "/" <<
            info.materialType << "in" << prof.elapsed() << "ms";

    return prog;
}
