/***************************************************************************
 *   Copyright (C) 2009 by Alexander Smirnov                               *
 *   alexander.v.smirnov@gmail.com                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/

#include "board.h"
#include "boardrenderer.h"

#include <kdebug.h>

Board* Board::self()
{
    static Board instance;
    return &instance;
}

/**
 * Constructor
 */
Board::Board():m_dice(numberOfDice())
{
    for (int i = 0; i < m_dice.size(); i++)
    {
            m_dice[i] = new Die(i);
    }
}

/**
 * Board has dies located on it.
 * This method returns specified die
 */
Die* Board::die(qint8 n)
{
    Q_ASSERT(n<m_numberOfDice);
    return self()->m_dice[n];
}

/**
 * Show dice on the board
 */ 
void Board::showDice()
{
    for (int i = 0; i < self()->m_dice.size(); i++)
    {
        self()->m_dice[i]->show();
    }
}

/**
 * Hide dice
 */ 
void Board::hideDice()
{
    kDebug() << "hide dice";
    for (int i = 0; i < self()->m_dice.size(); i++)
    {
        self()->m_dice[i]->hide();
    }
}

/*void Board::hideDie( qint8 n )
{
    foreach( Die* die, self()->m_dice )
    {
        if( die->number() == n )
        {
            die->hide();
            break;
        }
    }
}*/

/**
 * Return number of checkers at specified cell
 * @param cellIdx - cell index, integer from 0 to 28
 */
qint8  Board::numberOfCheckersAt(qint8 cellIdx)
{
    Q_ASSERT(self()->indexValid( cellIdx ));
    return self()->m_checkers.count(cellIdx);
}

/**
 * Return list of checker GraphicItems associated with cellIdx.
 *
 * @param cellIdx - cell index to check, integer from 0 to 28
 * @return list of checker graphics item associated with specified cell
 */
QList<Checker*> Board::checkersAt(qint8 cellIdx)
{
    Q_ASSERT(self()->indexValid( cellIdx ));
    return self()->m_checkers.values(cellIdx);
}

/**
 * Adds checker to specified position
 * @param cellIdx - cell index to check,integer from 0 to 28
 * @param checker - checker to add
  */
void Board::addChecker(qint8 cellIdx, Checker * checker)
{
    Q_ASSERT( self()->indexValid( cellIdx ) );
    self()->m_checkers.insert(cellIdx, checker);
    checker->setCell (cellIdx);
    checker->setOrder( numberOfCheckersAt(cellIdx) );
}

Checker* Board::getLastChecker(qint8 cellIdx)
{
    kDebug() << "cellIdx=" << cellIdx;
    Q_ASSERT( self()->indexValid( cellIdx ) );
    Q_ASSERT( !self()->m_checkers.values(cellIdx).isEmpty() );

    return self()->m_checkers.values(cellIdx).first();
    //return 0;
}

void Board::setCheckerSize(int size )
{
    for (int cellIdx = 0; cellIdx < m_numberOfCells; ++cellIdx)
    {
        const QList<Checker*> checkers = checkersAt(cellIdx);
        foreach (Checker* checker, checkers)
        {
            checker->updatePixmap(size);
        }
    }
}

/**
 * Moves checker from one board position to another
 * This is a low-level method and it is not aware of game rules
 * which are described in BackgammonGame class //TODO; update where rules are implemented
 * @param cellFrom - cell to move checker from
 * @param cellTo - cell to move checker to
 */
void Board::moveChecker(Checker * checker, qint8 cellTo)
{
    if ( !self()->indexValid( cellTo ) )
    {
        kDebug() << "index#"<<cellTo<<"  out of range, move not possible...";
        return;
    }

    self()->m_checkers.remove(checker->cell(), checker);
    self()->m_checkers.insert(cellTo, checker);

    checker->setCell(cellTo);
    checker->setOrder( numberOfCheckersAt(cellTo) );
}

/**
 * Depending on parameter sign, return Black/White/NoColor
 * TODO: move to fibs module??
 */
Checker::Color Board::checkerColor( qint8 c )
{
    if ( c > 0 )
    {
        return Checker::Black;
    }
    else if ( c < 0 )
    {
        return Checker::White;
    }
    else
    {
        return Checker::NoColor;
    }
}

/**
 * Check if board index is valid
 * @param cellIdx - cell index to check,integer from 0 to 28
 * @return true if index is valid, false otherwise
 */
bool Board::indexValid(qint8 cellIdx)
{
    return ( cellIdx >= 0 && cellIdx < m_numberOfCells );
}

/**
 * Return color of checker located at specified cell
 * @param idx - cell index to check,integer from 0 to 28
 * @return White if checkers, located at specified cell is white.
 *         Black if checkers, located at specified cell is black,
 *         NoColor.if cell is empty.
 */
Checker::Color Board::cellColor(qint8 cellIdx)
{
    Q_ASSERT(self()->indexValid( cellIdx ));
    //if cell is empty, return NoColor
    if ( self()->m_checkers.count(cellIdx) == 0 )
    {
        return Checker::NoColor;
    }
    else //return color of any checker at this cell
    {
        return self()->m_checkers.values(cellIdx).at(0)->color();
    }
}

qint8 Board::home(Checker::Color color)
{
    Q_ASSERT( color != Checker::NoColor );

    if (color == Checker::White)
    {
        return WHITE_POOL;
    }
    else //if (color == Checker::Black)
    {
        return BLACK_POOL;
    }
}

qint8 Board::bar(Checker::Color color)
{
    Q_ASSERT( color != Checker::NoColor );

    if (color == Checker::White)
    {
        return WHITE_BAR;
    }
    else //if (color == Checker::Black)
    {
        return BLACK_BAR;
    }
}


QString Board::state()
{
    QString result;
    QString delimeter="";
    for (int i = 0; i < 28; i++)
    {
        int number = self()->m_checkers.count(i);
        if (number != 0)
        {
            number *= self()->m_checkers.values(i).at(0)->color();
        }
        result.append(delimeter).append(QString::number(number));
        delimeter=":";
    }
    return result;
}

/**
 * Handle resize event
 * Recalculate board-related values, such as cell size and board rectangle
 */
void Board::resize(int width, int height )
{
    kDebug() << "resizing board...";
    static const qreal boardFactor = 0.93; //TODO: move it to preferences?

    QSize boardSize = BoardRenderer::defaultBoardSize();
    boardSize.scale(boardFactor*width, boardFactor*height, Qt::KeepAspectRatio);

    self()->m_boardRect.setRect( (1-boardFactor)/2*width,
                         (1-boardFactor)/2*height,
                         boardSize.width(),
                         boardSize.height() );

    for (int i = 0; i < m_numberOfCells; ++i)
    {
       self()->m_cellBounds[i].setWidth(-1);//this invalidates rectangle and forces recalculation on next request(lazy)
    }
}

/**
 * Returns rectangle for specified board cell
 * Board should have cells with ids(in terms of SVG) "cell1", "cell2"..."cell24"
 * 
 * @param cellIdx - number of cell the rectangle will be returned for. Should be number from 1 to 24
 */
const QRectF& Board::cellRect(const qint8 cellIdx)
{
    if (cellIdx <0 || cellIdx >= m_numberOfCells) // TODO:use indexValid
    {
        kDebug() << "ERROR: Wrong cell number requested: " << cellIdx;
        return self()->m_cellBounds[m_numberOfCells+1];
    }

    if (self()->m_cellBounds[cellIdx].isValid())
    {
        return self()->m_cellBounds[cellIdx];
    }

    kDebug() << "recalculate cell#" << cellIdx << " bounds.";
    const QRectF cellRect = BoardRenderer::defaultCellRect(cellIdx);
    const QRectF boardRect = BoardRenderer::defaultBoardRect();
    // a bit of math
    //TODO: should be cleary documented(with pictures) where these formulas come from
    //TODO: is there a more simple/clear way for finding the rectangle?May be, may be...if you know it, please advice
    //first we scale cell rectangle to correspond specified board size
    const qreal w = self()->m_boardRect.width() * cellRect.width() / boardRect.width();
    const qreal h = self()->m_boardRect.height() * cellRect.height() / boardRect.height();
    //and then move scaled rectangle to scene coordinates
    const qreal x = (cellRect.x()-boardRect.x()) / boardRect.width() * self()->m_boardRect.width() + self()->m_boardRect.x();
    const qreal y = (cellRect.y()-boardRect.y()) / boardRect.height() * self()->m_boardRect.height() + self()->m_boardRect.y();

    self()->m_cellBounds[cellIdx].setX(x);
    self()->m_cellBounds[cellIdx].setY(y);
    self()->m_cellBounds[cellIdx].setWidth(w);
    self()->m_cellBounds[cellIdx].setHeight(h);

    return self()->m_cellBounds[cellIdx];
}

/**
 * @return cell number what contains specified point. Number in range [0;27] If there is no cell, containing the point, then -1 is returned
 */
qint8 Board::cellNumber(const QPointF& point)
{
    for (int i = 0; i < m_numberOfCells; ++i)
    {
        if ( cellRect(i).contains(point) )
            return i;
    }
    return -1;
}
