/***************************************************************************
 *   Copyright (C) 2005 by Piotr Banasiak <piotrbanasiak@onet.pl>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to                             *
 *   the Free Software Foundation, 51 Franklin Street, Fifth Floor,        *
 *   Boston, MA 02110-1301, USA.                                           *
 ***************************************************************************/
#include "boardmodel.h"
#include "board.h"
#include "config.h"
#include "game.h"
#include "piecedelegate.h"
#include "humanplayer.h"

// using game because we needo not only borad reprezentation but also possible moves
// and porbably last moved piece and so on
BoardModel::BoardModel( Game *pGame )
 : QAbstractTableModel()
{
    m_Player=0;
    currentPiece=QPoint( -1,-1 );
    path.clear();

    m_Game=pGame;
    bool ok=connect( m_Game, SIGNAL(documentChanged()), this, SLOT(allDataChanged()));
}

void BoardModel::setGame( Game *pGame )
{
    currentPiece=QPoint( -1,-1 );
    path.clear();
    disconnect( m_Game , 0,0,0);
    m_Game=pGame;
    bool ok=connect( m_Game, SIGNAL(documentChanged()), this, SLOT(allDataChanged()));
}

void BoardModel::setPlayer( humanPlayer * pPlayer)
{
  m_Player=pPlayer;
}


BoardModel::~BoardModel()
{
}

int BoardModel::rowCount ( const QModelIndex & /*parent*/  ) const
{
    return Conf().m_boardSize;
}

int BoardModel::columnCount ( const QModelIndex & /*parent*/  ) const
{
    return Conf().m_boardSize;
}

QVariant BoardModel::data ( const QModelIndex & index, int role  ) const
{
   Board *pBoard= m_Game->getCurrentBoard();
   Q_ASSERT( pBoard );
   
    int r=index.row();
    int c=index.column();
    int x=index.column();
    int y=Conf().m_boardSize-r-1;
    QPoint point=QPoint( x,y );
   
    if (!index.isValid())
        return QVariant();

    if (r >= Conf().m_boardSize )
        return QVariant();
    if (c >= Conf().m_boardSize )
        return QVariant();

    //qDebug( "BoardModel::data() x: %d, y: %d   role:%d",x,y ,role);
    if (role == Qt::DisplayRole)
    {
        int piece= pBoard->get( x,y );
        if (piece) return piece;
        //if empty square, return light or dark
        if( (x+y)%2) return Board::WHITE;
        return Board::BLACK;
    }
    if( role ==Qt::ToolTipRole )
    {
        bool can=canMove( x, y );
        if (!can) return QVariant();
        return QVariant("can move");
    }
    
    if (role == PieceDelegate::canMoveRole)
    {
      bool can=canMove( x, y );
      return  QVariant( can );
    }
    
    if (role == PieceDelegate::currentPieceRole)
    {
      if(!m_Player)return QVariant();
      
      bool eq = point==currentPiece;
      return  QVariant( eq );
    }
    
    if (role == PieceDelegate::lastMoveRole)
    {
      const LegalMove *const pLM=m_Game->getLastMove();
      if(!pLM)return  QVariant( 0 );
      for(int j=0;j<pLM->path.count();j++)
      {
        const QPoint xy = pLM->path[j];
        if( xy==point )return  QVariant( 1 );
      }
      for(int j=0;j<pLM->takenPieces.count();j++)
      {
        const QPoint xy = pLM->takenPieces[j];
        if( xy==point )return  QVariant( 1 );
      }
      return  QVariant( 0 );
    }
    
    if (role == PieceDelegate::nextStepRole)
    {
      if(!m_Player)return QVariant();
      
      QPoint xy= QPoint( x,y );
      int moveNumber;
      bool ok = checkNextMove( xy ,moveNumber);
      return  QVariant( ok );
    }
    return  QVariant();
}

QVariant BoardModel::headerData ( int section, Qt::Orientation orientation, int role ) const
{
    if (role != Qt::DisplayRole)return  QVariant();
    if( orientation==Qt::Horizontal )
    {
        QChar c( 'A'+section );
        QString header (c);
        return header;
    }
    
    if( orientation==Qt::Vertical )
    {
        //return QString("Column %1").arg(section);
        return Conf().m_boardSize-section;
    }
    return QVariant();
}

/*!
    \fn BoardModel::canMove(int x, int y, Board *pB)
 */
bool BoardModel::canMove(int x, int y)const
{
    const QList<LegalMove>* pMoveList=m_Game->getPossibleMoves();
    QList<LegalMove>::const_iterator i;
    for (i = pMoveList->constBegin(); i != pMoveList->constEnd(); ++i)
    {
        QPoint start=i->path[0];
        //qDebug( "start: (%d, %d)",start.x , start.y );
        QPoint A(x,y);
        if (A==start)return true;
    }
    return false;
}
//----------------------------------------------------------------------------------------
// means square was cliked
bool BoardModel::setData ( const QModelIndex & index, const QVariant & /*value*/, int /*role*/ )
{
    int x=index.column();
    int y=Conf().m_boardSize-index.row()-1;
    
    //check if this is moveable piece
    if( canMove(x,  y) )
    {
      currentPiece=QPoint( x , y );
      path.clear();
      path.append( QPoint(x,y));
      allDataChanged();
      return true;
    }
    
    //check if this is move completed
    int moveNumber=-1;
    int cmp=checkNextMove( QPoint(x,y) ,moveNumber );
    if(cmp==2)
    { 
       path.clear();
       currentPiece=QPoint(-1,-1);
       const QList<LegalMove>* pMoveList=m_Game->getPossibleMoves();
       
       LegalMove LM=pMoveList->at(moveNumber);
       if(m_Player)
          m_Player->enteredMove( LM );
       //allDataChanged();///@todo should be emit all pieces and decorated
       return true;
    }
    //just next part of complex move
    if(cmp)
    {
       path.append( QPoint(x,y));
       allDataChanged();
    }
    return true;
}
//-----------------------------------------------------------------------
// moveNr is number of the move from Game
int BoardModel::checkNextMove( const QPoint & xy , int& moveNr)const
{
    const QList<LegalMove>* pMoveList=m_Game->getPossibleMoves();
    QList<LegalMove>::const_iterator i;
    
    if( path.isEmpty() ) return 0;
    
    int ok=0;
    for (int j=0;j< pMoveList->count(); j++)
    {
        LegalMove LM = (*pMoveList)[j];
        int cmp=LM.comparePart( path, xy );
        if( cmp==2 ) 
        {
           moveNr=j;
           return 2;
        }
        if(cmp)ok=true;
    }
    return ok;
}
//-----------------------------------------------------------------------
/*!
    \fn BoardModel::allDataChanged()
 */
void BoardModel::allDataChanged()
{
  ///@todo remove that dirrty trick, of repainting all
  QModelIndex idx0=createIndex ( 0, 0 );
  QModelIndex idx2=createIndex ( Conf().m_boardSize-1, Conf().m_boardSize-1 );
  emit dataChanged( idx0, idx2 );
}

#include "boardmodel.moc"
