/***************************************************************************
 *   Copyright (C) 2005 by Piotr Banasiak <piotrbanasiak@onet.pl>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to                             *
 *   the Free Software Foundation, 51 Franklin Street, Fifth Floor,        *
 *   Boston, MA 02110-1301, USA.                                           *
 ***************************************************************************/
#include "boardview.h"
#include "piecedelegate.h"

#include <QAbstractItemView>
#include <QTreeView>
#include <QPainter>
#include <QtGui>

BoardView::BoardView(QWidget *parent, PieceDelegate *delegate):
QAbstractItemView( parent )
{
    vHeader=0;
    hHeader=0;
    hHeader2=0;
    vHeader2=0;
    m_showGrid=false;
    m_inverted=false;
    
    setItemDelegate(delegate);
    m_cellSizeHint=itemDelegate()->sizeHint( QStyleOptionViewItem() ,QModelIndex() );
    
    setVerticalHeader();
    setHorizontalHeader();

    setHorizontalScrollBarPolicy ( Qt::ScrollBarAlwaysOff );
    setVerticalScrollBarPolicy ( Qt::ScrollBarAlwaysOff );
}

BoardView::~BoardView()
{
}

/*!
    Sets the widget to use for the vertical header to \a header.

    \sa horizontalHeader() setVerticalHeader()
*/
void BoardView::setHorizontalHeader()
{
    hHeader = new QHeaderView(Qt::Horizontal, this);
    hHeader2 = new QHeaderView(Qt::Horizontal, this);
    
    hHeader->setModel(model());
    hHeader2->setModel(model());

    connect(hHeader,SIGNAL(sectionResized(int,int,int)),  this, SLOT(columnResized(int,int,int)));
    connect(hHeader, SIGNAL(sectionCountChanged(int,int)), this, SLOT(columnCountChanged(int,int)));

    connect(hHeader2,SIGNAL(sectionResized(int,int,int)),  this, SLOT(column2Resized(int,int,int)));
    connect(hHeader2, SIGNAL(sectionCountChanged(int,int)), this, SLOT(columnCountChanged(int,int)));
    //hHeader->setFocusProxy(this);
    
    int sections=hHeader->count();
    int grid=m_showGrid?1:0;
    for( int j=0;j<sections;j++ )
    {
       hHeader->resizeSection( j , m_cellSizeHint.width() +grid);
       hHeader2->resizeSection( j , m_cellSizeHint.width() +grid);
    }
}

/*!
    Sets the widget to use for the horizontal header to \a header.

    \sa verticalHeader() setHorizontalHeader()
*/
void BoardView::setVerticalHeader()
{
    vHeader  = new QHeaderView(Qt::Vertical, this);
    vHeader2 = new QHeaderView(Qt::Vertical, this);
    
    vHeader->setClickable(false);
    vHeader2->setClickable(false);

    vHeader->setModel(model());
    vHeader2->setModel(model());

    connect(vHeader, SIGNAL(sectionResized(int,int,int)),   this, SLOT(rowResized(int,int,int)));
    connect(vHeader, SIGNAL(sectionCountChanged(int,int)),  this, SLOT(rowCountChanged(int,int)));

    connect(vHeader2, SIGNAL(sectionResized(int,int,int)),   this, SLOT(row2Resized(int,int,int)));
    connect(vHeader2, SIGNAL(sectionCountChanged(int,int)),  this, SLOT(rowCountChanged(int,int)));
    //vHeader->setFocusProxy(this);
    
    int sections=vHeader->count();
    for( int j=0;j<sections;j++ )
    {
       vHeader->resizeSection( j , m_cellSizeHint.height() );
       vHeader2->resizeSection( j , m_cellSizeHint.height() );
    }
}


/*!
    This slot is called whenever columns are added or deleted. The
    previous number of columns is specified by \a oldCount, and the new
    number of columns is specified by \a newCount.
*/
void BoardView::columnCountChanged( int /*oldCount*/, int newCount)
{
    cols=newCount;
    
    int sections=hHeader->count();
    for( int j=0;j<sections;j++ )
    {
       hHeader->resizeSection( j , m_cellSizeHint.width() );
    }
    
    int sections2=hHeader2->count();
    for( int j=0;j<sections2;j++ )
    {
       hHeader2->resizeSection( j , m_cellSizeHint.width() );
    }
    hHeader->setResizeMode(QHeaderView::Custom);
    //hHeader2->setResizeMode(QHeaderView::Custom);
    updateGeometries();
    viewport()->update();
}

/*!
    This slot is called whenever rows are added or deleted. The
    previous number of rows is specified by \a oldCount, and the new
    number of rows is specified by \a newCount.
*/
void BoardView::rowCountChanged(int /*oldCount*/, int newCount )
{
    
    rows=newCount;
    int sections=vHeader->count();
    for( int j=0;j<sections;j++ )
    {
       vHeader->resizeSection( j , m_cellSizeHint.height() );
    }
    
    int sections2=vHeader2->count();
    for( int j=0;j<sections2;j++ )
    {
       vHeader2->resizeSection( j , m_cellSizeHint.height() );
    }
    
    vHeader->setResizeMode(QHeaderView::Custom);
    //vHeader2->setResizeMode(QHeaderView::Custom);
    updateGeometries();
    viewport()->update();
}


QSize BoardView::minimumSizeHint() const
{
   int y=rowViewportPosition( 0 );
   int h = y+rows*m_cellSizeHint.height();
   int hHeight=hHeader->height();
   
   
   int x=columnViewportPosition( 0 );
   int w =  x+cols*m_cellSizeHint.width();
   int vWidth=vHeader->width();
    
   
   return QSize( w+2*vWidth, h+2*hHeight );
}

QRect BoardView::visualRect(const QModelIndex &index) const
{
    if (!index.isValid() ||index.parent() != rootIndex() || isIndexHidden(index) )
        return QRect();

    const int i = m_showGrid ? 1 : 0;
    return QRect(columnViewportPosition(index.column()), rowViewportPosition(index.row()),
                 columnWidth(index.column()) - i,  rowHeight(index.row()) - i);
}

/*!
    We don't need that function at all.
*/
void BoardView::scrollTo(const QModelIndex & /*index*/, ScrollHint /*hint*/ )
{
   return;
}

/*!
    Returns the index position of the model item corresponding to the
    table item at position \a pos in contents coordinates.
*/
QModelIndex BoardView::indexAt(const QPoint &pos) const
{
    //d_func()->executePostedLayout();
    int r = rowAt(pos.y());
    int c = columnAt(pos.x());
    
    if( m_inverted )
    {
        c = hHeader->count()-1-c;
        r = vHeader->count()-1-r;
    }
    if (r >= 0 && c >= 0)
        return model()->index(r, c, rootIndex());
    return QModelIndex();
}

QModelIndex BoardView::moveCursor(CursorAction , Qt::KeyboardModifiers )
{
   return QModelIndex();
}

/*!
    Returns the horizontal offset of the items in the table view.

    Note that the table view uses the horizontal header section
    positions to determine the positions of columns in the view.

    \sa verticalOffset()
*/
int BoardView::horizontalOffset() const
{
    return horizontalHeader()->offset();
}

/*!
    Returns the vertical offset of the items in the table view.

    Note that the table view uses the vertical header section
    positions to determine the positions of rows in the view.

    \sa horizontalOffset()
*/
int BoardView::verticalOffset() const
{
    return verticalHeader()->offset();
}

//is never hidden
bool BoardView::isIndexHidden(const QModelIndex &) const
{
    return false;
}

/*!
  \reimp
*/
void BoardView::setSelectionModel(QItemSelectionModel *selectionModel)
{
    Q_ASSERT(selectionModel);
    vHeader->setSelectionModel(selectionModel);
    horizontalHeader()->setSelectionModel(selectionModel);
    QAbstractItemView::setSelectionModel(selectionModel);
}

/*!
    \fn void BoardView::setSelection(const QRect &rect, QItemSelectionModel::SelectionFlags command)

    don't support any selection.
*/
void BoardView::setSelection(const QRect &, QItemSelectionModel::SelectionFlags )
{
   return;
}

QRegion BoardView::visualRegionForSelection(const QItemSelection &) const
{
    return QRegion();
}

/*!
    Returns true if the given \a row is hidden; otherwise returns false.
    
    \sa setRowHidden
*/
bool BoardView::isRowHidden(int row) const
{
    return vHeader->isSectionHidden(row);
}

/*!
    Returns true if the given \a column is hidden; otherwise returns false.

    \sa setColumnHidden
*/
bool BoardView::isColumnHidden(int column) const
{
    return horizontalHeader()->isSectionHidden(column);
}

/*!
    Returns the y-coordinate in contents coordinates of the given \a
    row.
*/
int BoardView::rowViewportPosition(int row) const
{
    return vHeader->sectionViewportPosition(row);
}

/*!
    Returns the x-coordinate in contents coordinates of the given \a
    column.
*/
int BoardView::columnViewportPosition(int column) const
{
    return horizontalHeader()->sectionViewportPosition(column);
}


/*!
    \since 4.1

    Sets the width of the given \a column to be \a width.
*/
// void BoardView::setColumnWidth(int column, int width)
// {
//     horizontalHeader()->resizeSection(column, width);
// }

/*!
    Returns the width of the given \a column.

    \sa resizeColumnToContents()
*/
int BoardView::columnWidth(int column) const
{
    return horizontalHeader()->sectionSize(column);
}



/*!
    Returns the height of the given \a row.

    \sa resizeRowToContents()
*/
int BoardView::rowHeight(int row) const
{
    return verticalHeader()->sectionSize(row);
}

/*!
    Returns the row in which the given y-coordinate, \a y, in contents
    coordinates is located.
*/
int BoardView::rowAt(int y) const
{
    return verticalHeader()->logicalIndexAt(y);
}
/*!
    Returns the column in which the given x-coordinate, \a x, in contents
    coordinates is located.
*/
int BoardView::columnAt(int x) const
{
    return horizontalHeader()->logicalIndexAt(x);
}

QHeaderView *BoardView::horizontalHeader() const
{
   return hHeader;
}

QHeaderView *BoardView::verticalHeader() const
{
   return vHeader;
}

/*!
    Paints the table on receipt of the given paint event \a event.
*/
void BoardView::paintEvent(QPaintEvent *event)
{
   // setup temp variables for the painting

    QStyleOptionViewItem option = viewOptions();
    const QPoint offset = QPoint();//scrollDelayOffset;// what the hell the offset is???
    //const bool m_showGrid = showGrid();
    const int gridSize = m_showGrid ? 1 : 0;
    const int gridHint = style()->styleHint(QStyle::SH_Table_GridLineColor, &option, this);
    const QColor gridColor = static_cast<QRgb>(gridHint);
    const QPen gridPen = QPen(gridColor, 0, m_gridStyle);
    const QItemSelectionModel *sels = selectionModel();
    //const QHeaderView *verticalHeader = vHeader;
    //const QHeaderView *horizontalHeader = hHeader;
    const QModelIndex current = currentIndex();

    const bool focus = (hasFocus() || viewport()->hasFocus()) && current.isValid();
    const QStyle::State state = option.state;

    QPainter painter( viewport() );

    // if there's nothing to do, clear the area and return
    //QAbstractItemModel *m=model();
    //int hCnt=hHeader->count();
    //int vCnt=verticalHeader->count();
    
    if (!model() || hHeader->count() == 0 || vHeader->count() == 0)
    {
        painter.fillRect(event->rect(), option.palette.brush(QPalette::Base));
        return;
    }

    QVector<QRect> rects = event->region().rects();
    for (int i = 0; i < rects.size(); ++i) 
    {
        QRect area = rects.at(i);
        area.translate(offset);

        // get the horizontal start and end sections (visual indexes)
        int left = hHeader->visualIndexAt(area.left());
        int right = hHeader->visualIndexAt(area.right());

        if (isRightToLeft()) {
            left = (left == -1 ? model()->columnCount(rootIndex()) - 1 : left);
            right = (right == -1 ? 0 : right);
        } else {
            left = (left == -1 ? 0 : left);
            right = (right == -1 ? model()->columnCount(rootIndex()) - 1 : right);
        }

        int tmp = left;
        left = qMin(left, right);
        right = qMax(tmp, right);

        // get the vertical start and end sections (visual indexes)
        int bottom = vHeader->visualIndexAt(area.bottom());
        bottom = (bottom == -1 ? model()->rowCount(rootIndex()) - 1 : bottom);
        if (bottom == -1)
            return; // empty model

        int top = 0;

        top = vHeader->visualIndexAt(area.top());
        if (top == -1)
             return; // no visible rows inside area

        Q_ASSERT(top >= 0 && top <= bottom);

        // do the actual painting
        for (int v = top; v <= bottom; ++v) {
            int row = vHeader->logicalIndex(v);
            if(row<0)continue;// FIXME
            if (vHeader->isSectionHidden(row))
                continue;
                
            int rowp = rowViewportPosition(row) + offset.y();
            int rowh = rowHeight(row) - gridSize;
            for (int h = left; h <= right; ++h) 
            {
                int col = hHeader->logicalIndex(h);
                if(col<0)continue;//  FIXME
                if (hHeader->isSectionHidden(col))
                    continue;
                int colp = columnViewportPosition(col) + offset.x();
                int colw = columnWidth(col) - gridSize;
                
                int col5=col;
                int row5=row;
                if( m_inverted )
                {
                   col5 = hHeader->count()-1-col;
                   row5 = vHeader->count()-1-row;
                }
                QModelIndex index = model()->index( row5, col5, rootIndex());
                if (index.isValid()) 
                {
                    option.rect = QRect(colp, rowp, colw, rowh);
                    option.state = state;
                    if (sels && sels->isSelected(index))
                        option.state |= QStyle::State_Selected;
//                     if (index == hover)
//                         option.state |= QStyle::State_MouseOver;
//                     else
//                         option.state &= ~QStyle::State_MouseOver;
                    QPalette::ColorGroup cg;
                    if ((model()->flags(index) & Qt::ItemIsEnabled) == 0) {
                        option.state &= ~QStyle::State_Enabled;
                        cg = QPalette::Disabled;
                    } else {
                        cg = QPalette::Normal;
                    }
                    if (focus && index == current)
                        option.state |= QStyle::State_HasFocus;
                        painter.fillRect(colp, rowp, colw, rowh,
                                         option.palette.brush(QPalette::Base));
                    itemDelegate()->paint(&painter, option, index);
                }
                if (v == top && m_showGrid) 
                {
                    QPen old = painter.pen();
                    painter.setPen(gridPen);
                    painter.drawLine(colp + colw, area.top(), colp + colw, area.bottom() + 1);
                    painter.setPen(old);
                }
            }
            if (m_showGrid) {
                QPen old = painter.pen();
                painter.setPen(gridPen);
                painter.drawLine(area.left(), rowp + rowh, area.right(), rowp + rowh);
                painter.setPen(old);
            }
        }

        int w = viewport()->width();
        int h = viewport()->height();
        int x = hHeader->length();
        int y = vHeader->length();
        QRect b(0, y, w, h - y);
        if (y < h && area.intersects(b))
            painter.fillRect(b, option.palette.brush(QPalette::Base));
        if (isRightToLeft()) 
        {
            QRect r(0, 0, w - x, h);
            if (x > 0 && area.intersects(r))
                painter.fillRect(r, option.palette.brush(QPalette::Base));
        } else 
        {
            QRect l(x, 0, w - x, h);
            if (x < w && area.intersects(l))
                painter.fillRect(l, option.palette.brush(QPalette::Base));
        }
    }
}


/*!
  \property QTableView::gridStyle
  \brief  the pen style used to draw the grid.

  This property holds the style used when drawing the grid (see \l{showGrid}).
*/
Qt::PenStyle BoardView::gridStyle() const
{
    return m_gridStyle;
}

void BoardView::setGridStyle(Qt::PenStyle style)
{
    if (gridStyle() != style) 
    {
        m_gridStyle = style;
        update();
    }
}

/*!
  \reimp
*/
void BoardView::setModel(QAbstractItemModel *model)
{
    vHeader->setModel(model);
    hHeader->setModel(model);
    vHeader2->setModel(model);
    hHeader2->setModel(model);
    QAbstractItemView::setModel(model);
}

void BoardView::updateGeometries()
{
    int width = !vHeader->isHidden() ? vHeader->sizeHint().width() : 0;
    int height = !hHeader->isHidden() ? hHeader->sizeHint().height() : 0;
    bool reverse = isRightToLeft();
    setViewportMargins(reverse ? 0 : width, height, reverse ? width : 0, 0);

    QRect vg = viewport()->geometry();

    int verticalLeft = reverse ? vg.right() : (vg.left() - width);
    vHeader->setGeometry(verticalLeft, vg.top(), width, vg.height());
    if (vHeader->isHidden())
        QMetaObject::invokeMethod(vHeader, "updateGeometries");
    vHeader->setOffset( verticalScrollBar()->value() );
    
    int horizontalTop = vg.top() - height;
    hHeader->setGeometry(vg.left(), horizontalTop, vg.width(), height);
    if (hHeader->isHidden())
        QMetaObject::invokeMethod( hHeader, "updateGeometries");
    hHeader->setOffset( horizontalScrollBar()->value() );

    // additional headers -----------------------------------
    int hw=hHeader->width();
    int vh =vHeader->height();
    //int grid=m_showGrid?1:0;
    
    vHeader2->setGeometry(verticalLeft+hw+4, vg.top(), width, vg.height());
    if (vHeader2->isHidden())
        QMetaObject::invokeMethod(vHeader2, "updateGeometries");
    
    hHeader2->setGeometry(vg.left(), horizontalTop+vh+4, vg.width(), height);
    if (hHeader2->isHidden())
        QMetaObject::invokeMethod( hHeader2, "updateGeometries");
    
    if (model()) 
    {
        //updateVerticalScrollbar();
        //updateHorizontalScrollbar();
    }

    QAbstractItemView::updateGeometries();
}

/*!
    This slot is called to change the height of the given \a row. The
    old height is specified by \a oldHeight, and the new height by \a
    newHeight.

    \sa columnResized()
*/
void BoardView::rowResized(int row, int, int)
{
    vHeader->resizeSection( row , m_cellSizeHint.height() );
}
void BoardView::row2Resized(int row, int, int)
{
    vHeader2->resizeSection( row , m_cellSizeHint.height() );
}

/*!
    This slot is called to change the width of the given \a column.
    The old width is specified by \a oldWidth, and the new width by \a
    newWidth.

    \sa rowResized()
*/
void BoardView::columnResized(int column, int, int)
{
    hHeader->resizeSection( column,m_cellSizeHint.width() );
}
void BoardView::column2Resized(int column, int, int)
{
    hHeader2->resizeSection( column,m_cellSizeHint.width() );
}

void BoardView::invertBoard( bool isUpsideDown )
{
    m_inverted=isUpsideDown;
    hide();
    show();
}

#include "boardview.moc"
