/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _FRUIT_H
#define _FRUIT_H

#include "GameMath.h"
#include "Particles.h"
#include "Instance.h" // temp

#include <Eigen/Core>
#include <QList>

class Model3D;
class Instance;
class HumanPlayer;

static const float FRUITATTACH[] = {-0.29,  0.16,  0.37, -71.83,
                    0.38, -0.06,  0.32, -12.29,
                    0.14,  0.47, -0.09,  43.24,
                    0   ,  0   ,  0   ,  1};
static const float FRUITBASKETHEIGHT = 1.0;
static const float FRUITBASKETSCALE = 0.07;

enum FruitType {
    ftApple,
    ftOrange,
    ftLemon,
    ftAnanas,
    ftStraw
};

class Fruit {
public:
    Fruit(Model3D* model, FruitType kind, int owner);
    Fruit();
    ~Fruit();

    void update();
    void update2();
    void render();
    void startWarp();
    void setBasketWarp(int bid);
    void disableWarpClip();
    void warpRender();
    void warpRender2();

    void scale(const Eigen::Vector3f& scale);
    void rotate(const Eigen::Vector3f& rot);
    void rotateX(float ang);
    void rotateY(float ang);
    void rotateZ(float ang);
    void position(const Eigen::Vector3f& pos);
    void positionBS(const Eigen::Vector3f& pos);
    Eigen::Vector3f scale();
    Eigen::Vector3f rotation();
    Eigen::Vector3f position();
    BS colBS();
    float midHeight();
    bool fruitMoved();

    void setId(int id) { this->m_id = id; }
    const Warp& warp() const { return m_warp; }
    void setMoveSpeed(const Eigen::Vector3f& moveSpeed) { this->m_moveSpeed = moveSpeed; }
    void setFallSpeed(float sp) { m_fallSpeed = sp; }
    void setVanish(float v) { m_vanish = v; }
    void setVanishing(bool v) { m_vanishing = v; }
    void setOwner(int owner) { this->m_owner = owner; }
    void setRemoteId(int id) { this->m_remoteId = id; }
    const Instance* instance() const { return m_inst; }
    Instance* instance() { return m_inst; }
    void setCarriedBy(int c) { m_carriedBy = c; }
    bool isGroundColOn() const { return m_groundColOn; }
    float vanish() const { return m_vanish; }
    int carriedBy() const { return m_carriedBy; }
    void setGroundColOn(bool b) { m_groundColOn = b; }
    Eigen::Vector3f moveSpeed() const { return m_moveSpeed; }
    float fallSpeed() const { return m_fallSpeed; }
    int score() const { return m_score; }
    FruitType kind() const { return m_kind; }
    bool isAttached() const { return m_inst->getAttach().attached; }
private:
    Instance* m_inst;
    int m_id;
    FruitType m_kind;
    BS m_colBS;
    int m_carriedBy;
    float m_grav;
    float m_fallSpeed;
    Eigen::Vector3f m_moveSpeed;
    unsigned int m_lastTime;
    float m_dt;
    float m_midHeight;
    Warp m_warp;
    bool m_groundColOn;
    int m_score;
    bool m_vanishing;
    float m_vanish;
    int m_owner;
    int m_remoteId;
    Eigen::Vector3f m_oldPos;
};

class FruitBasket {
public:
    FruitBasket(Model3D* model);
    ~FruitBasket();

    void addFruit(FruitType fid);
    void render();
    void addFruit(const Fruit* fruit) { addFruit(fruit->kind()); addScore(fruit->score()); starShow.activate(); }
    void scale(const Eigen::Vector3f& scale);
    void rotate(const Eigen::Vector3f& rot);
    void rotateX(float ang);
    void rotateY(float ang);
    void rotateZ(float ang);
    void position(const Eigen::Vector3f& pos);
    void positionBS(const Eigen::Vector3f& pos);
    Eigen::Vector3f getScale();
    Eigen::Vector3f getRot();
    Eigen::Vector3f getPos();
    BS getColBS();
    float getMidHeight();
    int getFruitID(int bid);
    inline int& getScore() { return score; }
    inline StarShow& getStarShow() { return starShow; }
    inline Instance* getInst() { return inst; }
    inline void setPlayId(int i) { playId = i; }
    inline void setTeamId(int i) { teamId = i; }
    void addScore(int delta) { score += delta; }
private:
    Instance* inst;
    int playId;
    int teamId;
    QList<int> contain;
    BS colBS;
    float midHeight;
    int score;
    StarShow starShow;
    int getContain(int id);
    void setContain(int id, int value);
};

struct FruitData {
    bool inGame;
    bool respawning;
};

class FruitSpawner {
public:
    FruitSpawner(int minCount, int maxCount, float minWait, float maxWait);
    FruitSpawner();
    ~FruitSpawner();

    void render();
    void warpRender();
    void warpRender2();
    void update();
    void spawnFruit();
    void spawnFruit(int type, int owner, const Eigen::Vector3f& pos, int id);
    void respawnFruit(int fid);

    void ownFruits(const HumanPlayer& owner);

    Fruit& getFruit(int n);
    Model3D* getFruitModel(int n);
    FruitData getFruitData(int n);
    inline int getMaxCount() const { return maxCount; }
    int getFruitDataCount() const { return maxCount * 5; }
private:
    int minCount;
    int maxCount;
    float minWait;
    float maxWait;
    QList<Fruit> fruit;
    QList<Model3D*> fruitModel;
    QList<FruitData> fruitData;
    int inGameCount;
    int thePrime;
    unsigned int lastTime;
    float dt;
};

#endif // _FRUIT_H
