/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "GameMath.h"

// FruitWars headers
#include "Math3D.h"

// System headers
#include <GL/glu.h>

USING_PART_OF_NAMESPACE_EIGEN
using Eigen::Quaternionf;
using Eigen::Transform3f;

Plane Plane::normalize() const
{
    float l = 1.00 / sqrt(sqr(a) + sqr(b) + sqr(c));
    return Plane(a * l, b * l, c * l, d * l);
}

void AABB::draw() const
{
    glEnable(GL_COLOR_MATERIAL);
    glColor3f(1.0, 1.0, 1.0);
    glBegin(GL_LINE_STRIP);
        glVertex3f(leftBottom.x(), leftBottom.y(), leftBottom.z());
        glVertex3f(rightTop.x(), leftBottom.y(), leftBottom.z());
        glVertex3f(rightTop.x(), rightTop.y(), leftBottom.z());
        glVertex3f(leftBottom.x(), rightTop.y(), leftBottom.z());
        glVertex3f(leftBottom.x(), leftBottom.y(), leftBottom.z());
    glEnd();
    glBegin(GL_LINE_STRIP);
        glVertex3f(leftBottom.x(), leftBottom.y(), rightTop.z());
        glVertex3f(rightTop.x(), leftBottom.y(), rightTop.z());
        glVertex3f(rightTop.x(), rightTop.y(), rightTop.z());
        glVertex3f(leftBottom.x(), rightTop.y(), rightTop.z());
        glVertex3f(leftBottom.x(), leftBottom.y(), rightTop.z());
    glEnd();
    glBegin(GL_LINES);
        glVertex3f(leftBottom.x(), leftBottom.y(), leftBottom.z());
        glVertex3f(leftBottom.x(), leftBottom.y(), rightTop.z());
        glVertex3f(leftBottom.x(), rightTop.y(), leftBottom.z());
        glVertex3f(leftBottom.x(), rightTop.y(), rightTop.z());
        glVertex3f(rightTop.x(), rightTop.y(), leftBottom.z());
        glVertex3f(rightTop.x(), rightTop.y(), rightTop.z());
        glVertex3f(rightTop.x(), leftBottom.y(), leftBottom.z());
        glVertex3f(rightTop.x(), leftBottom.y(), rightTop.z());
    glEnd();
    glDisable(GL_COLOR_MATERIAL);
}


OBB::OBB(const AABB &bb, const Vector3f &pos, const Vector3f &scale, const Quaternionf &rot)
{
    Transform3f trans;
    trans.fromPositionOrientationScale(pos, rot, scale);
    points[0] = trans * Vector3f(bb.leftBottom.x(), bb.leftBottom.y(), bb.leftBottom.z());
    points[1] = trans * Vector3f(bb.rightTop.x()  , bb.leftBottom.y(), bb.leftBottom.z());
    points[2] = trans * Vector3f(bb.leftBottom.x(), bb.rightTop.y()  , bb.leftBottom.z());
    points[3] = trans * Vector3f(bb.rightTop.x()  , bb.rightTop.y()  , bb.leftBottom.z());
    points[4] = trans * Vector3f(bb.leftBottom.x(), bb.leftBottom.y(), bb.rightTop.z());
    points[5] = trans * Vector3f(bb.rightTop.x()  , bb.leftBottom.y(), bb.rightTop.z());
    points[6] = trans * Vector3f(bb.leftBottom.x(), bb.rightTop.y()  , bb.rightTop.z());
    points[7] = trans * Vector3f(bb.rightTop.x()  , bb.rightTop.y()  , bb.rightTop.z());
}

OBB::OBB(const AABB &bb)
{
    points[0] = Vector3f(bb.leftBottom.x(), bb.leftBottom.y(), bb.leftBottom.z());
    points[1] = Vector3f(bb.rightTop.x()  , bb.leftBottom.y(), bb.leftBottom.z());
    points[2] = Vector3f(bb.leftBottom.x(), bb.rightTop.y()  , bb.leftBottom.z());
    points[3] = Vector3f(bb.rightTop.x()  , bb.rightTop.y()  , bb.leftBottom.z());
    points[4] = Vector3f(bb.leftBottom.x(), bb.leftBottom.y(), bb.rightTop.z());
    points[5] = Vector3f(bb.rightTop.x()  , bb.leftBottom.y(), bb.rightTop.z());
    points[6] = Vector3f(bb.leftBottom.x(), bb.rightTop.y()  , bb.rightTop.z());
    points[7] = Vector3f(bb.rightTop.x()  , bb.rightTop.y()  , bb.rightTop.z());
}

void BS::draw() const
{
    GLUquadric *bol = gluNewQuadric();
    glPushMatrix();
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_COLOR, GL_ONE);
        glTranslatef(pos.x(), pos.y(), pos.z());
        gluSphere(bol, radius, 20, 20);
        glDisable(GL_BLEND);
    glPopMatrix();
    gluDeleteQuadric(bol);
}

Frustum::Frustum()
{
}

Frustum Frustum::calc() {
    Frustum f;
    Transform3f modelView, proj;
    Matrix4f clip;
    glGetFloatv(GL_PROJECTION_MATRIX, proj.data());
    glGetFloatv(GL_MODELVIEW_MATRIX, modelView.data());
    clip = (proj * modelView).matrix();

    f.plane[0].a = clip[3] + clip[0];
    f.plane[0].b = clip[7] + clip[4];
    f.plane[0].c = clip[11] + clip[8];
    f.plane[0].d = clip[15] + clip[12];
    f.plane[0] = f.plane[0].normalize();

    f.plane[1].a = clip[3] - clip[0];
    f.plane[1].b = clip[7] - clip[4];
    f.plane[1].c = clip[11] - clip[8];
    f.plane[1].d = clip[15] - clip[12];
    f.plane[1] = f.plane[1].normalize();

    f.plane[2].a = clip[3] + clip[1];
    f.plane[2].b = clip[7] + clip[5];
    f.plane[2].c = clip[11] + clip[9];
    f.plane[2].d = clip[15] + clip[13];
    f.plane[2] = f.plane[2].normalize();

    f.plane[3].a = clip[3] - clip[1];
    f.plane[3].b = clip[7] - clip[5];
    f.plane[3].c = clip[11] - clip[9];
    f.plane[3].d = clip[15] - clip[13];
    f.plane[3] = f.plane[3].normalize();

    f.plane[4].a = clip[3] + clip[2];
    f.plane[4].b = clip[7] + clip[6];
    f.plane[4].c = clip[11] + clip[10];
    f.plane[4].d = clip[15] + clip[14];
    f.plane[4] = f.plane[4].normalize();

    f.plane[5].a = clip[3] - clip[2];
    f.plane[5].b = clip[7] - clip[6];
    f.plane[5].c = clip[11] - clip[10];
    f.plane[5].d = clip[15] - clip[14];
    f.plane[5] = f.plane[5].normalize();

    return f;
}

float distToPlane(const Plane &plane, const Vector3f &pos)
{
    return plane.a * pos.x() + plane.b * pos.y() + plane.c * pos.z() + plane.d;
}

bool bsInFrustum(const Frustum &frustum, const BS &bs)
{
    return true;
    for (int p = 0; p < 6; p++) {
        if (distToPlane(frustum.plane[p], bs.pos) + bs.radius <= 0) {
            return false;
        }
    }
    return true;
}

bool aabbInFrustum(const Frustum &frustum, const AABB &bb)
{
    return true;
    return obbInFrustum(frustum, OBB(bb));
}

bool obbInFrustum(const Frustum &frustum, const OBB &obb)
{
    return true;
    for (int p = 0; p < 6; p++) {
        bool between = true;
        for (int q = 0; q < 8; q++) {
            if (distToPlane(frustum.plane[p], obb.points[q]) > 0) {
                between = false;
            }
        }
        if (between) {
            return false;
        }
    }
    return true;
}

bool pointInFrustum(const Frustum &frustum, const Vector3f &point)
{
    for (int p = 0; p < 6; p++) {
        if (distToPlane(frustum.plane[p], point) <= 0) {
            return false;
        }
    }
    return true;
}
