/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _INSTANCE_H
#define _INSTANCE_H

#include "Array.h"
#include "GameMath.h"

#include <QVector>
#include <Eigen/Core>

class Instance;
class Model3D;
class StaticModel;
class FrameModel;
class AnimModel;
class BoneModel;

static const int MAXKEYVERTS = 32;
extern Frustum instFrust;

class HelpFrame {
    public:
        float blendMult, blendSum;
        int priority;
        Eigen::Vector3f animPos, animScale;
        Eigen::Quaternionf animRot;
        Eigen::Transform3f transform;

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

class BoneInstance;

class BoneAttach {
    public:
        BoneInstance *inst;
        int frame;
        Eigen::Transform3f offset;
        bool attached;

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

class KeyVert {
    public:
        int meshID;
        int vertID;
        Eigen::Vector3f pos;
};

class Instance {
    protected:
        Eigen::Vector3f pos;
        Eigen::Vector3f rot;
        Eigen::Vector3f scale;
        QVector<int> texture;
        BoneAttach attach;
        BS bs;
    public:
        static Instance* load(Model3D *model);
        Instance();
        virtual ~Instance();
        virtual void render();
        virtual void setFrame(unsigned int set, float time);
        virtual void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        virtual void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        virtual void calcAnim();
        void setTexture(unsigned int mat, int tex);
        void position(const Eigen::Vector3f &pos);
        void scaleObj(const Eigen::Vector3f &scale);
        void rotateX(float xAng);
        void rotateY(float yAng);
        void rotateZ(float zAng);
        void rotate(const Eigen::Vector3f &rot);
        void setBS(const BS &bs);
        void attachToFrame(BoneInstance *inst, int frame);
        void specialAttach(BoneInstance *inst, int frame, const Eigen::Transform3f &offset);
        void detach();
        Eigen::Transform3f transform();
        Eigen::Transform3f attachTransform();
        int getTex(unsigned int n);
        Eigen::Vector3f getPos();
        Eigen::Vector3f getRot();
        Eigen::Vector3f getScale();
        BoneAttach getAttach();
        BS getBS();

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

class StaticInstance : public Instance {
    private:
        StaticModel *model;
    public:
        StaticInstance(StaticModel *model);
        void render();
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        StaticModel* getModel();
};

class FrameInstance : public Instance {
    private:
        FrameModel *model;
    public:
        FrameInstance(FrameModel *model);
        void render();
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        FrameModel* getModel();
};

class AnimInstance : public Instance {
    private:
        AnimModel *model;
        Array<HelpFrame> frame;
    public:
        AnimInstance(AnimModel *model);
        void render();
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        AnimModel* getModel();
};

class BoneInstance : public Instance {
    private:
        BoneModel *model;
        Array<HelpFrame> frame;
        int nKeyVerts;
        KeyVert *keyVert;
        void calcCombinedMatrix(int n);
    public:
        BoneInstance(BoneModel *model);
        BoneInstance();
        ~BoneInstance();
        void render();
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        void addKeyVert(int mid, int vid);
        Eigen::Vector3f getKeyPos(int kid) const;
        HelpFrame getFrame(int fid);
        BoneModel* getModel();
        int getKeyVertCount();

        friend class Instance;
        friend class StaticInstance;
        friend class FrameInstance;
        friend class AnimInstance;
};

#endif // _INSTANCE_H
