/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _PARTICLES_H
#define _PARTICLES_H

#include <Eigen/Core>
#include <Eigen/Geometry>
#include <GL/gl.h>
#include <vector>

class Camera;
class ThirdPersonCamera;

static const float WARPDIF = 5.0;
static const float WARPWAITTIME = 0.50;
static const float WARPSCALESPEED = 2.0;
static const float AIMSIZE = 0.70;
static const float AIMSMOOTH = 0.15;
static const float TERRAINAIMSTEPSIZE = 0.25;
static const float KEVILCURVEC = 2000.0;
static const float KEVILCURVEK = 1.00;
static const float KEVILCURVEMW = 100.0;
static const float KEVILCURVED = 0.10;
static const float STARSIZE = 0.25;
static const float STARSPINSPEED = 2.0 * M_PI;
static const float STARRELPOS[] = {-10.0, 45.0, -35.0};
static const float STARSHOWSIZE = 0.25;

extern GLuint warpTex;
extern GLuint shadowTex;
extern GLuint aimTex;
extern GLuint starTex;
extern GLuint trailTex;
extern GLuint starShowTex[3];
extern GLuint indicTex;

class BillBoard {
private:
    Eigen::Vector3f pos;
    float spin;
    float size;
    Eigen::Vector2f scale;
    int tex;
    bool blended;
    unsigned long int blendSFac;
    unsigned long int blendDFac;
public:
    BillBoard(const Eigen::Vector3f& pos, float size, int tex);
    BillBoard();
    void render(const Camera& cam) const;

    inline float getSize() const { return size; }
    inline void setSize(float size) { this->size = size; }
    inline Eigen::Vector2f getScale() const { return scale; }
    inline void setScale(const Eigen::Vector2f& scale) { this->scale = scale; }
    inline Eigen::Vector3f getPos() const { return pos; }
    inline void setPos(const Eigen::Vector3f& pos) { this->pos = pos; }
    inline float getSpin() const { return spin; }
    inline void setSpin(float spin) { this->spin = spin; }
    inline int getTex() const { return tex; }
    inline void setTex(int tex) { this->tex = tex; }
    inline bool getBlended() const { return blended; }
    inline void setBlended(bool blended) { this->blended = blended; }
    inline unsigned long int getBlendSFac() const { return blendSFac; }
    inline void setBlendSFac(unsigned long int blendSFac) { this->blendSFac = blendSFac; }
    inline unsigned long int getBlendDFac() const { return blendDFac; }
    inline void setBlendDFac(unsigned long int blendDFac) { this->blendDFac = blendDFac; }
};

class Warp {
private:
    Eigen::Vector3f pos;
    float size;
    float scale;
    int state;
    float scaleSpeed;
    float waitTime;
    int clipSide;
public:
    Warp(const Eigen::Vector3f& pos, float size);
    ~Warp();
    void render();
    void update(float dt);

    inline Eigen::Vector3f getPos() const { return pos; }
    inline void setPos(const Eigen::Vector3f& pos) { this->pos = pos; }
    inline float getSize() const { return size; }
    inline void setSize(float size) { this->size = size; }
    inline float getScale() const { return scale; }
    inline void setScale(float scale) { this->scale = scale; }
    inline int getState() const { return state; }
    inline void setState(int state) { this->state = state; }
    inline float getScaleSpeed() const { return scaleSpeed; }
    inline void setScaleSpeed(float scaleSpeed) { this->scaleSpeed = scaleSpeed; }
    inline float getWaitTime() const { return waitTime; }
    inline void setWaitTime(float waitTime) { this->waitTime = waitTime; }
    inline int getClipSide() const { return clipSide; }
    inline void setClipSide(int clipSide) { this->clipSide = clipSide; }
};

class Aim {
private:
    BillBoard bill;
    float depth;
    float tDepth;
    float terDepth;
    float smooth;
    Eigen::Transform3f viewMat;
    Eigen::Transform3f projMat;
    unsigned int playId;
public:
    Aim(unsigned int id);
    Aim();
    void render();
    void calcAim();

    EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

class Stars {
private:
    int playId;
    float size;
    float spin;
    BillBoard bill[4];
    float alpha;
    Eigen::Transform3f trans;
public:
    Stars(int id, float size);
    ~Stars();
    void render();
    void update(float dt);
    inline void setTrans(const Eigen::Transform3f& trans) { this->trans = trans; }
    inline void setAlpha(float a) { alpha = a; }

    EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

struct TrailSeg {
    Eigen::Vector3f segA, segB;
};

class Trail {
private:
    int trailBegin;
    std::vector<TrailSeg> trails;
public:
    Trail(int count);
    ~Trail();
    void update(const Eigen::Vector3f& pos1, const Eigen::Vector3f& pos2);
    void render();
};

struct StarPart {
    float angle;
    float radius;
    float height;
    float hSpeed;
    float angSpeed;
    float acc;
    float alpha;
    float alphaSpeed;
    BillBoard bill;
};

class StarShow {
private:
    std::vector<StarPart> parts;
    float maxRadius;
    Eigen::Vector3f spawnPos;
    float dt;
    unsigned int lastTime;
    bool active;
public:
    StarShow(const Eigen::Vector3f& pos, float rad, int count);
    ~StarShow();
    void activate();
    void update();
    void render();

    inline void setSpawnPos(const Eigen::Vector3f& spawnPos) { this->spawnPos = spawnPos; }
};

class Indicator {
private:
    float spin;
    BillBoard bill;
    int playId;
    int baskId;
    bool show;
public:
    Indicator(float size, int pid, int bid);
    Indicator();
    ~Indicator();
    void render();
};

extern void createWarpTex();
extern void createShadowTex();
extern void createCartoonShadowTex();
extern void createAimTex();
extern void createStarTex();
extern void createStarShowTex();
extern void createTrailTex();
extern void createIndicTex();
extern void createCartoonTex();

#endif // _PARTICLES_H
