/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Players.h"

// FruitWars headers
#include "Camera.h"
#include "Globals.h"
#include "Instance.h"
#include "Math3D.h"
#include "Particles.h"

USING_PART_OF_NAMESPACE_EIGEN
using Eigen::AngleAxisf;
using Eigen::Scaling3f;
using Eigen::Transform3f;
using Eigen::Translation3f;

HumanPlayer::HumanPlayer(Model3D* model, int id)
    : dt(0.0)
    , lastTime(time())
    , grav(GLOBALGRAV)
    , jumpForce(12.0)
    , jumpSpeed(0.0)
    , dodgeForce(10.0)
    , dodgeSpeed(0, 0, 0)
    , lastDodgeTime(time())
    , moveSpeed(PLAYERWALKSPEED)
    , turnSpeed(1.0)
    , isJumping(false)
    , playId(id)
    , teamId(-1)
    , holds(-1)
    , stunTime(0.0)
    , dead(false)
    , deadScale(1.0)
    , isAI(false)
    , baskId(0)
    , relTeamId(0)
    , colorId(0)
    , banana(bananaModel, id)
    , stars(id, 25.0)
    , warp(Vector3f(0, 0, 0), PLAYERWARPSIZE)
    , holdTime(0)
    , nickName(QString("Player %1").arg(id))
{
    inst = (BoneInstance*) Instance::load(model);
    anim.resize(inst->getModel()->animSets.size());
    inst->setBS(BS(Vector3f(PLAYERBSOFFSET[0], PLAYERBSOFFSET[1], PLAYERBSOFFSET[2]), PLAYERBSRAD));
    oldPos = inst->getPos();
    colBS = inst->getBS();
    colBS.radius = PLAYERCOLRADIUS;
    for (unsigned int i = 0; i < anim.size(); i++) {
        anim[i].frame = 0.0;
        anim[i].tFrame = 0.0;
        anim[i].frameSmooth = 0.85;
        anim[i].frameSpeed = 20.0;
        anim[i].dFrame = 0.0;
        anim[i].blend = 0.0;
        anim[i].setId = i;
        anim[i].priority = 0;
        anim[i].maxFrame = 19.0;
    }
    anim[0].blend = 1.0;
    anim[PLSTUNANIM].frameSpeed = PLSTUNANIMFRAMESPEED;
    anim[PLSTUNANIM].maxFrame = 29.0;

    inst->addKeyVert(0, PLAYERLEFTARMVERT);
    inst->addKeyVert(0, PLAYERRIGHTARMVERT);
    //inst->setFrame(0, 0);
    //inst->calcAnim();

    warp.setState(3);
}

HumanPlayer::HumanPlayer()
    : banana(bananaModel, 0)
    , stars(0, 25.0)
    , warp(Vector3f(0, 0, 0), PLAYERWARPSIZE)
{
}

HumanPlayer::~HumanPlayer()
{
}

void HumanPlayer::startJump()
{
    jumpSpeed = jumpForce;
}

void HumanPlayer::strafe(float speed)
{
    Vector3f dir = getDirVec();
    dir = Vector3f(-dir.z(), 0.0, dir.x());
    position(getPos() + dir * speed);
}

void HumanPlayer::turn(float angle)
{
    rotateY(getRot().y() + angle);
}

void HumanPlayer::move(float speed)
{
    position(getPos() + getDirVec() * speed);
}

void HumanPlayer::dodgeRight()
{
    Vector3f dir = getDirVec();
    dir = Vector3f(-dir.z(), 0.0, dir.x());
    dodgeSpeed = dir * dodgeForce;
}

void HumanPlayer::dodgeLeft()
{
    Vector3f dir = getDirVec();
    dir = Vector3f(dir.z(), 0.0, -dir.x());
    dodgeSpeed = dir * dodgeForce;
}

void HumanPlayer::dodgeForward()
{
    Vector3f dir = getDirVec();
    dodgeSpeed = dir * dodgeForce;
}

void HumanPlayer::dodgeBackward()
{
    Vector3f dir = getDirVec();
    dir = Vector3f(-dir.x(), 0.0, -dir.z());
    dodgeSpeed = dir * dodgeForce;
}

void HumanPlayer::pickUpFruit(int fid)
{
    Transform3f m = fruitAttachMat;
    Vector3f pos = fruitSpawner.getFruit(fid).instance()->getBS().pos * -1;
    m = m * Scaling3f(fruitSpawner.getFruit(fid).scale()) * Translation3f(pos);
    fruitSpawner.getFruit(fid).instance()->specialAttach(inst, PLAYERLEFTARMFRAME, m);
}

void HumanPlayer::tossFruit()
{
    if (holds >= 0) {
        fruitSpawner.getFruit(holds).setCarriedBy(-2);
        fruitSpawner.getFruit(holds).instance()->detach();
        fruitSpawner.getFruit(holds).setFallSpeed(FRUITTHROWSPEED + throwInfluence * jumpSpeed);
        fruitSpawner.getFruit(holds).setMoveSpeed(getDirVec() * FRUITMOVESPEED + dodgeSpeed * throwInfluence);
        holds = -1;
        // XXX send TOSSFRUIT msg
    }
}

void HumanPlayer::respawn()
{
    banana.getIt();
    Vector3f thePos;
    if (spawnRandom) {
        float leftBX = myTer.getPos().x();
        float leftBZ = myTer.getPos().z();
        float rightBX = myTer.getPos().x() + myTer.getXSegs() * myTer.getXSegSize();
        float rightBZ = myTer.getPos().z() - myTer.getZSegs() * myTer.getZSegSize();

        float h = SEAHEIGHT - 1;
        int check = 0;
        while (h < SEAHEIGHT && check < 10000) {
            thePos = Vector3f(leftBX + frandom() * (rightBX - leftBX), 0.0, leftBZ + frandom() * (rightBZ - leftBZ));
            h = myTer.getGroundHeight(Vector2f(thePos.x(), thePos.z()));
            check++;
        }
    } else {
        thePos = myBaskets[baskId].getPos() * (1.0 - ((myBaskets[baskId].getColBS().radius + getColBS().radius) / myBaskets[baskId].getPos().norm()));
        if (teamMode < 2) {
            Transform3f rotMat(AngleAxisf(((float) relTeamId / myTeams[teamId].getPlayerCount()) * 2.0 * M_PI, Vector3f::UnitY()));
            Transform3f posMat1(Translation3f(myBaskets[baskId].getPos()));
            Transform3f posMat2(Translation3f(myBaskets[baskId].getPos() * -1));
            Transform3f theMat(/*fastMul(fastMul(*/Transform3f(posMat1 * rotMat) * posMat2);
            thePos = theMat * thePos;
        }
    }
    thePos.y() = myTer.getGroundHeight(Vector2f(thePos.x(), thePos.z())) + PLAYERWARPHEIGHT + PLAYERWARPDIF;
    position(thePos);
    thePos.y() = thePos.y() - PLAYERWARPDIF;
    warp.setPos(thePos);
    warp.setClipSide(1);
    warp.setState(0);

    if (!isAI) {
        myCam.setTargetAngle(Vector3f(myCam.targetAngle().x(), atan2(-thePos.x(), thePos.z()), myCam.targetAngle().z()));
    }
    rotateY(atan2(-thePos.x(), thePos.z()));
    jumpSpeed = 0;
    isJumping = true;
    dodgeSpeed = Vector3f(0, 0, 0);
}

void HumanPlayer::die()
{
    tossFruit();
    dead = true;
}

void HumanPlayer::update()
{
    dt = (time() - lastTime) / 1000.0;
    lastTime = time();

    jumpSpeed -= grav * dt;
    stunTime -= dt;
    if (stunTime <= 0.0) stunTime = -0.01;

    anim[PLSTUNANIM].dFrame = anim[PLSTUNANIM].frameSpeed * dt;

    if (dead) {
        deadScale -= DEADSCALESPEED * dt;
        scale(Vector3f(PLAYERSCALE * deadScale, PLAYERSCALE * deadScale, PLAYERSCALE * deadScale));
        if (deadScale <= 0.0) {
            deadScale = 1.0;
            dead = false;
            scale(Vector3f(PLAYERSCALE * deadScale, PLAYERSCALE * deadScale, PLAYERSCALE * deadScale));
            respawn();
        }
    }

    if (stunTime <= 0.0 && !dead && !gameEnded) {
        if (time() - holdTime > 1000 && holds > -1 && holdMode == 2) {
            holdTime = time();
            myBaskets[baskId].getScore()++;
        }

        anim[PLWALKANIM].blend = qMin(1.0f, anim[PLWALKANIM].blend + PLBLENDSPEED * dt);
        anim[PLSTUNANIM].blend = 1.0 - anim[PLWALKANIM].blend;

        if (keyState[' '] && !isJumping) {
            jumpSpeed = jumpForce;
            isJumping = true;
        }

        if (!keyState[' '] && dodgeSpeed.x() == 0 && dodgeSpeed.z() == 0) {
            if (jumpSpeed > 0) jumpSpeed = 0;
        }

        if (keyState['W']) {
            move(moveSpeed * dt);
            anim[PLWALKANIM].dFrame = anim[PLWALKANIM].frameSpeed * dt;
        }

        if (keyState['S']) {
            move(-moveSpeed * dt);
            anim[PLWALKANIM].dFrame = anim[PLWALKANIM].frameSpeed * dt;
        }

        if (keyState['D']) {
            strafe(moveSpeed * dt);
            anim[PLWALKANIM].dFrame = anim[PLWALKANIM].frameSpeed * dt;
        }

        if (keyState['A']) {
            strafe(-moveSpeed * dt);
            anim[PLWALKANIM].dFrame = anim[PLWALKANIM].frameSpeed * dt;
        }

        if (keyState['E']) {
            tossFruit();
        }

        if (keyState['Q']) {
            banana.throwIt(getThrowVec());
        }

        if (dKeyState['D'] && !isJumping && time() - lastDodgeTime > PLAYERDODGETIMEDIF) {
            dKeyState['D'] = false;
            dLastTime['D'] = time() - DCDIF;
            jumpSpeed = PLAYERDODGEJUMPFAC * jumpForce;
            isJumping = true;
            dodgeRight();
        }

        if (dKeyState['A'] && !isJumping && time() - lastDodgeTime > PLAYERDODGETIMEDIF) {
            dKeyState['A'] = false;
            dLastTime['A'] = time() - DCDIF;
            jumpSpeed = PLAYERDODGEJUMPFAC * jumpForce;
            isJumping = true;
            dodgeLeft();
        }

        if (dKeyState['W'] && !isJumping && time() - lastDodgeTime > PLAYERDODGETIMEDIF) {
            dKeyState['W'] = false;
            dLastTime['W'] = time() - DCDIF;
            jumpSpeed = PLAYERDODGEJUMPFAC * jumpForce;
            isJumping = true;
            dodgeForward();
        }

        if (dKeyState['S'] && !isJumping && time() - lastDodgeTime > PLAYERDODGETIMEDIF) {
            dKeyState['S'] = false;
            dLastTime['S'] = time() - DCDIF;
            jumpSpeed = PLAYERDODGEJUMPFAC * jumpForce;
            isJumping = true;
            dodgeBackward();
        }
    }

    if (stunTime > 0.0 || dead || gameEnded || !(keyState['W'] || keyState['S'] || keyState['A'] || keyState['D'])) {
        anim[PLWALKANIM].dFrame = 0.0;
    }

    // Update stars
    if (stunTime > 0.0) {
        stars.update(dt);
    }

    if (stunTime > 0.0 || dead) {
        anim[PLWALKANIM].blend = qMax(0.0f, anim[PLWALKANIM].blend - PLBLENDSPEED * dt);
        anim[PLSTUNANIM].blend = 1.0 - anim[PLWALKANIM].blend;
    }

    banana.update(dt);
    if (warp.getState() < 3) warp.update(dt);

    position(getPos() + Vector3f(0.0, dt * jumpSpeed, 0.0) + dodgeSpeed * dt);

    updateAnim();
}

void HumanPlayer::nwUpdate()
{
    dt = (time() - lastTime) / 1000.0;
    lastTime = time();

    anim[PLSTUNANIM].dFrame = anim[PLSTUNANIM].frameSpeed * dt;

    jumpSpeed -= grav * dt;
    position(getPos() + Vector3f(0.0, dt * jumpSpeed, 0.0) + dodgeSpeed * dt);


    // Update stars
    if (stunTime > 0.0) {
        stars.update(dt);
    }

    if (stunTime > 0.0 || dead) {
        anim[PLWALKANIM].blend = qMax(0.0f, anim[PLWALKANIM].blend - PLBLENDSPEED * dt);
        anim[PLSTUNANIM].blend = 1.0 - anim[PLWALKANIM].blend;
    } else {
        anim[PLWALKANIM].blend = qMin(1.0f, anim[PLWALKANIM].blend + PLBLENDSPEED * dt);
        anim[PLSTUNANIM].blend = 1.0 - anim[PLWALKANIM].blend;

        if (!gameEnded && time() - holdTime > 1000 && holds > -1 && holdMode == 2) {
            holdTime = time();
            myBaskets[baskId].getScore()++;
        }
    }

    banana.position(banana.getPos() + banana.getThrowSpeed() * dt);

    Vector3f posA, posB;
    if (!banana.getInst()->getAttach().attached) {
        posA = banana.getInst()->transform() * Vector3f(WEAPONBANANATRAILOFFSET[0], WEAPONBANANATRAILOFFSET[1], WEAPONBANANATRAILOFFSET[2]);
        posB = banana.getInst()->transform() * Vector3f(WEAPONBANANATRAILOFFSET[3], WEAPONBANANATRAILOFFSET[4], WEAPONBANANATRAILOFFSET[5]);
    } else {
        posA = banana.getInst()->attachTransform() * Vector3f(WEAPONBANANATRAILOFFSET[0], WEAPONBANANATRAILOFFSET[1], WEAPONBANANATRAILOFFSET[2]);
        posB = banana.getInst()->attachTransform() * Vector3f(WEAPONBANANATRAILOFFSET[3], WEAPONBANANATRAILOFFSET[4], WEAPONBANANATRAILOFFSET[5]);
    }

    banana.getTrail().update(posA, posB);

    position(getPos() + Vector3f(0.0, dt * jumpSpeed, 0.0) + dodgeSpeed * dt);

    scale(Vector3f(PLAYERSCALE * deadScale, PLAYERSCALE * deadScale, PLAYERSCALE * deadScale));
    updateAnim();
}

void HumanPlayer::nwUpdate(long long rtime)
{
    (void) rtime;
    // XXX network
}

void HumanPlayer::updateAnim()
{
    for (unsigned int i = 0; i < anim.size(); i++) {
        anim[i].tFrame = anim[i].tFrame + anim[i].dFrame;
        anim[i].frame = anim[i].tFrame * (1.0 - anim[i].frameSmooth) + anim[i].frameSmooth * anim[i].frame;
        if (anim[i].frame >= anim[i].maxFrame) {
            anim[i].frame -= anim[i].maxFrame;
            anim[i].tFrame -= anim[i].maxFrame;
        }
        if (anim[i].frame <= 0.0) {
            anim[i].frame += anim[i].maxFrame;
            anim[i].tFrame += anim[i].maxFrame;
        }
        inst->setAnimSetBlend(anim[i].setId, anim[i].frame, anim[i].blend, anim[i].priority);
    }
    inst->calcAnim();
}

void HumanPlayer::render()
{
    if (warp.getState() < 3 && warp.getClipSide() != 0) {
        glEnable(GL_CLIP_PLANE0);
        GLdouble eq[4] = {0.0, -1.0 * warp.getClipSide(), 0.0, warp.getPos().y() * warp.getClipSide()};
        glClipPlane(GL_CLIP_PLANE0, eq);

        if (teamMode < 2) {
            glEnable(GL_COLOR_MATERIAL);
            Vector3f theColor = myPlayerColor[myTeams[teamId].getColorId()];
            theColor = lerp(theColor, Vector3f(1, 1, 1), 0.5);
            glColor3f(theColor.x(), theColor.y(), theColor.z());
        }
        inst->render();
        banana.render();
        if (teamMode < 2) {
            glColor3f(1.0, 1.0, 1.0);
            glDisable(GL_COLOR_MATERIAL);
        }
        glDisable(GL_CLIP_PLANE0);
    } else {
        if (teamMode < 2) {
            glEnable(GL_COLOR_MATERIAL);
            Vector3f theColor = myPlayerColor[myTeams[teamId].getColorId()];
            theColor = lerp(theColor, Vector3f(1, 1, 1), 0.5);
            glColor3f(theColor.x(), theColor.y(), theColor.z());
        }
        inst->render();
        banana.render();
        if (teamMode < 2) {
            glColor3f(1.0, 1.0, 1.0);
            glDisable(GL_COLOR_MATERIAL);
        }
    }
}

void HumanPlayer::warpRender()
{
    if (warp.getState() < 3 && warp.getPos().y() > SEAHEIGHT) warp.render();
}

void HumanPlayer::warpRender2()
{
    if (warp.getState() < 3 && warp.getPos().y() <= SEAHEIGHT) warp.render();
    banana.trailRender();
    if (stunTime > 0.0) {
        glDisable(GL_LIGHTING);
        glPushMatrix();
        Transform3f theMat = inst->transform();
        theMat = /*fastMul(*/theMat * inst->getFrame(9).transform;
        stars.setTrans(theMat);
        stars.setAlpha(qMin(1.0, 2.0 * stunTime / playerStunTime));
        stars.render();
        glPopMatrix();
        glEnable(GL_LIGHTING);
    }
}

void HumanPlayer::renderShadow()
{
    int tileX = (int)floor((getPos().x() - PLAYERSHADOWRAD - myTer.getPos().x()) / myTer.getXSegSize());
    int tileZ = (int)floor((myTer.getPos().z() - getPos().z() - PLAYERSHADOWRAD) / myTer.getZSegSize());

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBindTexture(GL_TEXTURE_2D, shadowTex);
    glEnable(GL_TEXTURE_2D);
    glPolygonOffset(-1.1, -4.0);
    glEnable(GL_POLYGON_OFFSET_FILL);

    float posX = tileX * myTer.getXSegSize() + myTer.getPos().x();
    float posZ = myTer.getPos().z() - tileZ * myTer.getZSegSize();
    float bu = 0.5 * (((posX - getPos().x()) / PLAYERSHADOWRAD) + 1.0);
    float bv = 0.5 * (((getPos().z() - posZ) / PLAYERSHADOWRAD) + 1.0);
    float stepu = 0.5 * myTer.getXSegSize() / PLAYERSHADOWRAD;
    float stepv = 0.5 * myTer.getZSegSize() / PLAYERSHADOWRAD;

    glBegin(GL_TRIANGLES);
        glNormal3f(-0.49, 0.81, -0.32);
        for (int i = 0; i < 2; i++) {
            for (int j = 0; j < 2; j++) {
                glTexCoord2f(bu + i * stepu, bv + j * stepv);
                glVertex3f(posX + i * myTer.getXSegSize(), myTer.getHeight(tileX + i, tileZ + j), posZ - j * myTer.getZSegSize());
                glTexCoord2f(bu + (i + 1) * stepu, bv + (j + 1) * stepv);
                glVertex3f(posX + (i + 1) * myTer.getXSegSize(), myTer.getHeight(tileX + i + 1, tileZ + j + 1), posZ - (j + 1) * myTer.getZSegSize());
                glTexCoord2f(bu + i * stepu, bv + (j + 1) * stepv);
                glVertex3f(posX + i * myTer.getXSegSize(), myTer.getHeight(tileX + i, tileZ + j + 1), posZ - (j + 1) * myTer.getZSegSize());

                glTexCoord2f(bu + i * stepu, bv + j * stepv);
                glVertex3f(posX + i * myTer.getXSegSize(), myTer.getHeight(tileX + i, tileZ + j), posZ - j * myTer.getZSegSize());
                glTexCoord2f(bu + (i + 1) * stepu, bv + j * stepv);
                glVertex3f(posX + (i + 1) * myTer.getXSegSize(), myTer.getHeight(tileX + i + 1, tileZ + j), posZ - j * myTer.getZSegSize());
                glTexCoord2f(bu + (i + 1) * stepu, bv + (j + 1) * stepv);
                glVertex3f(posX + (i + 1) * myTer.getXSegSize(), myTer.getHeight(tileX + i + 1, tileZ + j + 1), posZ - (j + 1) * myTer.getZSegSize());
            }
        }
    glEnd();

    glDisable(GL_POLYGON_OFFSET_FILL);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}

void HumanPlayer::sendUpdate()
{
    // XXX network stuff
}

void HumanPlayer::scale(const Vector3f& scale)
{
    inst->scaleObj(scale);
}

void HumanPlayer::rotate(const Vector3f& rot)
{
    inst->rotate(rot);
}

void HumanPlayer::rotateX(float ang)
{
    inst->rotateX(ang);
}

void HumanPlayer::rotateY(float ang)
{
    inst->rotateY(ang);
}

void HumanPlayer::rotateZ(float ang)
{
    inst->rotateZ(ang);
}

void HumanPlayer::position(const Vector3f& pos)
{
    inst->position(pos);
}

void HumanPlayer::positionBS(const Vector3f& pos)
{
    inst->position(pos - inst->transform().linear() * colBS.pos);
}

Vector3f HumanPlayer::getDirVec()
{
    return Vector3f(sin(inst->getRot().y() - PLAYERROTOFFSETY),
            0.0,
            -cos(inst->getRot().y() - PLAYERROTOFFSETY));
}

Vector3f HumanPlayer::getThrowVec()
{
    return Vector3f(sin(inst->getRot().y() - PLAYERROTOFFSETY) * cos(myCam.targetAngle().x() + PLAYERLOOKANGLE),
                sin(myCam.targetAngle().x() + PLAYERLOOKANGLE),
                -cos(inst->getRot().y() - PLAYERROTOFFSETY) * cos(myCam.targetAngle().x() + PLAYERLOOKANGLE));

}

Vector3f HumanPlayer::getScale()
{
    return inst->getScale();
}

Vector3f HumanPlayer::getRot()
{
    return inst->getRot();
}

Vector3f HumanPlayer::getPos()
{
    return inst->getPos();
}

BS HumanPlayer::getColBS()
{
    return BS(inst->transform() * colBS.pos, colBS.radius * getScale().x());
}

Vector3f HumanPlayer::getStrafePos()
{
    Vector3f dir = getDirVec();
    dir = Vector3f(-dir.z(), 0.0, dir.x());
    return getPos() + dir;
}

Team::Team(int tid)
    : nPlayers(0)
    , players(8 / nTeams)
    , baskId(-1)
    , teamId(tid)
    , teamName(QString("Team %1").arg(tid + 1))
    , colorId(0)
{
}

Team::~Team()
{
}

void Team::addPlayer(int pid)
{
    if (nPlayers < players.size()) {
        players[nPlayers++] = pid;
    }
}

int Team::getPlayId(int index)
{
    return players[index];
}

int Team::getScore()
{
    return myBaskets[baskId].getScore();
}

void Team::setScore(int s)
{
    myBaskets[baskId].getScore() = s;
}

bool Team::isEmpty()
{
    return nPlayers == 0;
}

int Team::getRelTeamId(int pid)
{
    for (unsigned int i = 0; i < nPlayers; i++) {
        if (players[i] == pid) return i;
    }
    return -1;
}


