/*
    gamefupluginmanager.cpp - Gamefu plugin manager

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#include "gamefupluginmanager.h"

// Qt includes
#include <qvaluestack.h>
#include <qtimer.h>

// KDE includes
#include <kapplication.h>
#include <ktrader.h>
#include <kstaticdeleter.h>
#include <kparts/componentfactory.h>
#include <kglobal.h>
#include <kconfig.h>
#include <kdebug.h>

// Gamefu includes
#include "gamefuplugin.h"

// TODO: Python plugin support
namespace Gamefu
{

static KStaticDeleter<PluginManager> pluginManagerDeleter;
PluginManager *PluginManager::s_self = 0;
PluginManager *PluginManager::self()
{
	if( !s_self )
	{
		pluginManagerDeleter.setObject(s_self, new PluginManager);
	}
	
	return s_self;
}

class PluginManager::Private
{
public:
	KPluginInfo::List foundPlugins;
	QValueStack<QString> pluginToLoad;
	typedef QMap<QString, Gamefu::Plugin*> PluginMap;
	PluginMap loadedPlugins;

	KPluginInfo *infoForPlugin(const QString &pluginName);
};

PluginManager::PluginManager(QObject *parent, const char *name)
 : QObject(parent, name), d(new Private)
{
	KConfig *gamefuConfig = KGlobal::config();
	// Look for all available plugins.
	d->foundPlugins = KPluginInfo::fromServices( KTrader::self()->query( QString::fromUtf8("Gamefu/Plugin") ), gamefuConfig, QString::fromUtf8("Plugins") );

	// We want to add a reference to the application's event loop so we
	// can remain in control when all windows are removed.
	// This way we can unload plugins asynchronously, which is more
	// robust if they are still doing processing.
	kapp->ref();
}

PluginManager::~PluginManager()
{
	// TODO: Check if doesn't crash.
	shutdown();
	delete d;
}

void PluginManager::load()
{
	// Check for enabled plugins
	KPluginInfo::List::ConstIterator it, itEnd = d->foundPlugins.constEnd();
	for(it = d->foundPlugins.constBegin(); it != itEnd; ++it)
	{
		KPluginInfo *info = static_cast<KPluginInfo*>(*it);
		
		// Load config to check if the plugin is enabled
		info->load();
		if( info->isPluginEnabled() )
		{
			d->pluginToLoad.push( info->pluginName() );
		}
		else
		{
			unloadPlugin( info->pluginName() );
		}
	}

	// Load enabled plugins
	QTimer::singleShot(0, this, SLOT(createNextPlugin()));
}

QValueList<KPluginInfo *> PluginManager::availablePlugins(const QString &category)
{
	KPluginInfo::List result;
	KPluginInfo::List::ConstIterator it, itEnd = d->foundPlugins.constEnd();
	for( it = d->foundPlugins.constBegin(); it != itEnd; ++it )
	{
		KPluginInfo *info = static_cast<KPluginInfo*>(*it);
		if( info->category() == category || category.isEmpty() )
		{
			result.append(info);
		}
	}

	return result;
}

void PluginManager::createNextPlugin()
{
	// Does all plugins are loaded ?
	if( d->pluginToLoad.isEmpty() )
	{
		kdDebug() << "All plugin finished to load." << endl;
		emit loadingFinished();
		return;
	}

	// Create the plugin
	QString nextPlugin = d->pluginToLoad.pop();
	createPluginInternal(nextPlugin);
	
	// Continue the loading plugin process.
	QTimer::singleShot(0, this, SLOT(createNextPlugin()));
}

void PluginManager::createPluginInternal(const QString &pluginName)
{
	// Get plugin info
	KPluginInfo *info = d->infoForPlugin(pluginName);
	if( !info )
	{
		kdWarning() << "Didn't found suitable info for plugin name '" << pluginName << "'" << endl;
		return;
	}

	// Do not load plugin twice.
	if( d->loadedPlugins.contains(pluginName) )
		return;

	// Query components factory to create an instance of the give plugin name.
	int error = 0;
	Gamefu::Plugin *plugin = KParts::ComponentFactory::createInstanceFromQuery<Gamefu::Plugin>( QString( "Gamefu/Plugin" ),
		QString( "[X-KDE-PluginInfo-Name]=='%1'" ).arg( pluginName ), this, 0, QStringList(), &error );

	if( plugin )
	{
		// Insert current plugin into loaded plugins map
		d->loadedPlugins.insert(pluginName, plugin);
		// Get plugin info to be acceded by plugins.
		plugin->setPluginInfo( info );

		kdDebug() << "Succesfully loaded Gamefu plugin '" << pluginName << "'" << endl;

		// Notify the exterior that this plugin is loaded.
		emit pluginLoaded(plugin);
	}
	else
	{
		switch( error )
		{
		case KParts::ComponentFactory::ErrNoServiceFound:
			kdWarning() << "No service implementing the given mimetype "
				<< "and fullfilling the given constraint expression can be found." << endl;
			break;

		case KParts::ComponentFactory::ErrServiceProvidesNoLibrary:
			kdWarning() << "the specified service provides no shared library." << endl;
			break;

		case KParts::ComponentFactory::ErrNoLibrary:
			kdWarning() << "the specified library could not be loaded." << endl;
			break;

		case KParts::ComponentFactory::ErrNoFactory:
			kdWarning() << "the library does not export a factory for creating components." << endl;
			break;

		case KParts::ComponentFactory::ErrNoComponent:
			kdWarning() << "the factory does not support creating components of the specified type." << endl;
			break;
		}

		kdWarning() << "Loading plugin '" << pluginName << "' failed, KLibLoader reported error: '" << endl <<
			KLibLoader::self()->lastErrorMessage() << "'" << endl;
	}
}

void PluginManager::unloadPlugin(const QString &pluginName)
{
	if( d->loadedPlugins.contains(pluginName) )
	{
		kdDebug() << "Unloading plugin " << pluginName << endl;
		Gamefu::Plugin *plugin = d->loadedPlugins[pluginName];

		emit pluginUnloaded(plugin);

		d->loadedPlugins.remove(pluginName);

		plugin->deleteLater();
	}
}

void PluginManager::shutdown()
{
	kdDebug() << endl;
	// Remove any pending plugins to load, we're shutting down now :)
	d->pluginToLoad.clear();
	
	// Ask all plugins to unload
	Private::PluginMap::Iterator it, itEnd = d->loadedPlugins.end();
	for ( it = d->loadedPlugins.begin(); it != itEnd; /* EMPTY */ )
	{
		// Plugins could emit their ready for unload signal directly in response to this,
		// which would invalidate the current iterator. Therefore, we copy the iterator
		// and increment it beforehand.
		Private::PluginMap::Iterator current( it );
		++it;
		current.data()->deleteLater();
	}
	
	QTimer::singleShot(3000, this, SLOT( shutdownDone() ) );
}

void PluginManager::shutdownDone()
{
	kapp->deref();
}

KPluginInfo *PluginManager::Private::infoForPlugin(const QString &pluginName)
{
	KPluginInfo::List::ConstIterator it, itEnd = foundPlugins.constEnd();
	for(it = foundPlugins.constBegin(); it != itEnd; ++it)
	{
		KPluginInfo *info = static_cast<KPluginInfo*>(*it);
		if( info->pluginName() == pluginName )
			return info;
	}

	return 0;
}

}

#include "gamefupluginmanager.moc"
