/*
    notifymanager.h - Gamefu Global Notify manager.

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#ifndef GAMEFUNOTIFYMANAGER_H
#define GAMEFUNOTIFYMANAGER_H

#include <qobject.h>

namespace KIO
{
	class Job;
}

namespace Gamefu 
{

class AbstractNotify;
/**
 * @brief Global notification framework
 * This is a singleton class that help libgamefu and plugins to display notifications in main interface.
 * You must set a AbstractNotify backend to actually make this work. See setNotifyBackend()
 *
 * Use this class with Gamefu::NotifyManager::self()
 *
 * @author Michaël Larouche <larouche@kde.org>
 */
class NotifyManager : public QObject
{
	Q_OBJECT
public:
	/**
	 * @brief Type of message
	 * Identify the kind of message displayed by notify backend. Same enum as KDE::StatusBar::MessageType
	 * Actually set a different icon in the resulting popup message.
	 *
	 * - Information: Informative message.
	 * - Question: Message is a question.
	 * - Sorry: Message that shows a limitation in the program.
	 * - Warning: You should pay attention to this message, but don't expect a normal behavior.
	 * - Error: Fatal error occured.
	 * - ShowAgainCheckBox: not used yet.
	 * - None: Show no icons.
	 */
	enum MessageType { Information, Question, Sorry, Warning, Error, ShowAgainCheckBox, None };

	/**
	 * d-tor
	 */
	~NotifyManager();

	/**
	 * @brief Singleton access to this class.
	 * @return the unique instance of this class
	 */
	static NotifyManager *self();

public slots:
	/**
	 * @brief Set notify backend.
	 * Must be set before using this class.
	 * @param backend Notify backend that inherit AbstractNotify.
	 */
	void setNotifyBackend(AbstractNotify *backend);

	/**
	 * @brief Display a long message in a popup message.
	 * Display a passive popup.
	 * @param message message to display.
	 * @param type message type. See @ref MessageType enum.
	 */
	void longMessage(const QString &message, int type = Information);
	/**
	 * @brief Display a short message with a longer message description.
	 * A short message is displayed and you can popup a longer message.
	 * @param shortMessage short message to display.
	 * @param longMessage longMessage to display.
	 * @param type message type for longMessage. See @ref MessageType enum.
	 */
	void shortLongMessage(const QString &shortMessage, const QString &longMessage, int type = Information);
	/**
	 * @brief Display a short message.
	 * A short message is displayed during a period of time.
	 * @param shortMessage shortMessage to display.
	 * @param longDelay if true, display the message during a greater period of time.
	 */
	void shortMessage(const QString &shortMessage, bool longDelay = false);

	/**
	 * @brief Create a new progress operation.
	 * Display a progress bar with an abort button.
	 * @param owner The owner is used to identity the progress. Owner is like the key to update the progress bar.
	 * @param description Description of the current operation.
	 * @param totalSteps Total steps required to complete the operation.
	 * @param receiver QObject receiver for the abort button.
	 * @param abortSlot Slot for the abort button.
	 */
	void createNewProgressOperation(QObject *owner, const QString &description, int totalSteps, QObject *receiver, const char *abortSlot);
	/**
	 * @brief Create a new progress operation for KIO::Job
	 * Same as above, but for KIO::Job.
	 * @param job given job
	 * @param description Description of the current operation
	 * @param totalSteps Total steps required to complete the operation.
	 */
	void createNewProgressOperation(KIO::Job *job, const QString &description, int totalSteps);
	/**
	 * @brief Increment progress for the given QObject operation.
	 * @param owner Given QObject.
	 */
	void incrementProgress(QObject *owner);
	/**
	 * @brief Set current progress for given QObject operation.
	 * @param owner Given QObject
	 * @param steps current progress.
	 */
	void setProgress(QObject *owner, int steps);
	/**
	 * @brief Set a human-readable status message for the given QObject operation.
	 * @param owner Given QObject.
	 * @param status status message.
	 */
	void setProgressStatus(QObject *owner, const QString &status);

private:
	/**
	 * Private constructor to present more than one instance of this class.
	 * Use Gamefu::NotifyManager::self() to access this class.
	 */
	NotifyManager();

	static NotifyManager *s_self;

	class Private;
	Private *d;
};

}

#endif
