/*
    imagefillerjob.cpp - Job that fill the database with images

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#include "imagefillerjob.h"

// Qt includes
#include <qfile.h>

// KDE includes
#include <kdebug.h>
#include <klocale.h>
#include <kurl.h>

// Gamefu includes
#include <collectionaccess.h>
#include <rommetainformation.h>

// Local includes
#include "imagefillerstrategy.h"

class ImageFillerJob::Private
{
public:
	Private()
	 : imageStrategy(0)
	{}

	~Private()
	{
		delete imageStrategy;
	}

	KURL imageDirectory;
	ImageFillerStrategy *imageStrategy;
	ImageEnum::ImageType imageType;
};

ImageFillerJob::ImageFillerJob(QObject *parent)
 : ThreadWeaver::DependentJob(parent, "ImageFillerJob"), d(new Private)
{
	setDescription( i18n("Filling database with images...") );
}

ImageFillerJob::~ImageFillerJob()
{
	delete d;
}

void ImageFillerJob::setImageDirectory(const KURL &imageDirectory)
{
	d->imageDirectory = imageDirectory;
}

void ImageFillerJob::setImageType(ImageEnum::ImageType type)
{
	d->imageType = type;
}

void ImageFillerJob::setImageFillerStrategy(ImageFillerStrategy *strategy)
{
	delete d->imageStrategy;
	d->imageStrategy = 0;

	d->imageStrategy = strategy;
}

bool ImageFillerJob::doJob()
{
	QValueList<Gamefu::RomMetaInformation> romList = Gamefu::CollectionAccess::self()->getAllRomsList();

	if( romList.isEmpty() )
	{
		setStatus( i18n("A transaction is in progress, image will not be filled in database.") );
		abort();
	}
	setProgressTotalSteps( romList.count() );

	QValueList<Gamefu::RomMetaInformation>::iterator it, itEnd = romList.end();
	for( it = romList.begin(); it != itEnd; ++it )
	{
		// Check if the job was aborted
		if( !isAborted() )
		{
			// Get the filename part of the ROM URL.
			Gamefu::RomMetaInformation tempRomInfo = *it;
			KURL romUrl(tempRomInfo.url());
			QString filename = romUrl.fileName();

			// Format image filename
			// TODO: Support other image extension than PNG
			QString imageFilename = d->imageStrategy->getImageFilename(filename, d->imageType);

			// Do nothing if the image strategy return an empty string
			if( !imageFilename.isEmpty() )
			{
				imageFilename = d->imageDirectory.path() + '/' + imageFilename;
	
				// Check if the image filename exists
				if( QFile::exists(imageFilename) )
				{
					switch(d->imageType)
					{
						// Titlescreen
						case ImageEnum::Titlescreen:
							tempRomInfo.setTitlescreenPath(imageFilename);
							break;
						// Screenshot
						case ImageEnum::Screenshot:
							tempRomInfo.setScreenshotPath(imageFilename);
							break;
						default:
							continue;
					}
					Gamefu::CollectionAccess::self()->modifyExistingRom( tempRomInfo );
				}
			}
			
			// Increment progress in status bar
			incrementProgress();
		}
		else
		{
			break;
		}
	}

	return !isAborted();
}

void ImageFillerJob::completeJob()
{
	// Do nothing.
}

#include "imagefillerjob.moc"
