/*
    imagefillerstrategy.h - Image filename format strategy

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#ifndef IMAGEFILLERSTRATEGY_H
#define IMAGEFILLERSTRATEGY_H

// Qt includes
#include <qstring.h>

#include "imagetypeenum.h"

/**
 * @brief Base class for image filename format strategy
 *
 * Strategy must implement getImageFilename() method.
 *
 * This is a pure virtual class.
 *
 * @author Michaël Larouche <larouche@kde.org>
 */
class ImageFillerStrategy
{
public:
	/**
	 * @brief Retrieve the image filename from the given filename and ImageType.
	 *
	 * @param fileName ROM filename.
	 * @param imageType Image type, either a titlescreen or a screenshot
	 */
	virtual QString getImageFilename(const QString &fileName, ImageEnum::ImageType imageType) = 0;
};

/**
 * @brief Image Filler strategy that use the same filename as the source ROM.
 * @author Michaël Larouche <larouche@kde.org>
 */
class FilenameImageFillerStrategy : public ImageFillerStrategy
{
public:
	/**
	 * @brief Retrieve the image filname from the given filename and ImageType.
	 *
	 * @param fileName ROM filename.
	 * @param imageType Image type, either a titlescreen or a screenshot
	 */
	virtual QString getImageFilename(const QString &fileName, ImageEnum::ImageType imageType);
};

/**
 * @brief Image Filler strategy that use the release number as the image filename.
 *
 * Titlescreen: 1a.png; Screenshot: 1b.png
 * @author Michaël Larouche <larouche@kde.org>
 */
class ReleaseImageFillerStrategy : public ImageFillerStrategy
{
public:
	/**
	 * @brief Retrieve the image filname from the given filename and ImageType.
	 *
	 * @param fileName ROM filename.
	 * @param imageType Image type, either a titlescreen or a screenshot
	 */
	virtual QString getImageFilename(const QString &fileName, ImageEnum::ImageType imageType);
};

#endif
