/*
    Copyright 2009  Stanislas Marquis <smarquis@chaosorigin.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "core.h"

void zodiacore_new( ZodiaCore** z )
{
    /*
     *  Allocate memory for ZodiaCore and set everything to null.
     */

    Trace( printf( "-- New ZodiaCore\n" ); )

#ifndef NDEBUG
    memusage_init();
#endif /* !NDEBUG */

    assert( *z == NULL );
    *z = Malloc( sizeof( ZodiaCore ) );

#ifndef SILENT
    (*z)->verboseFlag = True;
#endif /* !SILENT */

    (*z)->dataDir =
    (*z)->shareDir = NULL;

    (*z)->dataCnx =
    (*z)->atlasCnx = NULL;

    (*z)->prefs = NULL;

    (*z)->aspectCollection =
    (*z)->planetCollection =
    (*z)->aspectFilterCollection =
    (*z)->planetFilterCollection =
    (*z)->orbFilterCollection =
    (*z)->aspectRestrCollection =
    (*z)->orbRestrCollection =
    (*z)->profileCollection =
    (*z)->contextCollection =
    (*z)->horoscopeCollection = NULL;

    (*z)->sortedHoroscopes =
    (*z)->chartCollection = NULL;
}

void zodiacore_init( ZodiaCore* z, int options )
{
    if ( options & zcINIT_HomeDir )
        zodiacore_init_data_dir( z );

    if ( options & zcINIT_ShareDir )
        zodiacore_init_share_dir( z );

    if ( options & zcINIT_UseDb )
        sqlite_init( z, NULL );

    if ( options & zcINIT_UseAtlas )
        atlas_init( z, NULL );

    if ( options & ( zcINIT_UseDb|zcINIT_LoadDb ) )
    {
        aspect_init( z );
        planet_init( z );
        aspectfilter_init( z );
        planetfilter_init( z );
        orbfilter_init( z );
        aspectrestr_init( z );
        orbrestr_init( z );
        profile_init( z );
        context_init( z );
        prefs_init( z );
        horoscope_init( z );
        chart_init( z );
    }

#ifndef NDEBUG
    memusage_debug();
#endif /* !NDEBUG */
}

void zodiacore_delete( ZodiaCore** z, Bool freeSwe )
{
    assert( *z );

    Free( (*z)->dataDir, sizeof( char ) * ( strlen( (*z)->dataDir ) + 1 ) );
    Free( (*z)->shareDir, sizeof( char ) * ( strlen( (*z)->shareDir ) + 1 ) );

    sqlite_fini( *z );
    atlas_fini( *z );

    aspect_fini( *z );
    planet_fini( *z );
    aspectfilter_fini( *z );
    planetfilter_fini( *z );
    orbfilter_fini( *z );
    aspectrestr_fini( *z );
    orbrestr_fini( *z );
    profile_fini( *z );
    context_fini( *z );
    prefs_fini( *z );
    horoscope_fini( *z );
    chart_fini( *z );

    Trace( printf( "-- Delete ZodiaCore\n" ); )
    Free( *z, sizeof( ZodiaCore ) );
    *z = NULL;

    if ( freeSwe ) swe_close();

#ifndef NDEBUG
    memusage_debug();
#endif /* NDEBUG */
}

void zodiacore_init_data_dir( ZodiaCore* z )
{
    /*
     *  The struct ZodiaCore expects a malloc'd string as path to home dir.
     *  This function tries to find it alone:
     *    - by env var ZODIACORE_HOME_DIR or
     *    - check inside home directory for directory ./zodiacore
     *      and create it if necessary
     */

    if ( z->dataDir ) return;

    int i;
    char* home = getenv( "ZODIACORE_HOME_DIR" );

#ifdef KAMALA
    assert( home );
#else
    if ( !home )
    {
#ifdef __unix
        home = getenv( "HOME" );
#else
        home = getenv( "HOMEPATH" );
#endif /* __unix */
        assert( home );
        const size_t sz = strlen( home );

        z->dataDir = Malloc( sizeof( char ) * ( sz + 12 ) );

        strncpy( z->dataDir, home, sz );
        strcpy( &z->dataDir[sz], "/.zodiacore" );
    }
    else
    {
#endif /* KAMALA */
        z->dataDir = Malloc( sizeof( char ) * ( strlen( home ) + 1 ) );
        strcpy( z->dataDir, home );
#ifndef KAMALA
    }
#endif /* !KAMALA */

    Verbose( z, printf( "-- Checking home directory %s\n", z->dataDir ); )

    struct stat st;
    i = stat( z->dataDir, &st );
    if ( i )
    {
        Verbose( z, printf( "-- Creating home directory %s\n", z->dataDir ); )
#ifdef __unix
        Debug( i = ) mkdir( z->dataDir, S_IRWXU );
#else
        Debug( i = ) mkdir( z->dataDir );
#endif /* __unix */
        assert( !i );
    }
    else
    {
        assert( S_ISDIR( st.st_mode ) );
        assert( st.st_mode & S_IRWXU );
    }
}

void zodiacore_init_share_dir( ZodiaCore* z )
{
    /*
     *  The struct ZodiaCore expects a malloc'd string as path to shared dir.
     *  This function tries to find it alone:
     *    - by env var ZODIACORE_SHARE_DIR
     *    - check in a range of already existing system directories
     */

    if ( z->shareDir ) return;

    int i;
    struct stat st;
    char* path = getenv( "ZODIACORE_SHARE_DIR" );

    if ( !path )
    {
#ifdef __unix
        const char* const paths[] =
        {
            "/usr/share/zodiacore",
            "/usr/local/share/zodiacore",
            "/share/zodiacore",
            "/usr/opt/zodiacore",
            "/usr/local/opt/zodiacore",
            "/opt/zodiacore",
            NULL
        };
#else
        const char* const paths[] =
        {
            "C:/ZodiaCore",
            "C:/ChaosOrigin",
            "C:/ChaosOrigin/ZodiaCore",
            "C:/Program Files/ZodiaCore",
            "C:/Program Files/ChaosOrigin",
            "C:/Program Files/ChaosOrigin/ZodiaCore",
            NULL
        };
#endif /* __unix */
        for ( i = 0; paths[i]; ++i )
        {
            if ( !access( paths[i], R_OK ) )
            {
                path = (char*) paths[i];
                break;
            }
        }
    }

    if ( path )
    {
        Verbose( z, printf( "-- Checking shared directory %s\n", path ); )
        i = stat( path, &st );
    }
    else
    {
        path = (char*) "?";
        i = 1;
    }

    if ( i )
    {
        Fatal
        (
            printf( "-- Fatal %d unable to stat shared directory %s\n"
                    "-- Check your installation\n",
                i, path );
        )
    }
    assert( S_ISDIR( st.st_mode ) );
#ifdef __unix
    assert( st.st_mode & ( S_IROTH | S_IXOTH ) );
#endif /* __unix */

    z->shareDir = Malloc( sizeof( char ) * ( strlen( path ) + 1 ) );
    strcpy( z->shareDir, path );
}

#ifndef NDEBUG

void zodiacore_unittest( ZodiaCore* z )
{
    /*btree_unittest();*/
    json_unittest();
    /*sqlite_unittest();*/
    str_unittest();
    /*vector_unittest();*/
}

#endif /* !NDEBUG */
