/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "texturepool.h"

#include <QGLContext>

Kolf::TexturePool::TexturePool()
{
	//Load common textures into the pool here.
}

Kolf::TexturePool::~TexturePool()
{
	foreach (QGLContext* context, m_contextData.keys())
		detachContext(context);
}

Kolf::TexturePool* Kolf::TexturePool::instance()
{
	static Kolf::TexturePool theInstance;
	return &theInstance;
}

void Kolf::TexturePool::addTexture(const QByteArray& key, const QImage& image)
{
	//check input
	if (key.isEmpty() || image.isNull())
		return;
	//remove old texture if necessary, then store new texture
	if (m_textureData.contains(key))
		removeTexture(key);
	m_textureData[key] = image;
}

void Kolf::TexturePool::removeTexture(const QByteArray& key)
{
	if (!m_textureData.contains(key))
		return;
	//remove from texture registry
	m_textureData.remove(key);
	//delete textures on OpenGL contexts
	QMutableMapIterator<QGLContext*, Kolf::TexturePool::ContextData> iterContextData(m_contextData);
	while (iterContextData.hasNext())
	{
		QGLContext* context = iterContextData.next().key();
		Kolf::TexturePool::ContextData& contextData = iterContextData.value();
		if (!contextData.textures.contains(key))
			continue;
		context->deleteTexture(contextData.textures[key]);
		contextData.textures.remove(key);
	}
}

GLuint Kolf::TexturePool::prepareTexture(const QByteArray& key, QGLContext* context)
{
	if (!m_textureData.contains(key))
		return 0;
	//the following will create a new ContextData structure if this context binds a texture for the first time
	Kolf::TexturePool::ContextData& contextData = m_contextData[context];
	//generate texture if necessary
	if (!contextData.textures.contains(key))
		contextData.textures[key] = context->bindTexture(m_textureData[key]);
	return contextData.textures[key];
}

void Kolf::TexturePool::detachContext(QGLContext* context)
{
	if (!m_contextData.contains(context))
		return;
	//delete textures on this OpenGL context
	Kolf::TexturePool::ContextData& contextData = m_contextData[context];
	foreach (GLuint texture, contextData.textures)
		context->deleteTexture(texture);
	m_contextData.remove(context);
}

#include "texturepool.moc"
