/***************************************************************************
 *   Copyright 2009 Zeng Huan <zh.issac@gmail.com>
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "blender-manager.h"
#include "blender-blendthread.h"
#include "blender-delegate.h"
#include "blender-model.h"
#include "blender-texture.h"
#include "blender-texturepool.h"
#include "hmcreator-drawingboard.h"
#include "../elements/utils-heightmap.h"
#include "../utils/selectiondialog.h"

#include <QListView>
#include <QPointer>
#include <KLocalizedString>

namespace Kolf
{
	const QString TextureFileNamePattern("kolf-ng/courses/%1");
}

Kolf::BlenderManager::BlenderManager(QObject* parent)
	: QObject(parent)
	, m_model(new Kolf::TextureModel)
	, m_delegate(new Kolf::TextureListDelegate)
	, m_view(new QListView)
	, m_outputSize(512, 512)
	, m_texturesModified(false)
	, m_blendThread(new BlendThread(this))
{
	m_view->setModel(m_model);
	m_view->setItemDelegate(m_delegate);
	m_view->setSelectionMode(QAbstractItemView::SingleSelection);
	connect(m_model, SIGNAL(texturesModified()), this, SLOT(setTexturesModified()));
}

Kolf::BlenderManager::~BlenderManager()
{
}

void Kolf::BlenderManager::addTextures()
{
	//create item view
	QListView* view = new QListView;
	view->setSelectionMode(QAbstractItemView::SingleSelection);
	view->setModel(Kolf::BlenderTexturePool::instance());
	Kolf::TextureListDelegate delegate;
	view->setItemDelegate(&delegate);
	//create dialog
	QPointer<Utils::SelectionDialog> dialog(new Utils::SelectionDialog(view));
	dialog->setCaption(i18n("Add texture"));
	foreach (const QModelIndex& index, dialog->exec())
		m_model->addTexture(index);
	delete dialog;
}

void Kolf::BlenderManager::delTextures()
{
	QModelIndexList indexes = m_view->selectionModel()->selectedIndexes();
	if (!indexes.isEmpty())
		m_model->removeTexture(indexes[0].row());
}

void Kolf::BlenderManager::startBlending(bool force)
{
	if (force || m_texturesModified)
	{
		m_blendThread->process();
		m_texturesModified = false;
	}
}

void Kolf::BlenderManager::processBlending()
{
	QMutexLocker locker(&m_mutex);
	m_blendResult = QImage(m_outputSize, QImage::Format_RGB32);
	uint* resultDataPtr = reinterpret_cast<uint*>(m_blendResult.bits());

	for (int x = 0; x < m_outputSize.width(); ++x)
		for(int y = 0; y < m_outputSize.height(); ++y)
		{
			qreal r = 0, g = 0, b = 0, a = 0;
			int height = m_heightmap.pixel(x, y);

			foreach (const BlenderTexture* texture, m_model->textures())
			{
				QColor texColor = texture->pixel(x, y, height, Kolf::PolylineCurve);
				//http://en.wikipedia.org/wiki/Alpha_compositing
				qreal alphaT = texColor.alphaF();
				r = texColor.redF() * alphaT + r * a * ( 1.0 - alphaT);
				g = texColor.greenF() * alphaT + g * a * ( 1.0 - alphaT);
				b = texColor.blueF() * alphaT + b * a * ( 1.0 - alphaT);
				a = alphaT + a - alphaT * a;
			}
			resultDataPtr[x + y * m_outputSize.width()] = QColor::fromRgbF(r, g, b, 1.0).rgba(); //NOTE: A = 255 is dictated by the image format.
		}
}

void Kolf::BlenderManager::setHeightmap(const Kolf::Heightmap& heightmap)
{
	m_mutex.lock();
	m_heightmap = heightmap.scaled(m_outputSize);
	m_mutex.unlock();
	startBlending(true);
}

#include "blender-manager.moc"
