/***************************************************************************
 *   Copyright 2009 Zeng Huan <zh.issac@gmail.com>
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "hmcreator-drawingboard.h"
#include "hmcreator-toolbutton.h"
#include "../elements/geometry-terrain.h"
#include "../elements/object-environment.h"
#include "../elements/scene.h"

#include <QMouseEvent>
#include <QPainter>
#include <QTransform>
#include <KLocalizedString>

const QPen Kolf::DrawingBoard::HelperLinesPen(Qt::red);

Kolf::DrawingBoard::DrawingBoard()
	: GameUIPageWidget(i18n("Terrain heightmap"))
	, m_currentToolButton(0)
	, m_isEditedFromLastBlending(false)
{
	initializeImage();
}

void Kolf::DrawingBoard::setImage(const QPixmap& image)
{
	m_workingImage = image;
	emit heightmapChanged(image.toImage());
	storeImage();
}

void Kolf::DrawingBoard::setSelection(const QPainterPath& selection)
{
	 m_selection = selection;
	 update();
}

void Kolf::DrawingBoard::currentSceneChanged(Kolf::Scene* oldScene, Kolf::Scene* newScene)
{
	Q_UNUSED(oldScene);
	if(!newScene)
		return;
	foreach (Kolf::EnvironmentObject* object, newScene->objects().allOfType<Kolf::EnvironmentObject>())
	{
		const Kolf::TerrainGeometry* geometry = object->terrainGeometry();
		const Kolf::HeightmapRenderer& heightmapRenderer = geometry->heightmapRenderer();
		m_workingImage = fromHeightmapRenderer(heightmapRenderer);
		storeImage();
		emit heightmapChanged(heightmapRenderer.mirrored());
		update();
	}
}

void Kolf::DrawingBoard::hideEvent(QHideEvent* event)
{
	Q_UNUSED(event)
	if (m_isEditedFromLastBlending && !m_workingImage.isNull())
	{
		emit heightmapChanged(m_workingImage.toImage());
		m_isEditedFromLastBlending = false;
	}
	Kolf::GameUIPageWidget::hideEvent(event);
}

void Kolf::DrawingBoard::mousePressEvent(QMouseEvent* event)
{

	if (event->button() == Qt::LeftButton && m_currentToolButton) 
	{
		m_lastPoint = event->posF();
		m_currentToolButton->work(this, event);
	}
	update();
}

void Kolf::DrawingBoard::mouseMoveEvent(QMouseEvent* event)
{
	if(m_currentToolButton)
		m_currentToolButton->work(this, event);
	update();
}

void Kolf::DrawingBoard::mouseReleaseEvent(QMouseEvent* event)
{
	if (event->button() == Qt::LeftButton && m_currentToolButton) 
		m_currentToolButton->work(this, event);
	update();

}

void Kolf::DrawingBoard::paintEvent(QPaintEvent* event)
{
	Q_UNUSED(event);
	QPainter painter(this);
	//heightmap
	painter.drawPixmap(QPointF(0, 0), m_workingImage);
	//selection
	painter.setPen(Kolf::DrawingBoard::HelperLinesPen);
	painter.drawPath(m_selection);
}

void Kolf::DrawingBoard::resizeEvent(QResizeEvent* event)
{
	resizeImage(event);
	resizeSelection(event);
	update();
	QWidget::resizeEvent(event);
}

void Kolf::DrawingBoard::currentToolButtonChanged(ToolButton* button)
{
	m_currentToolButton = button;
	if(m_currentToolButton)
		m_currentToolButton->onSelected(this);
	update();
}

void Kolf::DrawingBoard::resizeSelection(QResizeEvent* event)
{
	if (event->oldSize().width() == 0 || event->oldSize().height() == 0)
		return;
	QTransform transform;
	qreal stretch_x = event->size().width() / qreal(event->oldSize().width());
	qreal stretch_y = event->size().height() / qreal(event->oldSize().height());
	m_selection = transform.scale(stretch_x, stretch_y).map(m_selection);
}

void Kolf::DrawingBoard::resizeImage(QResizeEvent* event)
{
	Q_UNUSED(event);
	//min size of painting area set to 128 * 128
	int newWidth = qMax(width(), 128);
	int newHeight = qMax(height(), 128);
	if(!m_workingImage.isNull())
		m_workingImage= m_workingImage.scaled(QSize(newWidth, newHeight));
	if(!m_storeImage.isNull())
		m_storeImage = m_storeImage.scaled(QSize(newWidth, newHeight));
}

void Kolf::DrawingBoard::storeImage()
{
	m_storeImage = m_workingImage;
	m_isEditedFromLastBlending = true;
}

void Kolf::DrawingBoard::restoreImage()
{
	m_workingImage = m_storeImage;
}

Kolf::HeightmapRenderer Kolf::DrawingBoard::toHeightmapRenderer()
{
	return Kolf::HeightmapRenderer(m_workingImage.toImage());
}

QPixmap Kolf::DrawingBoard::fromHeightmapRenderer(const Kolf::HeightmapRenderer& heightmapRenderer)
{
	//call method mirrored() to make image fit our coordinate system.
	return QPixmap::fromImage(heightmapRenderer.mirrored());
}

void Kolf::DrawingBoard::clearImage()
{
	m_workingImage.fill(qRgb(0, 0, 0));
}

void Kolf::DrawingBoard::initializeImage()
{
	m_workingImage = QPixmap(size());
	m_storeImage = QPixmap(size());
	clearImage();
	storeImage();
	update();
}

#include "hmcreator-drawingboard.moc"
