/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_MANAGEDOBJECT_H
#define KOLF_MANAGEDOBJECT_H

#include <QDebug>
#include <QMap>
#include <QPointer>
class KConfigGroup;

namespace Kolf
{
	class ManagedObject;
	class Object;
	class Scene;

	struct ManagedObjectStateRef
	{
		Kolf::ManagedObject* object; //0 marks invalid state references
		int stateIndex; //-1 denotes invalid state references

		inline ManagedObjectStateRef()
			: object(0)
			, stateIndex(-1)
		{
		}
		inline ManagedObjectStateRef(Kolf::ManagedObject* _object, int _stateIndex)
			: object(_object)
			, stateIndex(_stateIndex)
		{
		}
	};

	/**
	 * \class ManagedObject
	 * \since 2.0
	 *
	 * This is a layer over Kolf::Object, which provides state management. You might want to use the interface of Kolf::ManagedObjectGroup instead, if you're managing multiple objects, or want a complete QUndoStack.
	 */
	class ManagedObject : public QObject
	{
		Q_OBJECT
		Q_DISABLE_COPY(ManagedObject)
		private:
			//forward declaration of inner class
			class State;
		public:
			ManagedObject(Kolf::Object* object, Kolf::Scene* scene);
			virtual ~ManagedObject();

			///Returns the Kolf::Object instance which is managed by this Kolf::ManagedObject instance, or 0 if the object is not existing in the current state.
			Kolf::Object* object() const;
			///Returns the scene for this object.
			Kolf::Scene* scene() const;

			///Returns how much states are currently saved for this object.
			///\sa state()
			int stateCount() const;
			///Returns the index of the state that this managed object is currently occupying. (This state might not necessarily represent the true state of the object, if the object's properties were changed after the last pushState() or activateState() call.)
			int currentState() const;
			///Returns the \a index-th state of this managed object. Note that index == 0 always gives an empty pre-creation state, and index == stateCount() might give a post-deletion state if this object has been deleted (or was deleted, but this was undone).
			Kolf::ManagedObject::State* state(int index) const;

			///Writes this state snapshot into the given KConfigGroup.
			void serialize(KConfigGroup* group);
			///Read a state snapshot from the given KConfigGroup. A null pointer is returned if the format of this group is empty.
			static Kolf::ManagedObject* deserialize(KConfigGroup* group, Kolf::Scene* scene);
		public Q_SLOTS:
			///Saves the object's current state in the managed object's history. If the current state of the object is not the last one in this history, all following states will be dropped by inserting the new state (as one expects from an undo stack).
			///\returns the index of the newly created state (which can be used to go to this state later on), or -1 if the state was not different from the current one and was therefore merged into the current one (set the \a allowStateMerging argument to false to suppress this behavior)
			///\warning Do not use this method directly when this Kolf::ManagedObject is in a Kolf::ManagedObjectGroup.
			Kolf::ManagedObjectStateRef pushState(bool allowStateMerging = true);
			///Goes back to the \a index-th state. (This might cause the object to be deleted or created if one of the states is empty.)
			///\warning Do not use this method directly when this Kolf::ManagedObject is managed by a Kolf::ManagedObjectGroup.
			void activateState(int index);
		Q_SIGNALS:
			///This signal is emitted after the Kolf::Object instance of this managed object has been reincarnated through an undo or redo operation.
			void objectCreated(Kolf::ManagedObject* object);
			///This signal is emitted after the Kolf::Object instance of this managed object is about to be deleted through an undo or redo operation.
			void objectAboutToBeDeleted(Kolf::ManagedObject* object);
		private:
			class State;
			///Used by Kolf::ManagedObject::deserialize.
			ManagedObject(Kolf::ManagedObject::State* state, Kolf::Scene* scene);

			mutable QPointer<Kolf::Object> m_object; //may need to be recreated in the const method object()
			Kolf::Scene* m_scene;
			QList<Kolf::ManagedObject::State*> m_states;
			int m_currentStateIndex;
			bool m_blockPushRequests;
	};
}

/**
 * \class ManagedObject::State
 * \since 2.0
 *
 * This object saves the complete state of a Kolf::Object, that is: the class name and the values of all QMetaProperties. It is internally used by Kolf::ManagedObject and Kolf::ManagedObjectGroup.
 *
 * \sa Kolf::ManagedObject
 */
class Kolf::ManagedObject::State
{
	Q_DISABLE_COPY(State)
	public:
		///Constructs an empty state snapshot.
		///\sa isEmpty()
		State();
		///Creates a new state snapshot of the given \a object, or an empty state snapshot if a null pointer is given.
		static Kolf::ManagedObject::State* fromObject(Kolf::Object* object);

		///Returns the class name for this state.
		QByteArray className() const;
		///Returns whether this state snapshot is empty. An empty state snapshot is regularly used to represent objects that were deleted or have not yet been created.
		bool isEmpty() const;
		///Finds whether the difference between the given three states of one object (specified in chronological order) is enough to justify that all three states are saved in a managed object.
		///The relevant condition is that some properties change between \a previousState and \a mediacyState which do not change between \a mediacyState and \a followingState. Of course, something _has_ to change between \a mediacyState and \a followingState, too.
		static bool isMediacySignificant(Kolf::ManagedObject::State* previousState, Kolf::ManagedObject::State* mediacyState, Kolf::ManagedObject::State* followingState);

		///Applies this state to the given \a object. If the object's class is the same as the object which this state snapshot was taken from, the object will afterwards be in an exactly identical state.
		void applyToObject(Kolf::Object* object) const;

		///Writes this state snapshot into the given KConfigGroup.
		void serialize(KConfigGroup* group);
		///Read a state snapshot from the given KConfigGroup into a new Kolf::ManagedObject::State snapshot. A null pointer is returned if the format of this group is empty.
		static Kolf::ManagedObject::State* deserialize(KConfigGroup* group);
	private:
		///Creates a new state snapshot of the given \a object.
		///\warning This method assumes the validity of the given pointer.
		State(Kolf::Object* object);
		///Helper function for Kolf::ManagedObject::State::isMediacySignificant.
		static QList<QByteArray> findChangingProperties(Kolf::ManagedObject::State* previousState, Kolf::ManagedObject::State* nextState);

		QByteArray m_className;
		mutable QMap<QByteArray, QVariant> m_propertyValues;
};

#include "object.h"
inline QDebug operator<<(QDebug debug, const Kolf::ManagedObjectStateRef& ref)
{
	debug.nospace() << "(" << ref.stateIndex << " of " << (QObject*)ref.object->object() << ")";
	return debug.space();
}

#endif // KOLF_MANAGEDOBJECT_H
