/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "managedobjectgroup.h"
#include "managedobject.h"
#include "object.h"
#include "scene.h"

#include <KConfigGroup>

Kolf::ManagedObjectGroup::ManagedObjectGroup(Kolf::Scene* scene)
	: m_scene(scene)
	, m_currentStateIndex(-1)
	, m_cleanStateIndex(-1)
	, m_baseStateIndex(-1)
{
}

Kolf::ManagedObjectGroup::~ManagedObjectGroup()
{
	qDeleteAll(m_managedObjects);
}

int Kolf::ManagedObjectGroup::managedObjectCount() const
{
	return m_managedObjects.count();
}

Kolf::ManagedObject* Kolf::ManagedObjectGroup::managedObject(int index) const
{
	return m_managedObjects.value(index);
}

void Kolf::ManagedObjectGroup::addObject(Kolf::ManagedObject* managedObject)
{
	if (m_managedObjects.contains(managedObject))
		return;
	//track state changes
	const bool prevClean = isClean(), prevCanRedo = canRedo(), prevCanUndo = canUndo();
	//add object
	m_managedObjects << managedObject;
	connect(managedObject, SIGNAL(objectCreated(Kolf::ManagedObject*)),
			this, SLOT(slotObjectCreated(Kolf::ManagedObject*)));
	connect(managedObject, SIGNAL(objectAboutToBeDeleted(Kolf::ManagedObject*)),
			this, SIGNAL(objectAboutToBeDeleted(Kolf::ManagedObject*)));
	//push all states of this object to the global queue (create a first state if necessary)
	if (managedObject->stateCount() == 0)
		managedObject->pushState();
	trimStateList();
	for (int stateIndex = 0; stateIndex < managedObject->stateCount(); ++stateIndex)
		m_stateOrder << Kolf::ManagedObjectStateRef(managedObject, stateIndex);
	m_currentStateIndex = m_stateOrder.count() - 1;
	//check state changes
	const bool newClean = isClean(), newCanRedo = canRedo(), newCanUndo = canUndo();
	if (prevClean != newClean)
		emit cleanChanged(newClean);
	if (prevCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
	if (prevCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
	//done
	emit objectCreated(managedObject);
	return;
}

Kolf::ManagedObject* Kolf::ManagedObjectGroup::addObject(Kolf::Object* object)
{
	//check if object is already managed
	foreach (Kolf::ManagedObject* managedObject, m_managedObjects)
		if (managedObject->object() == object)
			return managedObject;
	//add object
	Kolf::ManagedObject* managedObject = new Kolf::ManagedObject(object, m_scene);
	addObject(managedObject);
	return managedObject;
}

void Kolf::ManagedObjectGroup::removeObject(Kolf::ManagedObject* managedObject)
{
	if (!m_managedObjects.contains(managedObject))
		return;
	emit objectAboutToBeDeleted(managedObject);
	//track state changes
	const bool prevClean = isClean(), prevCanRedo = canRedo(), prevCanUndo = canUndo();
	//remove all states of this managed object from the state order
	//we go through the states in reverse order because the indexes change during the loop
	for (int i = m_stateOrder.count() - 1; i >= 0; --i)
	{
		Kolf::ManagedObjectStateRef stateRef = m_stateOrder[i];
		if (stateRef.object == managedObject)
		{
			m_stateOrder.removeAt(i);
			if (m_currentStateIndex >= i)
				--m_currentStateIndex; //the indexes above i have been shifted!
			if (m_cleanStateIndex == i)
				m_cleanStateIndex = -1; //there is no clean state any more!
			if (m_baseStateIndex == i)
				m_baseStateIndex = -1; //there is no base state any more!
		}
	}
	//check state changes
	const bool newClean = isClean(), newCanRedo = canRedo(), newCanUndo = canUndo();
	if (prevClean != newClean)
		emit cleanChanged(newClean);
	if (prevCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
	if (prevCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
	//delete object
	m_managedObjects.removeAll(managedObject);
	delete managedObject;
}

void Kolf::ManagedObjectGroup::removeObject(Kolf::Object* object)
{
	foreach (Kolf::ManagedObject* managedObject, m_managedObjects)
		if (managedObject->object() == object)
		{
			removeObject(managedObject);
			return;
		}
}

void Kolf::ManagedObjectGroup::clear()
{
	qDeleteAll(m_managedObjects);
	m_managedObjects.clear();
	m_stateOrder.clear();
	m_currentStateIndex = m_cleanStateIndex = m_baseStateIndex = -1;
}

void Kolf::ManagedObjectGroup::serialize(KConfigGroup* group)
{
	static const QString objectGroupCaption = QLatin1String("Object %1");
	for (int i = 0; i < m_managedObjects.count(); ++i)
	{
		KConfigGroup objectGroup(group, objectGroupCaption.arg(i + 1));
		m_managedObjects[i]->serialize(&objectGroup);
	}
}

void Kolf::ManagedObjectGroup::deserialize(KConfigGroup* group)
{
	static const QString objectGroupCaption = QLatin1String("Object %1");
	const QStringList groupList = group->groupList();
	for (int i = 1; groupList.contains(objectGroupCaption.arg(i)); ++i)
	{
		KConfigGroup objectGroup(group, objectGroupCaption.arg(i));
		Kolf::ManagedObject* managedObject = Kolf::ManagedObject::deserialize(&objectGroup, m_scene);
		addObject(managedObject);
		//The following lines are quite a hack. They should be obsoleted by the complete editing history being saved in the course file.
		this->markAsBaseState();
		this->setClean();
	}
}

void Kolf::ManagedObjectGroup::pushState(Kolf::ManagedObject* managedObject)
{
	if (!m_managedObjects.contains(managedObject))
		return;
	//track state changes
	const bool prevClean = isClean(), prevCanRedo = canRedo(), prevCanUndo = canUndo();
	//get information about the current state
	Kolf::ManagedObjectStateRef currentState = (m_currentStateIndex == -1) ? Kolf::ManagedObjectStateRef() : m_stateOrder[m_currentStateIndex];
	//allow state merging only if no other objects were modified in the meantime
	const bool allowStateMerging = currentState.object == managedObject;
	//push state
	trimStateList();
	Kolf::ManagedObjectStateRef newStateRef = managedObject->pushState(allowStateMerging);
	if (newStateRef.stateIndex != -1)
	{
		m_stateOrder << newStateRef;
		++m_currentStateIndex;
	}
	//check state changes
	const bool newClean = isClean(), newCanRedo = canRedo(), newCanUndo = canUndo();
	if (prevClean != newClean)
		emit cleanChanged(newClean);
	if (prevCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
	if (prevCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
}

void Kolf::ManagedObjectGroup::pushState(Kolf::Object* object)
{
	foreach (Kolf::ManagedObject* managedObject, m_managedObjects)
		if (managedObject->object() == object)
		{
			pushState(managedObject);
			return;
		}
}

void Kolf::ManagedObjectGroup::markAsBaseState()
{
	//track state changes
	const bool prevClean = isClean(), prevCanRedo = canRedo(), prevCanUndo = canUndo();
	//change base state
	m_baseStateIndex = m_currentStateIndex;
	//check state changes
	const bool newClean = isClean(), newCanRedo = canRedo(), newCanUndo = canUndo();
	if (prevClean != newClean)
		emit cleanChanged(newClean);
	if (prevCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
	if (prevCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
}

void Kolf::ManagedObjectGroup::trimStateList()
{
	//remove everything after m_currentState
	for (int i = m_currentStateIndex + 1; i < m_stateOrder.count(); ++i)
		m_stateOrder.removeAt(i);
	//reset m_cleanStateIndex if the clean state was deleted
	if (m_cleanStateIndex > m_currentStateIndex)
		m_cleanStateIndex = -1;
	//reset m_baseStateIndex if the base state was deleted
	if (m_baseStateIndex > m_currentStateIndex)
		m_baseStateIndex = -1;
}

bool Kolf::ManagedObjectGroup::canRedo() const
{
	return m_currentStateIndex != -1 && m_currentStateIndex + 1 < m_stateOrder.count();
}

void Kolf::ManagedObjectGroup::redo()
{
	if (!canRedo())
		return;
	//track state changes
	const bool prevClean = isClean(), prevCanRedo = true, prevCanUndo = canUndo();
	//redo following state
	++m_currentStateIndex;
	Kolf::ManagedObjectStateRef stateRef = m_stateOrder[m_currentStateIndex];
	stateRef.object->activateState(stateRef.stateIndex);
	//check state changes
	const bool newClean = isClean(), newCanRedo = canRedo(), newCanUndo = canUndo();
	if (prevClean != newClean)
		emit cleanChanged(newClean);
	if (prevCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
	if (prevCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
}

bool Kolf::ManagedObjectGroup::canUndo() const
{
	if (m_baseStateIndex == -1)
		return m_currentStateIndex > 0;
	else
		return m_currentStateIndex > m_baseStateIndex;
}

void Kolf::ManagedObjectGroup::undo()
{
	if (!canUndo())
		return;
	//track state changes
	const bool prevClean = isClean(), prevCanRedo = canRedo(), prevCanUndo = true;
	//undo current state
	Kolf::ManagedObjectStateRef stateRef = m_stateOrder[m_currentStateIndex];
	stateRef.object->activateState(stateRef.stateIndex - 1);
	--m_currentStateIndex;
	//check state changes
	const bool newClean = isClean(), newCanRedo = canRedo(), newCanUndo = canUndo();
	if (prevClean != newClean)
		emit cleanChanged(newClean);
	if (prevCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
	if (prevCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
}

bool Kolf::ManagedObjectGroup::isClean() const
{
	return m_cleanStateIndex != -1 && m_cleanStateIndex == m_currentStateIndex;
}

void Kolf::ManagedObjectGroup::setClean()
{
	m_cleanStateIndex = m_currentStateIndex;
}

void Kolf::ManagedObjectGroup::slotObjectCreated(Kolf::ManagedObject* managedObject)
{
	Kolf::Object* object = managedObject->object();
	connect(object, SIGNAL(updated()), this, SLOT(objectUpdated()));
	emit objectCreated(managedObject);
}

void Kolf::ManagedObjectGroup::objectUpdated()
{
	Kolf::Object* caller = qobject_cast<Kolf::Object*>(sender());
	if (caller)
		pushState(caller);
}

#include "managedobjectgroup.moc"
