/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "object-circleblock.h"
#include "geometry-cylinder.h"
#include "objectregistry.h"
#include "overlay-circle.h"

#include <QBrush>
#include <QGraphicsEllipseItem>
#include <QPen>

REGISTER_KOLF_OBJECT(Kolf::CircleBlockObject, ki18n("Circular block"), KIcon())

Kolf::CircleBlockObject::CircleBlockObject(dWorldID worldID, dSpaceID spaceID)
	: Kolf::Object(worldID, spaceID)
	, m_geometry(new Kolf::CylinderGeometry(spaceID, this))
	, m_item(new QGraphicsEllipseItem(QRectF(), this))
	, m_radius(0)
{
	m_item->setPen(QPen(color()));
	m_item->setBrush(color());
}

Kolf::CircleBlockObject::~CircleBlockObject()
{
	delete m_item;
}

qreal Kolf::CircleBlockObject::height() const
{
	return m_geometry->height();
}

void Kolf::CircleBlockObject::setHeight(qreal height)
{
	m_geometry->setHeight(height);
}

qreal Kolf::CircleBlockObject::radius() const
{
	return m_radius;
}

void Kolf::CircleBlockObject::setRadius(qreal radius)
{
	if (m_radius != radius)
	{
		m_radius = radius;
		const QRectF newRect(-radius, -radius, 2 * radius, 2 * radius);
		m_item->setRect(newRect);
		m_item->setPen(color());
		m_item->setBrush(color());
		m_geometry->setRadius(radius);
		propagateUpdate();
	}
}

void Kolf::CircleBlockObject::objectEvent(int type, const QVariant& data)
{
	if (type == Kolf::Object::ColorChangeEvent)
	{
		const QColor color = data.value<QColor>();
		m_item->setPen(QPen(color));
		m_item->setBrush(color);
	}
}

Kolf::Overlay* Kolf::CircleBlockObject::createOverlay()
{
	return new Kolf::CircleOverlay(this);
}

void Kolf::CircleBlockObject::createSpatialGeometry()
{
	float x = position().x();
	float y = position().y();
	float z = position().z();
	
	//Has to be a a number that is a multiple of 2 and is divideable by 360 (so 2 works, but 15 doesn't).
	int numSides = 18;

	m_vertexArray = new float[numSides * 12 + (2 * (numSides / 2) * 12)];
	m_normalArray = new float[numSides * 12 + (2 * (numSides / 2) * 12)];
	numVerts = numSides * 12 + (2 * (numSides / 2) * 12);

	int facecounter = 0;
	//Hackish way to get pi.
	float pi = acos(-1.0f);

	//For-loop for the sides of the cylinder.
	for(int i = 0; i < 360; i += (360/numSides))
	{
		float sinAngle = (float)sin(i*pi / 180);
		float nextSinAngle = (float)(sin((i + (360/numSides))*pi / 180));
		float cosAngle = (float)cos(i*pi / 180);
		float nextCosAngle = (float)(cos((i + (360/numSides))*pi / 180));

		m_vertexArray[facecounter*12] = (float)(x + ((float)sinAngle * radius())); 
		m_vertexArray[facecounter*12 + 1] = (float)(y + ((float)cosAngle * radius()));
		m_vertexArray[facecounter*12 + 2] = (float)(z + height());
		m_vertexArray[facecounter*12 + 3] = (float)(x + ((float)sinAngle * radius())); 
		m_vertexArray[facecounter*12 + 4] = (float)(y + ((float)cosAngle * radius()));
		m_vertexArray[facecounter*12 + 5] = (float)(z);
		m_vertexArray[facecounter*12 + 6] = (float)(x + ((float)nextSinAngle * radius()));
		m_vertexArray[facecounter*12 + 7] = (float)(y + ((float)nextCosAngle * radius()));
		m_vertexArray[facecounter*12 + 8] = (float)(z);
		m_vertexArray[facecounter*12 + 9] = (float)(x + ((float)nextSinAngle * radius()));
		m_vertexArray[facecounter*12 + 10] = (float)(y + ((float)nextCosAngle * radius()));
		m_vertexArray[facecounter*12 + 11] = (float)(z + height());

		m_normalArray[facecounter*12] = (float)(sinAngle*-1);
		m_normalArray[facecounter*12 + 1] = (float)(cosAngle);
		m_normalArray[facecounter*12 + 2] = (float)(0);
		m_normalArray[facecounter*12 + 3] = (float)(sinAngle*-1);
		m_normalArray[facecounter*12 + 4] = (float)(cosAngle);
		m_normalArray[facecounter*12 + 5] = (float)(0);
		m_normalArray[facecounter*12 + 6] = (float)(nextSinAngle*-1);
		m_normalArray[facecounter*12 + 7] = (float)(nextCosAngle);
		m_normalArray[facecounter*12 + 8] = (float)(0);
		m_normalArray[facecounter*12 + 9] = (float)(nextSinAngle*-1);
		m_normalArray[facecounter*12 + 10] = (float)(nextCosAngle);
		m_normalArray[facecounter*12 + 11] = (float)(0);

		facecounter++;
	}

	//For-loop for the top of the cylinder.
	for(int i = 0; i < 360; i += (360/(numSides/2)))
	{
		float sinAngle = (float)sin(i*pi / 180);
		float nextSinAngle = (float)(sin((i + (360/numSides))*pi / 180));
		float thirdSinAngle = (float)(sin((i + (2*(360/numSides)))*pi / 180));
		float cosAngle = (float)cos(i*pi / 180);
		float nextCosAngle = (float)(cos((i + (360/numSides))*pi / 180));
		float thirdCosAngle = (float)(cos((i + (2*(360/numSides)))*pi / 180));

		m_vertexArray[facecounter*12] = (float)(x);
		m_vertexArray[facecounter*12 + 1] = (float)(y);
		m_vertexArray[facecounter*12 + 2] = (float)(z + height());
		m_vertexArray[facecounter*12 + 3] = (float)(x + ((float)sinAngle * radius())); 
		m_vertexArray[facecounter*12 + 4] = (float)(y + ((float)cosAngle * radius()));
		m_vertexArray[facecounter*12 + 5] = (float)(z + height());
		m_vertexArray[facecounter*12 + 6] = (float)(x + ((float)nextSinAngle * radius())); 
		m_vertexArray[facecounter*12 + 7] = (float)(y + ((float)nextCosAngle * radius()));
		m_vertexArray[facecounter*12 + 8] = (float)(z + height());
		m_vertexArray[facecounter*12 + 9] = (float)(x + ((float)thirdSinAngle * radius()));
		m_vertexArray[facecounter*12 + 10] = (float)(y + ((float)thirdCosAngle * radius()));
		m_vertexArray[facecounter*12 + 11] = (float)(z + height());

		m_normalArray[facecounter*12] = (float)(0);
		m_normalArray[facecounter*12 + 1] = (float)(0);
		m_normalArray[facecounter*12 + 2] = (float)(1);
		m_normalArray[facecounter*12 + 3] = (float)(0);
		m_normalArray[facecounter*12 + 4] = (float)(0);
		m_normalArray[facecounter*12 + 5] = (float)(1);
		m_normalArray[facecounter*12 + 6] = (float)(0);
		m_normalArray[facecounter*12 + 7] = (float)(0);
		m_normalArray[facecounter*12 + 8] = (float)(1);
		m_normalArray[facecounter*12 + 9] = (float)(0);
		m_normalArray[facecounter*12 + 10] = (float)(0);
		m_normalArray[facecounter*12 + 11] = (float)(1);

		facecounter++;
	}
	
	//For-loop for the bottom of the cylinder.
	for(int i = 0; i < 360; i += (360/(numSides/2)))
	{
		float sinAngle = (float)sin(i*pi / 180);
		float nextSinAngle = (float)(sin((i + (360/numSides))*pi / 180));
		float thirdSinAngle = (float)(sin((i + (2*(360/numSides)))*pi / 180));
		float cosAngle = (float)cos(i*pi / 180);
		float nextCosAngle = (float)(cos((i + (360/numSides))*pi / 180));
		float thirdCosAngle = (float)(cos((i + (2*(360/numSides)))*pi / 180));

		m_vertexArray[facecounter*12] = (float)(x);
		m_vertexArray[facecounter*12 + 1] = (float)(y);
		m_vertexArray[facecounter*12 + 2] = (float)(z);
		m_vertexArray[facecounter*12 + 3] = (float)(x + ((float)sinAngle * radius())); 
		m_vertexArray[facecounter*12 + 4] = (float)(y + ((float)cosAngle * radius()));
		m_vertexArray[facecounter*12 + 5] = (float)(z);
		m_vertexArray[facecounter*12 + 6] = (float)(x + ((float)nextSinAngle * radius())); 
		m_vertexArray[facecounter*12 + 7] = (float)(y + ((float)nextCosAngle * radius()));
		m_vertexArray[facecounter*12 + 8] = (float)(z);
		m_vertexArray[facecounter*12 + 9] = (float)(x + ((float)thirdSinAngle * radius()));
		m_vertexArray[facecounter*12 + 10] = (float)(y + ((float)thirdCosAngle * radius()));
		m_vertexArray[facecounter*12 + 11] = (float)(z);

		m_normalArray[facecounter*12] = (float)(0);
		m_normalArray[facecounter*12 + 1] = (float)(0);
		m_normalArray[facecounter*12 + 2] = (float)(-1);
		m_normalArray[facecounter*12 + 3] = (float)(0);
		m_normalArray[facecounter*12 + 4] = (float)(0);
		m_normalArray[facecounter*12 + 5] = (float)(-1);
		m_normalArray[facecounter*12 + 6] = (float)(0);
		m_normalArray[facecounter*12 + 7] = (float)(0);
		m_normalArray[facecounter*12 + 8] = (float)(-1);
		m_normalArray[facecounter*12 + 9] = (float)(0);
		m_normalArray[facecounter*12 + 10] = (float)(0);
		m_normalArray[facecounter*12 + 11] = (float)(-1);

		facecounter++;
	}
}

float* Kolf::CircleBlockObject::getNormalGeometry()
{
	return m_normalArray;
}

float* Kolf::CircleBlockObject::getVertexGeometry()
{
	return m_vertexArray;
}

#include "object-circleblock.moc"
