/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "physicsengine.h"
#include "ball.h"
#include "object.h"
#include "scene.h"

#include <ode/ode.h>
#include <QTimerEvent>

//constants
const int FramesPerSecond = 40;
const int MillisecondsPerStep = 1;
//derived constants
const qreal SecondsPerFrame = 1.0 / FramesPerSecond;
const int MillisecondsPerFrame = qMax(1000.0 * SecondsPerFrame, 1.0);
const qreal SecondsPerStep = MillisecondsPerStep / 1000.0;
const int StepsPerFrame = qMax(SecondsPerFrame / SecondsPerStep, 1.0);

Kolf::PhysicsEngine::PhysicsEngine(Kolf::Scene* scene)
	: m_timerID(-1)
	, m_scene(scene)
	, m_contactJointGroupID(dJointGroupCreate(0))
	, m_ballsMoving(false)
{
}

Kolf::PhysicsEngine::~PhysicsEngine()
{
	dJointGroupDestroy(m_contactJointGroupID);
}

Kolf::Scene* Kolf::PhysicsEngine::scene() const
{
	return m_scene;
}

void Kolf::PhysicsEngine::startSimulation()
{
	if (m_timerID == -1)
		m_timerID = startTimer(MillisecondsPerFrame);
}

void Kolf::PhysicsEngine::stopSimulation()
{
	if (m_timerID != -1)
	{
		killTimer(m_timerID);
		m_timerID = -1;
	}
}

void Kolf::PhysicsEngine::timerEvent(QTimerEvent* event)
{
	if (event->timerId() != m_timerID)
		return;
	//ensure that objects have correct gravity mode
	foreach (Kolf::Object* object, m_scene->objects())
		object->prepareKinematicState();
	//calculate next frame
	for (int i = 0; i < StepsPerFrame; ++i)
	{
		//create contact joints
		void* dataPointer = (void*) this;
		dSpaceCollide(m_scene->spaceID(), dataPointer, &Kolf::PhysicsEngine::resolveCollisionCallback);
		//execute custom interactions
		foreach (const Kolf::InteractionPair& pair, m_pendingCustomInteractions)
		{
			if (pair.first->simulationFlags() & Kolf::Object::CustomInteraction)
				pair.first->doCustomInteraction(pair.second);
			if (pair.second->simulationFlags() & Kolf::Object::CustomInteraction)
				pair.second->doCustomInteraction(pair.first);
		}
		m_pendingCustomInteractions.clear();
		//simulate next step
		dWorldQuickStep(m_scene->worldID(), SecondsPerStep);
		//remove contact joints for this step
		dJointGroupEmpty(m_contactJointGroupID);
	}
	//write results back to objects (esp. to their representations)
	foreach (Kolf::Object* object, m_scene->objects())
		object->updateKinematicState();
	//temporary friction implementation for balls
	bool ballsMovingNew = false;
	foreach (Kolf::Ball* ball, m_scene->balls())
	{
		Kolf::VectorF newVelocity = ball->velocity();
		QVectorF newVelocity2D = newVelocity.to2D();
		if (newVelocity2D * newVelocity2D > 1)
			//the "2" below and the "0.001" above are arbitrary values, which seem to produce good results
			newVelocity2D -= 2 * newVelocity2D / sqrt(newVelocity2D * newVelocity2D);
		else
			newVelocity2D = QVectorF();
		newVelocity.setX(newVelocity2D.x());
		newVelocity.setY(newVelocity2D.y());
		ball->setVelocity(newVelocity);
		if (!newVelocity2D.isNull())
			ballsMovingNew = true;
	}
	//detection of end of turn
	if (m_ballsMoving && !ballsMovingNew)
		emit turnEnded();
	m_ballsMoving = ballsMovingNew;
	//done
	emit nextFrameAvailable();
}

void Kolf::PhysicsEngine::resolveCollisionCallback(void* data, dGeomID o1, dGeomID o2)
{
	//find pointers to involved objects
	Kolf::PhysicsEngine* engine = reinterpret_cast<Kolf::PhysicsEngine*>(data);
	Kolf::Geometry* geometry1 = reinterpret_cast<Kolf::Geometry*>(dGeomGetData(o1));
	Kolf::Geometry* geometry2 = reinterpret_cast<Kolf::Geometry*>(dGeomGetData(o2));
	engine->resolveCollision(geometry1, geometry2);
}

void Kolf::PhysicsEngine::resolveCollision(Kolf::Geometry* geometry1, Kolf::Geometry* geometry2)
{
	Kolf::Object* object1 = geometry1->object();
	Kolf::Object* object2 = geometry2->object();
	const Kolf::Object::SimulationFlags sFlags1 = object1->simulationFlags();
	const Kolf::Object::SimulationFlags sFlags2 = object2->simulationFlags();
	//only continue if the given objects can participate in collisions
	const bool bothParticipate = (sFlags1 & sFlags2) & Kolf::Object::DynamicSimulation;
	const bool withCustomInteraction = (sFlags1 | sFlags2) & Kolf::Object::CustomInteraction;
	if (!bothParticipate && !withCustomInteraction)
		return;
	//let contact influence the bodies' movement only if requested by the object
	dBodyID body1 = dGeomGetBody(geometry1->id());
	dBodyID body2 = dGeomGetBody(geometry2->id());
	if (!(sFlags1 & Kolf::Object::KinematicSimulation))
		body1 = 0;
	if (!(sFlags2 & Kolf::Object::KinematicSimulation))
		body2 = 0;
	//contact joints are senseless if no object wants to have its moment calculated, except for when custom interaction is required
	if (body1 == 0 && body2 == 0 && !withCustomInteraction)
		return;
	//find contact points
	static const int maxContactNumber = 0x0010; //make at most 16 collision joints
	dContactGeom contactPoints[maxContactNumber];
	int contactNumber = dCollide(geometry1->id(), geometry2->id(), maxContactNumber, contactPoints, sizeof(dContactGeom));
	if (contactNumber == 0) //no contact points - no collision
		return;
	//add contact joints to simulation if necessary
	if (body1 != 0 || body2 != 0)
	{
		for (int i = 0; i < contactNumber; ++i)
		{
			//TODO: allow objects to modify the dContact instance (for different kinds of surfaces etc.)
			//specify behavior of contact joint
			dContact contact;
			contact.surface.mode = dContactApprox1;
			contact.surface.mu = 0;
			contact.geom = contactPoints[i];
			//create contact joint
			dJointID joint = dJointCreateContact(m_scene->worldID(), m_contactJointGroupID, &contact);
			dJointAttach(joint, body1, body2);
		}
	}
	//schedule custom interaction if necessary
	//ODE cannot execute custom interactions (such as position settings) while resolving collisions
	if (withCustomInteraction)
		m_pendingCustomInteractions << Kolf::InteractionPair(object1, object2);
}

#include "physicsengine.moc"
