/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "scene.h"
#include "ball.h"
#include "managedobjectgroup.h"
#include "object-coursemeta.h"
#include "putter.h"

#include <ode/ode.h>

QRectF Kolf::Scene::SceneRect(0, 0, 1000, 1000);

Kolf::Scene::Scene(QObject* parent)
	: QGraphicsScene(parent)
	, m_worldID(dWorldCreate())
	, m_spaceID(dHashSpaceCreate(0))
	, m_physicsEngine(new Kolf::PhysicsEngine(this))
	, m_objectGroup(new Kolf::ManagedObjectGroup(this))
	, m_putter(new Kolf::Putter(this))
{
	setSceneRect(Kolf::Scene::SceneRect);
	dWorldSetGravity(m_worldID, 0.0, 0.0, 0.0);//-9.81);
	//TODO: the following stuff is temporary
	m_physicsEngine->startSimulation();
	connect(m_physicsEngine, SIGNAL(turnEnded()), this, SIGNAL(turnEnded()));
	connect(m_objectGroup, SIGNAL(objectCreated(Kolf::ManagedObject*)), this, SLOT(objectCreated(Kolf::ManagedObject*)));
	connect(m_objectGroup, SIGNAL(objectAboutToBeDeleted(Kolf::ManagedObject*)), this, SLOT(objectAboutToBeDeleted(Kolf::ManagedObject*)));
}

Kolf::Scene::~Scene()
{
	delete m_physicsEngine;
	delete m_objectGroup;
	delete m_putter;
	dWorldDestroy(m_worldID);
	dSpaceDestroy(m_spaceID);
}

dWorldID Kolf::Scene::worldID() const
{
	return m_worldID;
}

dSpaceID Kolf::Scene::spaceID() const
{
	return m_spaceID;
}

QList<Kolf::Ball*> Kolf::Scene::balls() const
{
	return m_balls;
}

Kolf::ObjectList Kolf::Scene::objects() const
{
	return m_objects;
}

Kolf::PhysicsEngine* Kolf::Scene::physicsEngine() const
{
	return m_physicsEngine;
}

Kolf::ManagedObjectGroup* Kolf::Scene::objectGroup() const
{
	return m_objectGroup;
}

Kolf::Putter* Kolf::Scene::putter() const
{
	return m_putter;
}

Kolf::Ball* Kolf::Scene::addBall()
{
	//create ball
	Kolf::Ball* ball = new Kolf::Ball(m_worldID, m_spaceID);
	//set ball to the position specified in the course meta object (if there is one)
	Kolf::CourseMetaObject* cmObject = m_objects.allOfType<Kolf::CourseMetaObject>().value(0);
	if (cmObject)
		ball->setPosition(cmObject->position());
	else
		ball->setPosition(Kolf::PointF(500, 500, 100)); //a sane fallback above the middle of the scene
	//add ball to scene
	addObject(ball);
	m_balls << ball;
	m_objects << ball;
	return ball;
}

void Kolf::Scene::removeBall(Kolf::Ball* ball)
{
	m_balls.removeAll(ball);
	m_objects.removeAll(ball);
}

void Kolf::Scene::objectCreated(Kolf::ManagedObject* managedObject)
{
	Kolf::Object* object = managedObject->object();
	if (!m_objects.contains(object))
	{
		addObject(object);
		m_objects << object;
	}
}

void Kolf::Scene::objectAboutToBeDeleted(Kolf::ManagedObject* managedObject)
{
	Kolf::Object* object = managedObject->object();
	if (m_objects.contains(object))
	{
		removeObject(object);
		m_objects.removeAll(object);
	}
}

void Kolf::Scene::addObject(Kolf::Object* object)
{
	//add object to scene
	addItem(object);
	//restore object position
	const Kolf::PointF position = object->position();
	object->setPos(position.x(), position.y());
	//signals/slots
	connect(object, SIGNAL(subObjectAdded(Kolf::Object*)), this, SLOT(addObject(Kolf::Object*)));
	connect(object, SIGNAL(subObjectAboutToBeRemoved(Kolf::Object*)), this, SLOT(removeObject(Kolf::Object*)));
	//descend to subobjects
	foreach (Kolf::Object* subObject, object->subObjects())
		addObject(subObject);
}

void Kolf::Scene::removeObject(Kolf::Object* object)
{
	//remove object from scene
	removeItem(object);
	//signals/slots
	disconnect(object);
	//descend to subobjects
	foreach (Kolf::Object* subObject, object->subObjects())
		removeObject(subObject);
}

#include "scene.moc"
