/***************************************************************************
 *   Copyright 2009 Casper van Donderen <casper.vandonderen@gmail.com>
 *   Copyright 2009 Zeng Huan <zh.issac@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "view3d.h"
#include "../3d/renderstatemanager.h"
#include "../3d/texturepool.h"
#include "../elements/ball.h"
#include "../elements/geometry-terrain.h"
#include "../elements/object-environment.h"
#include "../elements/scene.h"

#include <GL/gl.h>
#include <GL/glu.h>
#include <QString>
#include <QTimer>

Kolf::View3D::View3D(QGLContext* context)
	: QGLWidget(context)
	, m_context(context)
	, m_scene(0)
	, m_rsm(new Kolf::RenderStateManager)
	, m_timer(new QTimer(this))
{
	setMinimumSize(100, 100);
	connect(m_timer, SIGNAL(timeout()), this, SLOT(updateGL()));
}

Kolf::View3D* Kolf::View3D::create()
{
	return new Kolf::View3D(new QGLContext(QGLFormat::defaultFormat()));
}

Kolf::View3D::~View3D()
{
	Kolf::TexturePool::instance()->detachContext(m_context);
	delete m_rsm;
}

void Kolf::View3D::setScene(Kolf::Scene* scene)
{
	if (m_scene == scene)
		return;
	m_scene = scene;
	if (m_scene)
		showEvent(0); //mimic show event to initialize objects for rendering
}

Kolf::RenderStateManager* Kolf::View3D::rsm() const
{
	return m_rsm;
}

void Kolf::View3D::stopTimer()
{ 
	m_timer->stop();
}

void Kolf::View3D::startTimer()
{ 
	m_timer->start(40);
}

/// This method stops the renderengine.
void Kolf::View3D::hideEvent(QHideEvent *)
{
	m_timer->stop();
}

void Kolf::View3D::initializeGL()
{
	qDebug("View3D:initializeGL");
	m_rsm->setDefaultRenderState();
	m_rsm->setGlobalBackgroundColor(Qt::black);
	//m_rsm->enableWireframeState();
}

void Kolf::View3D::paintGL()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	if (m_scene)
	{
		drawStaticObjects();
		drawAnimatedObjects();
		drawLandscape();
		drawBall();
	}
	glFlush();

	qDebug() << "frame end";
}

void Kolf::View3D::resizeGL(int w, int h)
{
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	double aspect = (double)w/h;
	gluPerspective(30, aspect, 1, 3000);
	m_rsm->setViewport(m_rsm->currentViewport());
	//m_rsm->setViewport(0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void Kolf::View3D::showEvent(QShowEvent *)
{
	//start rendering at 25 fps
	if (m_scene)
		initializeObjects();
	m_timer->start(40);
}

void Kolf::View3D::drawAnimatedObjects()
{
	//method to draw the animated objects. These objects can be vertex arrays, but then a glTranslatef() has to be called first and the Identity has to be reloaded. This operation is too expensive to do on every object when they are not animated.
}

void Kolf::View3D::drawBall()
{
	//method to draw the ball, do not know if this is possible at the moment. I need to 'find' the Ball object, get its X,Y,Z and the render a gluSphere at this position.
	foreach (Kolf::Ball* ball, m_scene->objects().allOfType<Kolf::Ball>())
	{
		GLUquadric *quad = gluNewQuadric(); //FIXME: possible memory leak
		glPushMatrix();
		QColor ballColor = ball->color();
		GLfloat materialColor[] = {ballColor.redF(), ballColor.greenF(), ballColor.blueF(), ballColor.alphaF()};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, materialColor);
		glTranslatef(ball->position().x(), ball->position().y(), ball->position().z());
		gluSphere(quad, 10, 10, 10);
		glPopMatrix();
	}
}

void Kolf::View3D::drawLandscape()
{
	//method to draw the landscape object
	foreach (Kolf::EnvironmentObject* object, m_scene->objects().allOfType<Kolf::EnvironmentObject>())
	{
		const Kolf::TerrainGeometry* geometry = object->terrainGeometry();
		const Kolf::HeightmapRenderer& heightmapRenderer = geometry->heightmapRenderer();
		if (heightmapRenderer.isNull())
			continue;

		//setup colors and lighting
		glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
		GLfloat materialColor[] = {1.0, 1.0, 1.0, 1.0};
		//GLfloat matSpecular[] = {1.0, 1.0, 1.0, 1.0};
		GLfloat matSpecular[] = {0.2, 0.2, 0.2, 0.2};
		GLfloat lowShininess[] = {5.0};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, materialColor);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpecular);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, lowShininess);

		//draw texture on heightmap
		GLuint textureID = Kolf::TexturePool::instance()->prepareTexture(geometry->textureKey(), m_context);
		glBindTexture(GL_TEXTURE_2D, textureID);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, heightmapRenderer.vertices().data());
		glNormalPointer(GL_FLOAT, 0 , heightmapRenderer.normals().data());
		glTexCoordPointer(2, GL_FLOAT, 0, heightmapRenderer.textureCoordinates().data());
		glDrawElements( GL_TRIANGLES, heightmapRenderer.indices().size(),
                                GL_UNSIGNED_INT, heightmapRenderer.indices().data());
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SINGLE_COLOR);

		//m_rsm->disableLightState();
		//glEnable(GL_COLOR_SUM);
		//glSecondaryColor3f(0.5, 0.5, 0.5);
		//glDisable(GL_COLOR_SUM);
		//m_rsm->enableLightState();
		glLoadIdentity();
	}

}

void Kolf::View3D::drawStaticObjects()
{
	foreach (Kolf::Object* object, m_scene->objects().allExceptType<Kolf::EnvironmentObject>())
	{
		if (!object->getNormalGeometry() || !object->getVertexGeometry())
			continue;
		//do not translate to the objects position, that position is already in the VertexGeometry.
		//glTranslatef(object->position().x(), object->position().y(), object->position().z());

		qDebug() << object->position() << (QObject*) object;

		GLfloat materialColor[] = {object->color().red() / 255.0, object->color().green() / 255.0, object->color().blue() / 255.0, object->color().alpha()}; 
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, materialColor);

		glEnableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_VERTEX_ARRAY);
		glNormalPointer(GL_FLOAT, 0, object->getNormalGeometry());
		glVertexPointer(3, GL_FLOAT, 0, object->getVertexGeometry());
		glDrawArrays(GL_QUADS, 0, object->numVerts / 3);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);

		glLoadIdentity();
    }
}

void Kolf::View3D::initializeObjects()
{
	foreach (Kolf::Object* object, m_scene->objects())
	{
		if(object->objectName() == QString("LightObject"))
		{
			GLfloat lightPosition[] = {object->position().x(), object->position().y(), object->position().z(), 0.0};
			GLfloat whiteLight[] = {1.0, 1.0, 1.0, 1.0};
			glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
			glLightfv(GL_LIGHT0, GL_DIFFUSE, whiteLight);
		}
		object->createSpatialGeometry();
	}
}

#include "view3d.moc"
