/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "managedobject.h"
#include "object.h"
#include "objectfactory.h"

Kolf::ManagedObject::ManagedObject(Kolf::ManagedObject::State* state, Kolf::ObjectFactory* objectFactory)
	: m_object(0)
	, m_objectFactory(objectFactory)
	, m_allowStateMerging(false)
{
	//initialize history with the current state
	beginPushState();
	m_states << state;
	endPushState();
	//create the object by kindly asking for it
	(void) object();
}

Kolf::ManagedObject::~ManagedObject()
{
	qDeleteAll(m_states);
	delete m_object;
}

Kolf::Object* Kolf::ManagedObject::object() const
{
	Kolf::ManagedObject::State* currentState = m_states[currentIndex()];
	if (!m_object && !currentState->isEmpty())
	{
		m_object = m_objectFactory->create(currentState->className());
		if (m_object)
		{
			connect(m_object, SIGNAL(updated()), this, SLOT(pushState()));
			currentState->applyToObject(m_object);
			Kolf::ManagedObject* mo = const_cast<Kolf::ManagedObject*>(this);
			emit mo->objectCreated(mo);
		}
	}
	return m_object;
}

Kolf::ManagedObject::State* Kolf::ManagedObject::state(int index) const
{
	return m_states.value(index);
}

void Kolf::ManagedObject::serialize(KConfigGroup* group)
{
	//TODO: For now, only the current state of the object is saved.
	m_states[currentIndex()]->serialize(group);
}

Kolf::ManagedObject* Kolf::ManagedObject::deserialize(KConfigGroup* group, Kolf::ObjectFactory* objectFactory)
{
	Kolf::ManagedObject::State* state = Kolf::ManagedObject::State::deserialize(group);
	return state ? new Kolf::ManagedObject(state, objectFactory) : 0;
}

void Kolf::ManagedObject::setAllowStateMerging(bool allowStateMerging)
{
	m_allowStateMerging = allowStateMerging;
}

void Kolf::ManagedObject::pushState()
{
	//create new state
	Kolf::ManagedObject::State* state = Kolf::ManagedObject::State::fromObject(m_object);
	//do not append state if both current and new state are empty
	const int currentIndex = this->currentIndex();
	if (m_states[currentIndex]->isEmpty() && state->isEmpty())
		return; //no new state has been added
	//do not append state, but merge it with the previous state if the difference to the previous state is not significant
	if (currentIndex != 0 && m_allowStateMerging)
	{
		Kolf::ManagedObject::State* previousState1 = m_states[currentIndex - 1];
		Kolf::ManagedObject::State* previousState2 = m_states[currentIndex];
		if (!Kolf::ManagedObject::State::isMediacySignificant(previousState1, previousState2, state))
		{
			m_states[currentIndex] = state;
			delete previousState2;
			return; //no new state has been added
		}
	}
	//append state
	beginPushState();
	m_states << state;
	endPushState();
}

void Kolf::ManagedObject::applyState(int index)
{
	//perform transition
	Kolf::ManagedObject::State* fromState = m_states[currentIndex()];
	Kolf::ManagedObject::State* toState = m_states[index];
	if (fromState->isEmpty())
	{
		if (!toState->isEmpty())
		{
			//create object
			m_object = m_objectFactory->create(toState->className());
			connect(m_object, SIGNAL(updated()), this, SLOT(pushState()));
			m_object->blockUpdates();
			if (m_object)
				toState->applyToObject(m_object);
			emit objectCreated(this);
		}
	}
	else // !fromState->isEmpty()
	{
		if (toState->isEmpty())
		{
			//delete object
			emit objectAboutToBeDeleted(this);
			delete m_object;
			m_object = 0;
		}
		else
			//apply new state to object
			toState->applyToObject(m_object);
	}
}

void Kolf::ManagedObject::removeStatesAbove(int index)
{
	while (index < m_states.count() - 1)
		m_states.removeLast();
}

#include "managedobject.moc"
