/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "managedobject.h"
#include "object.h"
#include "utils-bytearraycodec.h"

#include <QMetaObject>
#include <QMetaProperty>
#include <KConfigGroup>

Kolf::ManagedObject::State::State()
{
}

Kolf::ManagedObject::State::State(Kolf::Object* object)
	: m_className(object->metaObject()->className())
{
	//ignore properties defined by this class (and its superclasses)
	static int propertyOffset = QObject::staticMetaObject.propertyCount();
	//write all properties
	const QMetaObject* meta = object->metaObject();
	for (int propertyIndex = propertyOffset; propertyIndex < meta->propertyCount(); ++propertyIndex)
	{
		QMetaProperty property = meta->property(propertyIndex);
		m_propertyValues[property.name()] = property.read(object);
	}
}

Kolf::ManagedObject::State* Kolf::ManagedObject::State::fromObject(Kolf::Object* object)
{
	return object ? new Kolf::ManagedObject::State(object) : new Kolf::ManagedObject::State;
}

QByteArray Kolf::ManagedObject::State::className() const
{
	return m_className;
}

bool Kolf::ManagedObject::State::isEmpty() const
{
	return m_className.isEmpty() || m_propertyValues.isEmpty();
}

//Helper function for Kolf::ManagedObject::State::isMediacySignificant.
QList<QByteArray> Kolf::ManagedObject::State::findChangingProperties(Kolf::ManagedObject::State* fromState, Kolf::ManagedObject::State* toState)
{
	const QList<QByteArray> fromKeys = fromState->m_propertyValues.keys();
	const QList<QByteArray> toKeys = fromState->m_propertyValues.keys();
	QList<QByteArray> keys(fromKeys);
	//The tactic is:
	// 1. We go through keys and remove everything that has not changed in toState.
	// 2. We add any missing keys from toKeys.
	QMutableListIterator<QByteArray> iterKeys(keys);
	while (iterKeys.hasNext())
	{
		const QByteArray& key = iterKeys.next();
		if (!toKeys.contains(key))
			continue;
		//this property is also in toState - check value (but "convert" invalid QVariants to null QVariants of the other QVariant's type before checking)
		QVariant& fromValue = fromState->m_propertyValues[key];
		QVariant& toValue = toState->m_propertyValues[key];
		if (!fromValue.isValid() && toValue.isValid())
			fromValue = QVariant(toValue.type());
		else if (fromValue.isValid() && !toValue.isValid())
			toValue = QVariant(fromValue.type());
		if (fromValue == toValue)
			iterKeys.remove();
	}
	foreach (const QByteArray& key, toKeys)
		if (!fromKeys.contains(key))
			keys << key;
	return keys;
}

bool Kolf::ManagedObject::State::isMediacySignificant(Kolf::ManagedObject::State* previousState, Kolf::ManagedObject::State* mediacyState, Kolf::ManagedObject::State* followingState)
{
	//security check: a changing class name is very significant ;-)
	if (previousState->m_className != mediacyState->m_className || mediacyState->m_className != followingState->m_className)
		return true;
	//find differences between the three states
	const QList<QByteArray> diff1 = findChangingProperties(previousState, mediacyState);
	const QList<QByteArray> diff2 = findChangingProperties(mediacyState, followingState);
	//condition for significancy: diff1 != diff2
	foreach (const QByteArray& propertyKey, diff1)
		if (!diff2.contains(propertyKey))
			return true;
	foreach (const QByteArray& propertyKey, diff2)
		if (!diff1.contains(propertyKey))
			return true;
	return false;
}

void Kolf::ManagedObject::State::applyToObject(Kolf::Object* object) const
{
	if (!object)
		return;
	object->blockUpdates();
	//ignore properties defined by the superclasses
	static int propertyOffset = QObject::staticMetaObject.propertyCount();
	//read all properties
	const QMetaObject* meta = object->metaObject();
	for (int propertyIndex = propertyOffset; propertyIndex < meta->propertyCount(); ++propertyIndex)
	{
		QMetaProperty property = meta->property(propertyIndex);
		const QByteArray propertyName = property.name();
		QVariant& value = m_propertyValues[propertyName];
		//if value has just been deserialized (i.e. is a ByteArray), try to convert it to the type of the target property
		if (value.type() == QVariant::ByteArray)
			value = Utils::ByteArrayCodec::decode(value.value<QByteArray>(), property.read(object));
		property.write(object, value);
	}
	object->unblockUpdates();
}

void Kolf::ManagedObject::State::serialize(KConfigGroup* group)
{
	QMapIterator<QByteArray, QVariant> iterValues(m_propertyValues);
	while (iterValues.hasNext())
	{
		const QVariant value = iterValues.next().value();
		group->writeEntry(iterValues.key().data(), Utils::ByteArrayCodec::encode(value));
	}
	group->writeEntry("Type", m_className);
}

Kolf::ManagedObject::State* Kolf::ManagedObject::State::deserialize(KConfigGroup* group)
{
	Kolf::ManagedObject::State* state = new Kolf::ManagedObject::State;
	foreach (const QString& key, group->keyList())
	{
		if (key == QLatin1String("Type"))
			state->m_className = group->readEntry(key, QByteArray());
		else
			state->m_propertyValues[key.toLatin1()] = group->readEntry(key, QByteArray());
	}
	//check validity of format
	if (state->isEmpty())
	{
		delete state;
		return 0;
	}
	else
		return state;
}
