/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "object.h"
#include "overlay.h"
#include "scene.h"
#include "shape.h"

#include <Box2D/Dynamics/b2Body.h>
#include <Box2D/Dynamics/b2World.h>

Kolf::Object::Object(b2World* world)
	: m_body(0)
	, m_overlay(0)
	, m_blockUpdates(false)
	, m_type(Kolf::Object::CollisionSimulation)
{
	b2BodyDef bodyDef;
	bodyDef.userData = this; //NOTE: this is used by Kolf::ContactListener
	m_body = world->CreateBody(&bodyDef);
	//objects will most definitely not have mouse interaction
	setAcceptedMouseButtons(0);
	//enable some notifications which we need
	setFlag(QGraphicsItem::ItemSendsGeometryChanges);
	//NOTE: Kolf::Scene::addObject() is not called here because it needs a fully constructed Kolf::Object (including all subclass constructors!).
}

Kolf::Object::~Object()
{
	emit aboutToBeDestroyed(this); //WARNING: This must be the first line in this destructor.
	//The overlay is deleted first, because it might interact with all other parts of the object.
	delete m_overlay;
	//The order in which the Box2D-related objects are destroyed is very important.
	qDeleteAll(m_subObjects);
	qDeleteAll(m_shapes);
	m_body->GetWorld()->DestroyBody(m_body);
}

Kolf::ObjectFactory* Kolf::Object::objectFactory() const
{
	Kolf::Scene* scene = qobject_cast<Kolf::Scene*>(this->scene());
	return scene ? scene->objectFactory() : 0;
}

//BEGIN QGraphicsItem reimplementation

QRectF Kolf::Object::boundingRect() const
{
	return QRectF();
}

void Kolf::Object::paint(QPainter*, const QStyleOptionGraphicsItem*, QWidget*)
{
}

//END QGraphicsItem reimplementation
//BEGIN physics simulation

b2Body* Kolf::Object::body() const
{
	return m_body;
}

QList<Kolf::Shape*> Kolf::Object::shapes() const
{
	return m_shapes;
}

void Kolf::Object::addShape(Kolf::Shape* shape)
{
	if (shape->attach(this, m_body)) //this will fail if the shape is already attached to some object
		m_shapes << shape;
}

QList<Kolf::Object*> Kolf::Object::subObjects() const
{
	return m_subObjects;
}

void Kolf::Object::addSubObject(Kolf::Object* object)
{
	if (!m_subObjects.contains(object))
	{
		m_subObjects << object;
		scene()->addItem(object);
	}
}

void Kolf::Object::removeSubObject(Kolf::Object* object)
{
	if (m_subObjects.contains(object))
	{
		m_subObjects.removeAll(object);
		scene()->removeItem(object);
	}
}

void Kolf::Object::setRotation(qreal rotation)
{
	if (rotation != this->rotation())
	{
		this->QGraphicsItem::setRotation(rotation);
		objectEvent(RotationChangeEvent, qVariantFromValue(rotation));
	}
}

Kolf::Object::SimulationType Kolf::Object::simulationType() const
{
	return m_type;
}

void Kolf::Object::setSimulationType(Kolf::Object::SimulationType type)
{
	if (m_type == type)
		return;
	m_type = type;
	//write type into b2Body
	b2BodyType b2type; bool b2active;
	switch (type)
	{
		case Kolf::Object::NoSimulation:
			b2type = b2_staticBody;
			b2active = false;
			break;
		case Kolf::Object::CollisionSimulation:
			b2type = b2_staticBody;
			b2active = true;
			break;
		case Kolf::Object::KinematicSimulation:
			b2type = b2_kinematicBody;
			b2active = true;
			break;
		case Kolf::Object::DynamicSimulation: default:
			b2type = b2_dynamicBody;
			b2active = true;
			break;
	}
	m_body->SetType(b2type);
	m_body->SetActive(b2active);
}

QPointF Kolf::Object::velocity() const
{
	return m_velocity;
}

void Kolf::Object::setVelocity(const QPointF& velocity)
{
	if (m_velocity == velocity)
		return;
	m_velocity = velocity;
	m_velocityChanged = true;
	//fire change event if needed
	objectEvent(Kolf::Object::VelocityChangeEvent, velocity);
}

void Kolf::Object::prepareKinematicState()
{
	const QPointF position = pos();
	const qreal rotation = this->rotation() * M_PI / 180;
	m_body->SetTransform(b2Vec2(position.x(), position.y()), rotation);
	if (m_velocityChanged)
		m_body->SetLinearVelocity(b2Vec2(m_velocity.x(), m_velocity.y()));
	m_velocityChanged = false;
	//descend to subobjects
	QList<Kolf::Object*>::const_iterator it1 = m_subObjects.begin(), it2 = m_subObjects.end();
	for (; it1 != it2; ++it1)
		(*it1)->prepareKinematicState();
}

void Kolf::Object::updateKinematicState()
{
	//read position
	b2Vec2 v = m_body->GetPosition();
	setPos(v.x, v.y);
	//read rotation
	setRotation(m_body->GetAngle() / M_PI * 180);
	//read velocity
	v = m_body->GetLinearVelocity();
	const QPointF velocity(v.x, v.y);
	setVelocity(velocity);
	//unset m_velocityChanged if there was no manual change since last simulation step (there might already have been changes when the object has acted upon the VelocityChangeEvent we have just triggered)
	if (velocity == m_velocity)
		m_velocityChanged = false;
	//descend to subobjects
	QList<Kolf::Object*>::const_iterator it1 = m_subObjects.begin(), it2 = m_subObjects.end();
	for (; it1 != it2; ++it1)
		(*it1)->updateKinematicState();
}

//END physics simulation
//BEGIN property management

QVariant Kolf::Object::itemChange(QGraphicsItem::GraphicsItemChange change, const QVariant& value)
{
	if (change == QGraphicsItem::ItemPositionHasChanged)
	{
		propagateUpdate();
		objectEvent(Kolf::Object::PositionChangeEvent, value);
	}
	return QGraphicsItem::itemChange(change, value);
}

void Kolf::Object::objectEvent(Kolf::Object::EventType, const QVariant&)
{
}

void Kolf::Object::blockUpdates()
{
	m_blockUpdates = true;
}

void Kolf::Object::unblockUpdates(bool propagateChangesAsUpdate)
{
	m_blockUpdates = false;
	if (propagateChangesAsUpdate)
		propagateUpdate();
}

//END property management
//BEGIN editor management

Kolf::Overlay* Kolf::Object::overlay(bool createIfNecessary)
{
	//the overlay is created once it is requested
	if (!m_overlay && createIfNecessary)
	{
		m_overlay = createOverlay();
		Q_ASSERT(m_overlay != 0);
		m_overlay->setZValue(100); //should be above object representation
		m_overlay->update(); //initialize the overlay's parameters
	}
	return m_overlay;
}

bool Kolf::Object::isEditable() const
{
	return true;
}

bool Kolf::Object::showRotatorHandleOnOverlay() const
{
	return false;
}

void Kolf::Object::propagateUpdate()
{
	//Note: Overlays still need to be updated even in blocking mode.
	if (m_overlay)
		m_overlay->update();
	if (!m_blockUpdates)
		emit updated();
}

//END editor management

#include "object.moc"
