/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "objectfactory.h"
#include "object.h"

#include <QMutexLocker>
#include <KGlobal>

//BEGIN object registry

/**
 * \internal \class ObjectRegistry
 * This thread-safe singleton class is a database of all known instantiable Kolf::Object subclasses.
 * It is filled by Kolf::ObjectRegistrar (and hence, the REGISTER_KOLF_OBJECT macro).
 * Its public interface is Kolf::ObjectFactory.
 */
struct ObjectRegistry
{
	public:
		QMutex mutex;
		QMap<Kolf::ObjectFactory::Class*, Kolf::ObjectCreator> creators;

		~ObjectRegistry()
		{
			QMap<Kolf::ObjectFactory::Class*, Kolf::ObjectCreator>::const_iterator
				it1 = creators.begin(), it2 = creators.end();
			for (; it1 != it2; ++it1)
				delete it1.key();
		}
};

K_GLOBAL_STATIC(ObjectRegistry, objectRegistry)

Kolf::ObjectFactory::Class::Class()
	: icon(KIcon(""))
	, editorMayCreateInstances(true)
{
}

Kolf::ObjectRegistrar::ObjectRegistrar(const QMetaObject* metaObject, Kolf::ObjectCreator creator, Kolf::MetadataProvider provider)
{
	//create metadata object
	Kolf::ObjectFactory::Class* classObj = new Kolf::ObjectFactory::Class;
	provider(classObj); //fill class object with metadata
	classObj->metaObject = metaObject;
	//insert into registry
	QMutexLocker l(&objectRegistry->mutex);
	objectRegistry->creators.insert(classObj, creator);
}

//END object registry
//BEGIN Kolf::ObjectFactory

Kolf::ObjectFactory::ObjectFactory(Kolf::Scene* scene, b2World* world)
	: m_scene(scene)
	, m_world(world)
{
}

QList<const Kolf::ObjectFactory::Class*> Kolf::ObjectFactory::knownObjectClasses() const
{
	QList<const Kolf::ObjectFactory::Class*> result;
	QMutexLocker l(&objectRegistry->mutex);
	QMap<Kolf::ObjectFactory::Class*, Kolf::ObjectCreator>::const_iterator
		it1 = objectRegistry->creators.begin(), it2 = objectRegistry->creators.end();
	for (; it1 != it2; ++it1)
		result << it1.key();
	return result;
}

const Kolf::ObjectFactory::Class* Kolf::ObjectFactory::objectClass(const QByteArray& objectClassName) const
{
	QMutexLocker l(&objectRegistry->mutex);
	QMap<Kolf::ObjectFactory::Class*, Kolf::ObjectCreator>::const_iterator
		it1 = objectRegistry->creators.begin(), it2 = objectRegistry->creators.end();
	for (; it1 != it2; ++it1)
		if (it1.key()->metaObject->className() == objectClassName)
			return it1.key();
	return 0; //no matching class found
}

Kolf::Object* Kolf::ObjectFactory::create(const Kolf::ObjectFactory::Class* objectClass, Kolf::Object* parentObject)
{
	Kolf::ObjectCreator creator;
	{
		QMutexLocker l(&objectRegistry->mutex);
		creator = objectRegistry->creators.value(const_cast<Kolf::ObjectFactory::Class*>(objectClass), 0);
	}
	return create(creator, parentObject);
}

Kolf::Object* Kolf::ObjectFactory::create(const QByteArray& objectClassName, Kolf::Object* parentObject)
{
	Kolf::ObjectCreator creator = 0;
	{
		QMutexLocker l(&objectRegistry->mutex);
		QMap<Kolf::ObjectFactory::Class*, Kolf::ObjectCreator>::const_iterator
			it1 = objectRegistry->creators.begin(), it2 = objectRegistry->creators.end();
		for (; it1 != it2; ++it1)
			if (it1.key()->metaObject->className() == objectClassName)
			{
				creator = it1.value();
				break;
			}
	}
	return create(creator, parentObject);
}

Kolf::Object* Kolf::ObjectFactory::create(Kolf::ObjectCreator creator, Kolf::Object* parentObject)
{
	if (!creator)
		return 0;
	Kolf::Object* obj = creator(m_world);
	obj->setParent(parentObject);
	emit objectCreated(obj);
	return obj;
}

//END Kolf::ObjectFactory

#include "objectfactory.moc"
