/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "overlay.h"
#include "object.h"
#include "overlayareaitem.h"
#include "overlayhandle.h"
#include "shape.h"
#include "utils-animateditem.h"

#include <QPen>
#include <cmath>

const qreal Kolf::Overlay::MinimumObjectDimension = 0.2;
const qreal Kolf::Overlay::RotatorHandleDistance = 1;

Kolf::Overlay::Overlay(Kolf::Object* object)
	: QGraphicsItem(object)
	, m_object(object)
	, m_state(Kolf::Overlay::Passive)
	, m_activatorItem(new Kolf::OverlayAreaItem(Kolf::OverlayAreaItem::Clickable | Kolf::OverlayAreaItem::Hoverable, this))
	, m_interactorAnimator(new Utils::AnimatedItem(this))
	, m_interactorItem(new Kolf::OverlayAreaItem(Kolf::OverlayAreaItem::Clickable | Kolf::OverlayAreaItem::Draggable, m_interactorAnimator))
	, m_handleAnimator(new Utils::AnimatedItem(this))
	, m_rotatorHandle(0)
	, m_rotatorHelper(0)
{
	//overlays have to be shown explicitly
	hide();
	//overlays themselves do not have mouse interaction
	setAcceptedMouseButtons(0);
	//initialize activator area item
	m_activatorItem->setZValue(1);
	connect(m_activatorItem, SIGNAL(hoverEntered()), this, SLOT(activatorEntered()));
	connect(m_activatorItem, SIGNAL(hoverLeft()), this, SLOT(activatorLeft()));
	connect(m_activatorItem, SIGNAL(clicked(int)), this, SLOT(activatorClicked(int)));
	//initialize interactor area item
	m_interactorAnimator->setZValue(2);
	m_interactorAnimator->setOpacity(0); //not visible at first
	m_interactorItem->setBrush(Qt::green);
	connect(m_interactorItem, SIGNAL(clicked(int)), this, SLOT(activatorClicked(int))); //note that interactor item is over activator in hovered mode
	connect(m_interactorItem, SIGNAL(dragged(const QPointF&)), this, SLOT(interactorDragged(const QPointF&)));
	//initialize handle manager
	m_handleAnimator->setZValue(3);
	m_handleAnimator->setHideWhenInvisible(true);
	m_handleAnimator->setOpacity(0); //not visible at first
	//create rotator if necessary - the position must only be set once, as the coordinate system of the object rotates the rotator handle as well
	if (object->showRotatorHandleOnOverlay())
	{
		QPointF rotatorHandlePos(Kolf::Overlay::RotatorHandleDistance, 0);
		m_rotatorHandle = new Kolf::OverlayHandle(Kolf::OverlayHandle::CircleShape, this);
		m_rotatorHandle->setPos(rotatorHandlePos);
		connect(m_rotatorHandle, SIGNAL(moveRequest(const QPointF&)), this, SLOT(rotatorDragged(const QPointF&)));
		m_rotatorHelper = new QGraphicsLineItem(QLineF(QPointF(), rotatorHandlePos), this);
		m_rotatorHelper->setPen(Qt::DashLine);
		//show rotator handle together with the other handles
		addHandle(m_rotatorHandle);
		addHandle(m_rotatorHelper);
	}
	//apply passive state - we need to change to some other state to prevent the setState method from returning early
	m_state = Active;
	setState(Passive);
}

Kolf::Object* Kolf::Overlay::object() const
{
	return m_object;
}

void Kolf::Overlay::update()
{
	//update geometry outlines
	QPainterPath activationOutline, interactionOutline;
	foreach (Kolf::Shape* shape, m_object->shapes())
	{
		activationOutline.addPath(shape->activationOutline());
		interactionOutline.addPath(shape->interactionOutline());
	}
	activationOutline.setFillRule(Qt::WindingFill);
	interactionOutline.setFillRule(Qt::WindingFill);
	m_activatorItem->setPath(activationOutline);
	m_interactorItem->setPath(interactionOutline);
}

void Kolf::Overlay::addHandle(QGraphicsItem* handle)
{
	handle->setParentItem(m_handleAnimator);
	handle->show();
}

void Kolf::Overlay::addHandle(Kolf::OverlayHandle* handle)
{
	addHandle((QGraphicsItem*) handle);
	connect(handle, SIGNAL(moveStarted()), this, SLOT(handleInteractionStarted()));
	connect(handle, SIGNAL(moveEnded()), this, SLOT(handleInteractionEnded()));
}

Kolf::Overlay::State Kolf::Overlay::state() const
{
	return m_state;
}

void Kolf::Overlay::setState(Kolf::Overlay::State state)
{
	if (m_state == state)
		return;
	m_state = state;
	//apply new inner properties
	switch (state)
	{
		case Kolf::Overlay::Passive:
			m_interactorAnimator->setOpacityAnimated(0.0);
			break;
		case Kolf::Overlay::Hovered:
			m_interactorAnimator->setOpacityAnimated(0.3);
			break;
		case Kolf::Overlay::Active:
			m_interactorAnimator->setOpacityAnimated(0.6);
			break;
	}
	m_handleAnimator->setOpacityAnimated(state == Active ? 1.0 : 0.0);
	//propagate changes
	emit stateChanged();
}

void Kolf::Overlay::handleInteractionStarted()
{
	m_object->blockUpdates();
}

void Kolf::Overlay::handleInteractionEnded()
{
	m_object->unblockUpdates(true); //true == propagate changes in object's properties as update
}

void Kolf::Overlay::activatorEntered()
{
	if (m_state == Kolf::Overlay::Passive)
		setState(Kolf::Overlay::Hovered);
}

void Kolf::Overlay::activatorLeft()
{
	if (m_state == Kolf::Overlay::Hovered)
		setState(Kolf::Overlay::Passive);
}

void Kolf::Overlay::activatorClicked(int button)
{
	Q_UNUSED(button)
	setState(Kolf::Overlay::Active);
}

void Kolf::Overlay::interactorDragged(const QPointF& distance)
{
	if (m_state == Kolf::Overlay::Active)
		m_object->setPos(m_object->pos() + distance);
}

void Kolf::Overlay::rotatorDragged(const QPointF& targetScenePos)
{
	const QPointF targetPos = mapFromScene(targetScenePos);
	//note that handle is always at (RotatorHandleDistance, 0) because it is rotated together with the object - measured angles are difference angles
	qreal rotationRadians = atan2(targetPos.y(), targetPos.x());
	int rotationDegrees = rotationRadians / M_PI * 180;
	m_object->setRotation(m_object->rotation() + rotationDegrees);
}

//default implementation for QGraphicsItem

QRectF Kolf::Overlay::boundingRect() const
{
	return QRectF();
}

void Kolf::Overlay::paint(QPainter*, const QStyleOptionGraphicsItem*, QWidget*)
{
}

#include "overlay.moc"
