/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_OVERLAY_H
#define KOLF_OVERLAY_H

#include <QGraphicsItem>

namespace Utils
{
	class AnimatedItem;
}

namespace Kolf
{
	class Object;
	class OverlayAreaItem;
	class OverlayExtenderItem;
	class OverlayHandle;

	/**
	 * \class Overlay
	 * \since 2.0
	 *
	 * Overlays are used by the editor to provide the on-screen editing interface. The most wellknown overlay is probably the rectangle with grips at every corner and edge. Overlays react on changes in the property set of their object and change specific properties through user interaction.
	 *
	 * By default, the Overlay renders interaction and activation outlines and allows to change the object's position; subclasses have to add handles to manipulate geometrical properties of the object. (This class may become pure virtual in the future, we therefore recommend to not instance this class. If you want an overlay without handles, use Kolf::EmptyOverlay instead.)
	 *
	 * \warning Do not add any handles or additional interface items through the addItem() method. Use the addHandle() method to ensure that handles are correctly shown and hidden.
	 * \sa Kolf::Object
	 */
	class Overlay : public QObject, public QGraphicsItem
	{
		Q_OBJECT
		Q_INTERFACES(QGraphicsItem)
		public:
			///This enum specifies states that an overlay can occupy.
			enum State
			{
				Passive = 0, ///< The overlay is invisible (not as in isVisible() == false, but all components are transparent or invisible).
				Hovered = 1, ///< The overlay's activation area has been hovered, and the interactor area is shown (though very translucent).
				Active = 2   ///< The interaction area is shown at a high opacity, and the overlay's handles are visible.
			};

			Overlay(Kolf::Object* object);
			///Returns a reference to the object that this overlay is associated with.
			Kolf::Object* object() const;
			///Returns the current state of this overlay.
			State state() const;
			///Performs a state transition from the current towards the given \a state.
			void setState(State state);
			///Lets the overlay initialize its properties or react on property changes in the base object. Subclasses of Kolf::Object should not use this method, but the Kolf::Object::propagateUpdate() method. Subclasses of Kolf::Overlay need to reimplement this to read all important properties of their object (aside from calling the base class implementation!).
			virtual void update();

			///Overlays should not allow to decrease an object's dimensions below this level, for the sake of usability.
			static const qreal MinimumObjectDimension;
			///This is the distance in scene coordinates that the rotator handle will always have from the object's reference point.
			static const qreal RotatorHandleDistance;

			virtual QRectF boundingRect() const;
			virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = 0);
		Q_SIGNALS:
			///This signal is emitted if the overlay's state changes.
			void stateChanged();
		protected:
			///Adds a handle to the overlay interface. Handles are only shown if the overlay is visible and activated.
			void addHandle(QGraphicsItem* handle);
			///Identical to the other addHandle overload, but the handle's moveStarted/moveEnded signals are automatically connected to the overlay's handleInteractionStarted() and handleInteractionEnded() slots.
			void addHandle(Kolf::OverlayHandle* handle);
		protected Q_SLOTS:
			///Trigger this slot when the user starts to interact with a handle.
			void handleInteractionStarted();
			///Trigger this slot when the user stops to interact with a handle.
			void handleInteractionEnded();
		private Q_SLOTS:
			void activatorEntered();
			void activatorLeft();
			void activatorClicked(int button);
			void interactorDragged(const QPointF& distance);
			void rotatorDragged(const QPointF& targetScenePos);
		private:
			Kolf::Object* m_object;
			Kolf::Overlay::State m_state;
			//pre-defined handles and area items
			Kolf::OverlayAreaItem* m_activatorItem;
			Utils::AnimatedItem* m_interactorAnimator;
			Kolf::OverlayAreaItem* m_interactorItem;
			Utils::AnimatedItem* m_handleAnimator;
			Kolf::OverlayHandle* m_rotatorHandle;
			QGraphicsLineItem* m_rotatorHelper;
	};
}

#endif // KOLF_OVERLAY_H
