/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "putter.h"
#include "object-ball.h"

#include <QGraphicsScene>
#include <QGraphicsSceneEvent>

class Kolf::Putter::Item : public QObject, public QGraphicsItem
{
	public:
		Item(QGraphicsScene* scene, Kolf::Putter* parent);
		virtual bool eventFilter(QObject* watched, QEvent* event);

		//empty QGraphicsItem reimplementation
		virtual QRectF boundingRect() const { return QRectF(); }
		virtual void paint(QPainter*, const QStyleOptionGraphicsItem*, QWidget*) {}
	protected:
		virtual void mouseMoveEvent(QGraphicsSceneMouseEvent* event);
		virtual void mousePressEvent(QGraphicsSceneMouseEvent* event);
	private:
		Kolf::Putter* m_parent;
};

//BEGIN Kolf::Putter::Item

Kolf::Putter::Item::Item(QGraphicsScene* scene, Kolf::Putter* parent)
	: m_parent(parent)
{
	scene->addItem(this);
	scene->installEventFilter(this);
	hide(); //until a ball is attached to the putter
	setAcceptedMouseButtons(0); //we rely on the subclass to specify which buttons we want
	setAcceptHoverEvents(true);
}

bool Kolf::Putter::Item::eventFilter(QObject* watched, QEvent* event)
{
	if (watched == scene() && event->type() == QEvent::Leave)
		m_parent->sendUpdate(0, QPointF());
	return false; //we are only monitoring, not filtering
}

void Kolf::Putter::Item::mouseMoveEvent(QGraphicsSceneMouseEvent* event)
{
	m_parent->sendUpdate(event->widget(), event->pos());
}

void Kolf::Putter::Item::mousePressEvent(QGraphicsSceneMouseEvent* event)
{
	m_parent->handleMousePress(event->button(), event->pos());
}

//END Kolf::Putter::Item

//BEGIN Kolf::Putter

Kolf::Putter::Putter(QGraphicsScene* scene)
	: m_ball(0)
	, m_item(new Kolf::Putter::Item(scene, this))
{
	m_lastUpdate.activeView = 0;
	//All the initialization is done in Kolf::Putter::Item::Item.
}

Kolf::Putter::~Putter()
{
	delete m_item;
}

bool Kolf::Putter::isActive() const
{
	return (bool) m_ball;
}

Kolf::BallObject* Kolf::Putter::ball() const
{
	return m_ball;
}

void Kolf::Putter::setBall(Kolf::BallObject* ball)
{
	if (m_ball && m_ball == ball)
		return;
	//detach from current ball
	if (m_ball)
	{
		Kolf::BallObject* oldBall = m_ball;
		disconnect(m_ball);
		m_ball = 0;
		//done detaching
		emit detached(oldBall);
	}
	//attach to new ball
	if (ball)
	{
		m_ball = ball;
		connect(m_ball, SIGNAL(destroyed(QObject*)), this, SLOT(ballDeleted()));
		//resend last update after switching to new coordinate system
		const QPointF posDiff = m_ball->pos() - m_item->pos();
		m_item->setPos(m_ball->pos());
		m_lastUpdate.cursorPosition -= posDiff;
		update(m_lastUpdate.activeView, m_lastUpdate.cursorPosition);
		//done attaching
		emit attached(ball);
	}
	//set new visibility
	const bool isVisible = isActive();
	if (isVisible == m_item->isVisible())
		return;
	m_item->setVisible(isVisible);
	//mouse grabbing
	QGraphicsItem* mouseGrabber = m_item->scene()->mouseGrabberItem();
	if (isVisible)
	{
		if (mouseGrabber != m_item)
		{
			if (mouseGrabber)
				mouseGrabber->ungrabMouse();
			m_item->grabMouse();
		}
	}
	else //if (!isVisible)
	{
		//NOTE: The ungrabbing should already have been happened when hiding m_item, but it's good to be sure.
		if (mouseGrabber == m_item)
			m_item->ungrabMouse();
	}
}

QGraphicsItem* Kolf::Putter::graphicsItem() const
{
	return m_item;
}

void Kolf::Putter::sendUpdate(QWidget* activeView, const QPointF& cursorPosition)
{
	//We store the arguments of this update call, to be able to resend the update when attaching to a new ball.
	m_lastUpdate.activeView = activeView;
	m_lastUpdate.cursorPosition = cursorPosition;
	update(activeView, cursorPosition);
}

void Kolf::Putter::ballDeleted()
{
	m_ball = 0;
	setBall(0);
}

void Kolf::Putter::setVelocity(const QPointF& velocity)
{
	if (!m_ball)
		return;
	m_ball->setVelocity(velocity);
	emit putted(m_ball, velocity);
	setBall(0); //detach from ball
}

//END Kolf::Putter

#include "putter.moc"
