/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_PUTTER_H
#define KOLF_PUTTER_H

#include <QGraphicsItem>

namespace Kolf
{
	class BallObject;

	/**
	 * \class Putter
	 * \since 2.0
	 *
	 * A putter provides the putting interface, and sets the ball velocity accordingly.
	 */
	class Putter : public QObject
	{
		Q_OBJECT
		public:
			///Creates a new putter on the given \a scene. The putter will not be visible until it is attached to a ball.
			Putter(QGraphicsScene* scene);
			virtual ~Putter();

			///Returns whether the putter is currently active. This is functionally equivalent to:
			///\code putter->ball() != 0 \endcode
			///as well as:
			///\code putter->isVisible() \endcode
			///This function is provided to improve the semantics of your code.
			bool isActive() const;
			///Returns the ball that this putter is currently attached to.
			Kolf::BallObject* ball() const;
			///Attaches the putter to the given \a ball. If this putter is currently attached to another ball, it will detach from this ball first. You may safely give a null pointer, the putter will then be hidden.
			void setBall(Kolf::BallObject* ball);
		Q_SIGNALS:
			///This signal is emitted after the putter has attached to a new \a ball.
			void attached(Kolf::BallObject* ball);
			///This signal is emitted after the putter has detached from a \a ball.
			void detached(Kolf::BallObject* ball);
			///This signal is emitted when the player has used the putter to assign a velocity to the a ball. When the signal is emitted, the ball's velocity has already been adjusted accordingly. After the signal has been emitted, the putter will detach from the ball.
			void putted(Kolf::BallObject* ball, const QPointF& velocity);
		protected:
			///Returns the root QGraphicsItem instance for this putter. Its pos() will always match the ball while the putter is active. Add the custom interface of your Kolf::Putter subclass below this item.
			QGraphicsItem* graphicsItem() const;
			///Reimplement this method to react on mouse interaction.
			///\warning To receive any events here, call QGraphicsItem::setAcceptedMouseButtons() on graphicsItem().
			virtual void handleMousePress(Qt::MouseButton button, const QPointF& cursorPosition) = 0;
			///Reimplement this method to update the appearance of the putter after the cursor has been moved, or when the putter has attached to a new ball and is shown for the first time. There are two possible cases:
			///\li \a activeView is a null pointer: The mouse is currently not hovering any view.
			///\li \a activeView is not a null pointer: The mouse cursor is currently above \a activeView at \a cursorPosition, given in the coordinate system of the putter's graphicsItem().
			virtual void update(QWidget* activeView, const QPointF& cursorPosition) = 0;
			///Call this method when the user has selected a velocity. The putter will hide itself, detach from the ball, and send out the appropriate signals.
			void setVelocity(const QPointF& velocity);
		private Q_SLOTS:
			void ballDeleted();
		private:
			void sendUpdate(QWidget* activeView, const QPointF& cursorPosition);

			Kolf::BallObject* m_ball;
			struct {
				QWidget* activeView;
				QPointF cursorPosition;
			} m_lastUpdate;
			//Kolf::Putter::Item behaves much like a Private implementation class
			class Item; friend class Item;
			Kolf::Putter::Item* m_item;
	};
}

#endif // KOLF_PUTTER_H
