/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "game-normal.h"
#include "course.h"
#include "player-local.h"
#include "scoremodel.h"
#include "../base/object-ball.h"
#include "../base/putter.h"
#include "../base/scene.h"

#include <QTimer>

Kolf::NormalGame::NormalGame(Kolf::Course* course, Kolf::Interface* interface)
	: Kolf::Game(course, interface)
	, m_timer(new QTimer(this))
	, m_currentPlayerIndex(-1)
{
	m_timer->setInterval(500);
	m_timer->setSingleShot(true);
}

Kolf::Player* Kolf::NormalGame::addLocalPlayer(const QString& name, int hue)
{
	Kolf::Player* player = new Kolf::LocalPlayer(name, hue);
	Kolf::Game::addPlayer(player);
	connect(player, SIGNAL(ballObjectCreated(Kolf::BallObject*)), this, SLOT(ballObjectCreated(Kolf::BallObject*)));
	return player;
}

void Kolf::NormalGame::bootHole(Kolf::Hole* hole)
{
	m_currentPlayerIndex = -1;
	if (hole)
	{
		hole->scene()->startSimulation();
		callDelayed(SLOT(selectNextPlayer()));
		connect(hole->putter(), SIGNAL(putted(Kolf::BallObject*, const QPointF&)), this, SLOT(putted(Kolf::BallObject*)));
	}
}

void Kolf::NormalGame::shutdownHole(Kolf::Hole* hole)
{
	m_currentPlayerIndex = -1;
	hole->scene()->stopSimulation();
	//TODO: save state of balls, and restore it when this hole is selected again
}

void Kolf::NormalGame::ballObjectCreated(Kolf::BallObject* ballObject)
{
	connect(ballObject, SIGNAL(stopped()), this, SLOT(ballStopped()));
}

void Kolf::NormalGame::ballStopped()
{
	//advance to next player unless a) putter is active or b) some balls are still moving
	Kolf::Hole* hole = currentHole();
	if (!hole)
		return;
	foreach (Kolf::Player* player, players())
	{
		if (player->hasControl())
			return;
		Kolf::BallObject* ball = player->ball(false);
		if (ball && !ball->velocity().isNull())
			return;
	}
	callDelayed(SLOT(selectNextPlayer()));
}

void Kolf::NormalGame::putted(Kolf::BallObject* ball)
{
	Kolf::Hole* hole = currentHole();
	//increase score for this player
	foreach (Kolf::Player* player, players())
		if (player->ball(false) == ball)
		{
			int score = scoreModel()->score(hole, player);
			scoreModel()->setScore(hole, player, score + 1);
		}
}

void Kolf::NormalGame::selectNextPlayer()
{
	const QList<Kolf::Player*> players = this->players();
	//find next player (i.e. a player who's ball is not in a hole yet)
	const int oldPlayerIndex = m_currentPlayerIndex;
	int newPlayerIndex = m_currentPlayerIndex;
	bool inHole;
	do
	{
		newPlayerIndex = (newPlayerIndex + 1) % players.count();
		inHole = players[newPlayerIndex]->ball()->isInHole();
		if (newPlayerIndex == oldPlayerIndex && inHole)
		{
			//looped through all players -> no player can play -> all players are in the hole -> proceed to the next hole
			callDelayed(SLOT(selectNextHole()));
			return;
		}
	}
	while (inHole);
	//turn control to this player
	m_currentPlayerIndex = newPlayerIndex;
	players[newPlayerIndex]->giveControl();
}

void Kolf::NormalGame::selectNextHole()
{
	setCurrentHoleIndex(currentHoleIndex() + 1);
}

void Kolf::NormalGame::callDelayed(const char* localSlot)
{
	m_timer->stop();
	m_timer->start();
	disconnect(m_timer, SIGNAL(timeout()), this, 0);
	connect(m_timer, SIGNAL(timeout()), this, localSlot);
}

#include "game-normal.moc"
