/*
 * KLevelEditor.cpp
 *
 * Copyright (C) 2007 Keith Neuse <KNeuse@Gmail.com>
 */ 

#include "KGLevelEditor.h"
#include <KDebug>

KGLevelEditor::KGLevelEditor(QWidget * parent, const char *) : 
QGraphicsView(parent)
{
    //setup default variables
    mouseX = 0;
    mouseY = 0;
    mainTile = NULL;
    mousePressed = false;
    
    //optimized viewport
    setViewportUpdateMode(QGraphicsView::SmartViewportUpdate);

    //init functions
    initGraphicsView();
    initLayers();

    KGContainer *container = KGSelectionInterface::instance()->container();
    KGObject *object = KGSelectionInterface::instance()->selectedObject();
    eraserMode = KGSelectionInterface::instance()->eraserMode();
    
    addLayer(TILE_LAYER);
    addLayer(OBJECT_LAYER);

    connect(KGSelectionInterface::instance(),SIGNAL(currentObjectChanged(KGObject *, int)),this,SLOT(setCurrentObject(KGObject *, int)));
    connect(KGSelectionInterface::instance(),SIGNAL(eraserModeChanged(bool)),this,SLOT(setEraserMode(bool)));
    connect(KGSelectionInterface::instance(),SIGNAL(layerIDChanged(int)),this,SLOT(setWorkingLayer(int)));
    setCurrentObject(object,container->layerIDFromObject(object));
    setWorkingLayer(KGSelectionInterface::instance()->layerID());
}
KGLevelEditor::~KGLevelEditor(void)
{
    if(mainTile)
	delete mainTile;

    while(scene.items().size() > 0)
    {
	KGObject *deletetile = (KGObject *)scene.items().front();
	scene.items().pop_front();
	delete deletetile;
    }
}

bool KGLevelEditor::open()
{ 
    KGContainer *container = KGSelectionInterface::instance()->container();
    QFile openLevel(m_location.path());
    QDomDocument doc;

    openLevel.open(QIODevice::ReadWrite);
    doc.setContent(&openLevel);
    openLevel.close();
    QDomElement element = doc.documentElement();
    QDomNodeList elementList = element.childNodes();

    for(unsigned int loop = 0; loop < elementList.length(); loop++)
    {
	QDomElement objectElement = elementList.item(loop).toElement();
	KGObject *object = KGSelectionInterface::instance()->container()->findObject(objectElement.attribute("reference",QString()));
	if(!object)
	    continue;
	
	QDomElement locationElement = objectElement.firstChildElement("location");
	QString y = locationElement.attribute("y",QString());
	QString x = locationElement.attribute("x",QString());	
	addObject(object,KGSelectionInterface::instance()->container()->layerIDFromObject(object),QPoint(x.toInt(),y.toInt()));
    }
    return true;
}
bool KGLevelEditor::save()
{
    kDebug() << "Saving level" << endl;
    QFile saveLevel(m_location.path());
    if(!saveLevel.open(QIODevice::ReadWrite | QIODevice::Truncate))
	return false;
    
    QDomDocument doc("KGML");
    QDomElement root = doc.createElement("KGML");
    doc.appendChild(root);

    //find the layer index
    foreach(KGLayerInfo layerInfo, layerList)
    {
	foreach(QGraphicsItem *item, layerInfo.layerObject->children())
	{
	    if(item == mainTile)
		continue;
	    QDomElement objectElement = doc.createElement("tile");
	    objectElement.setAttribute("reference",((KGObject *)item)->reference());

	    QDomElement locationElement = doc.createElement("location");
	    locationElement.setAttribute("x",item->x());
	    locationElement.setAttribute("y",item->y());
	    objectElement.appendChild(locationElement);

	    root.appendChild(objectElement);
	}
    }
    
    QTextStream textStream(&saveLevel);
    textStream << doc.toString();
    saveLevel.close();
    return true;
}
void KGLevelEditor::setReference(QString newReference)
    { reference = newReference; }
QString KGLevelEditor::getReference()
    { return reference; }

void KGLevelEditor::setLayerHide(int layerID, bool hide)
{
    //find layer by layerID
    int layerNum = findLayer(layerID);

    //if found
    if(layerNum != -1)
    {
	//store the new hide state
	layerList[layerNum].hiden = hide;

	//remove layer from scene to hide it
	if(hide)
	    scene.removeItem(layerList[layerNum].layerObject);
	else //add layer to scene to show it
	    layerGroup->addToGroup(layerList[layerNum].layerObject);
    }
}
void KGLevelEditor::setWorkingLayer(int layerID)
{
    //find the layer index
    int newWorkingLayer = findLayer(layerID);

    //if found set workingLayer to that index
    if(newWorkingLayer != -1)
	workingLayer = newWorkingLayer;
}
void KGLevelEditor::initGraphicsView(void)
{
    //add tracking
    setMouseTracking (true);

    //setup QGraphicsView
    setAlignment(Qt::AlignLeft | Qt::AlignTop);
    
    //setup scene
    scene.setSceneRect(0, 0, 5000, 5000);
    setScene(&scene);

    //scroll to top left corner
    centerOn(0,0);
}
int KGLevelEditor::addLayer(int layerID)
{
    //if layer id already found don't add it
    if(findLayer(layerID) != -1)
	return -1;

    //setup new layer object
    KGLayerInfo newLayer;
    newLayer.layerObject = new QGraphicsItemGroup();
    newLayer.layerNum = layerList.size();
    newLayer.layerID = layerID;
    newLayer.hiden = false;
    
    //add layer
    layerList.push_back(newLayer);
    layerGroup->addToGroup(newLayer.layerObject);

    //setup layer properties within QGraphicsView
    newLayer.layerObject->setZValue(layerList.size()-1);
    newLayer.layerObject->setData(KEY_TYPE,PARENT_ITEM);
    
    return newLayer.layerNum;
}
int KGLevelEditor::findLayer(int layerID)
{
    //test each layer for the layerID
    foreach(KGLayerInfo layerItem, layerList)
    {
	if(layerItem.layerID == layerID)
	    return layerItem.layerNum;
    }

    return -1;
}
void KGLevelEditor::initLayers(void)
{
    //create master Lyaer
    layerGroup = new QGraphicsItemGroup();
    layerGroup->setData(KEY_TYPE,PARENT_ITEM);
    scene.addItem(layerGroup);

    //create mouse layer
    mouseLayer = addLayer(MOUSE_ID);
    layerList[mouseLayer].layerObject->setZValue(1000);
    
    //set the current working layer
    workingLayer = NULL_LAYER;
}

void KGLevelEditor::setEraserMode(bool argEraserMode)
{
    eraserMode = argEraserMode;
}
void KGLevelEditor::setCurrentObject(KGObject *newTile, int layerID)
{
    if(!newTile)
	return;
    //first remove the current mouse object
    if(mainTile)
    {
	scene.removeItem(mainTile);
	layerList[mouseLayer].layerObject->removeFromGroup(mainTile);
	delete mainTile;
    }

    setWorkingLayer(layerID);

    //then add the new mouse object
    mainTile = newTile->clone();
    layerList[mouseLayer].layerObject->addToGroup(mainTile);
    mainTile->setData(KEY_TYPE,PARENT_ITEM);
    mainTile->setPos(mouseX,mouseY);
    //setEraserMode(false);
}
void KGLevelEditor::addObject(KGObject *object, int layerID, QPoint pos)
{
	KGObject *newTile = object->clone();
	newTile->setData(KEY_TYPE,TILE_ITEM);
	layerList[findLayer(layerID)].layerObject->addToGroup(newTile);
	newTile->setPos(pos);
}
//This function generates tiles when the user places down a new tile
void KGLevelEditor::placeObject(QPoint mousePos)
{
    if(mousePos.x() < 0 || mousePos.y() < 0 || workingLayer == NULL_LAYER || layerList[workingLayer].hiden || !mainTile)
	return;
    QList<QGraphicsItem *> itemList = scene.items(mousePos);

//******************************************************************************
//Note: This allows tiles to replace other tiles

    foreach(QGraphicsItem *graphicItem,itemList)
    {
	if(graphicItem->parentItem() == layerList[workingLayer].layerObject)
	    delete graphicItem;
    }
//******************************************************************************


//******************************************************************************
//Note: this doesn't allow tiles to replace other tiles

    /*if(eraserMode)
    {
	foreach(QGraphicsItem *graphicItem,itemList)
	{
	    if(graphicItem->parentItem() == layerList[workingLayer].layerObject)
		delete graphicItem;
	}
    }
    else
    {
	foreach(QGraphicsItem *graphicItem,itemList)
	{
	    
	    if(graphicItem->parentItem() == layerList[workingLayer].layerObject)
		return;
	}
    }*/

//******************************************************************************
    if(!eraserMode)
    {
	KGObject *newTile = mainTile->clone();
	newTile->setData(KEY_TYPE,TILE_ITEM);
	layerList[workingLayer].layerObject->addToGroup(newTile);
	newTile->setPos(mouseX,mouseY);
    }
}

//**************************************
//Events
//**************************************

void KGLevelEditor::mouseReleaseEvent ( QMouseEvent *){mousePressed = false;}

void KGLevelEditor::mousePressEvent ( QMouseEvent *)
{
    mousePressed = true;
    placeObject(QPoint(mouseX,mouseY));
}
QList<KGObject *> KGLevelEditor::getObjects(int layerID)
{
    QList<KGObject *> objectList;

    //find the layer index
    if(-1 == findLayer(layerID))
	return objectList;

    foreach(QGraphicsItem *item, layerList[workingLayer].layerObject->children())
	objectList.push_back(qgraphicsitem_cast<KGObject *>(item));

    return objectList;
}
void KGLevelEditor::mouseMoveEvent ( QMouseEvent * event )
{
    //vars
    QPointF newMouse;
    
    //convert mouse cordinates to viewport cordinates
    newMouse = mapToScene(event->pos());

    //inforce a grid layout here
    mouseX = (int)(newMouse.x()/50)*50;
    mouseY = (int)(newMouse.y()/50)*50;

    if(mainTile)
	mainTile->setPos(mouseX,mouseY);

    //place object if user holds and drags mouse
    if(mousePressed)
	placeObject(QPoint(mouseX,mouseY));
}

#include "KGLevelEditor.moc"
