/****************************************************************************
 *   Copyright (C) 2005-2007 by Albert Astals Cid <aacid@kde.org>           *
 * Copyright (C) 2007-2008 by Michal Sciubidlo <michal.sciubidlo@gmail.com> *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 ****************************************************************************/

#include "boardwidget.h"
#include "boardwidget.moc"

#include <cmath>

#include <QApplication>
#include <QAction>
#include <QDrag>
#include <QDragMoveEvent>
#include <QDropEvent>
#include <QMenu>
#include <QMouseEvent>
#include <QPainter>
#include <QStringList>
#include <QGraphicsSceneMouseEvent>
#include <QStyleOptionGraphicsItem>
#include <QPointF>
#include <QRectF>

#include <kdebug.h>

#include "tile.h"
#include "tilefactory.h"
#include "tilewidget.h"
#include "utils.h"
#include "settings.h"
#include "blanktiledialog.h"
#include "renderer.h"

boardWidget::boardWidget(blankTileDialog * dlg, QMenu * itemMenu)
{
	m_tileSize = 32;
	m_activeSquare = QPoint(-1, -1);
	m_startdrag = QPoint(-1, -1);
	m_menu = itemMenu;
	m_dlg = dlg;
	//setMouseTracking(true);
	//setAcceptDrops(true);

	//setMinimumSize(15 * 30, 15 * 30);
	//setTileSize(30);
}

boardWidget::~boardWidget()
{
}

void boardWidget::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
	Q_UNUSED(widget);

	QRect updatingRect = option->exposedRect.toRect();
	
	for (int i = 0; i < 15; i++)
	{
		QRect column((int)(m_tileSize * i), 0, (int)m_tileSize, (int)(m_tileSize * 15));
		if (column.intersects(updatingRect))
		{
			for (int j = 0; j < 15; j++)
			{
				QRect tile ((int)(m_tileSize * i), (int)(m_tileSize * j), (int)m_tileSize, (int)m_tileSize);
				if (tile.intersects(updatingRect))
				{
					painter->drawPixmap( QPointF(i*m_tileSize, j*m_tileSize), renderer::self()->boardElement(m_squares[i][j].kind()));
					if (m_squares[i][j].ourTile())
					{
						painter->drawPixmap( QPointF(i*m_tileSize, j*m_tileSize), renderer::self()->tileBackground( m_squares[i][j].ourTile()->isBlank() ));
						painter->drawText(QRectF(i*m_tileSize, j*m_tileSize, m_tileSize, m_tileSize), Qt::AlignCenter, m_squares[i][j].ourTile()->string());
					}
				}
			}
		}
	}
}

QRectF boardWidget::boundingRect() const
{
	return QRectF(0, 0, m_tileSize*15, m_tileSize*15);
}

void boardWidget::resize(double size)
{
	prepareGeometryChange();
	m_tileSize = size;
}

void boardWidget::nextPlayer(const QString &, bool isItMe)
{
	// TODO (un)block all mouse and keyboard acctions ?
	setAcceptDrops(isItMe);
}

void boardWidget::slotSetPlayerTiles(const tilePositionList &tiles)
{
	setPlayedTiles(tiles);
	update();
}

void boardWidget::rejectPlayerTiles()
{
	finalizePlayedTiles(DELETE);
	update();
}

void boardWidget::slotLockTilesAdded()
{
	finalizePlayedTiles(ACCEPT);
	m_addedtiles.clear();
	update();
}

void boardWidget::removeDroppedTile(tile * t)
{
	if (Utils::isValidPoint(m_startdrag))
	{
		QPoint tmp = Utils::calcMousePositionPoint(m_startdrag, m_tileSize);
		delete m_squares[tmp.x()][tmp.y()].ourTile();
		removeTileFromNewlyAdded(tmp);
		m_squares[tmp.x()][tmp.y()].setTile(t);

		if (t)
		{
			tilePosition tp;
			tp.setX(tmp.x());
			tp.setY(tmp.y());
			tp.setTile(t);
			m_addedtiles << tp;
		}

		m_startdrag = QPoint(-1, -1);
		update();
	}
}

const QList< QPair<QString, int > > &boardWidget::getTileStringList() const
{
	return m_tilesStringList;
}

void boardWidget::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
	QPoint oldSquare = m_activeSquare;
	m_activeSquare = Utils::calcMousePositionPoint(event->pos().toPoint(), m_tileSize);
	if ((oldSquare != m_activeSquare) && Settings::highlightActiveSquare())
	{
		int sizeBig = (int)ceil(m_tileSize);
		update((int)floor(m_tileSize * oldSquare.x()), (int)floor(m_tileSize * oldSquare.y()), sizeBig, sizeBig);
		update((int)floor(m_tileSize * m_activeSquare.x()), (int)floor(m_tileSize * m_activeSquare.y()), sizeBig, sizeBig);
	}
	if (!(event->buttons() & Qt::LeftButton)) return;

	if ((event->pos().toPoint() - m_startdrag).manhattanLength() < QApplication::startDragDistance()) return;

	QPoint tmp = Utils::calcMousePositionPoint(m_startdrag, m_tileSize);
	bool locked = m_squares[tmp.x()][tmp.y()].isLocked();
	tile *t = m_squares[tmp.x()][tmp.y()].ourTile();
	if (t && !locked)
	{
		QDrag *drag = Utils::createDragForTile(event->widget(), (QObject*)this, renderer::self()->tileBackground( t->isBlank() ), t, m_tileSize);
		Qt::DropAction dropAction = drag->start(Qt::MoveAction);
		Q_UNUSED(dropAction);
		m_startdrag = event->pos().toPoint();
	}
	else
	{
		m_startdrag = QPoint(-1, -1);
	}
}

void boardWidget::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
	if (event->button() == Qt::LeftButton)
		m_startdrag = event->pos().toPoint();
	else if (event->button() == Qt::RightButton)
	{
		QPoint tmp = Utils::calcMousePositionPoint(event->pos().toPoint(), m_tileSize);
		tile *t = m_squares[tmp.x()][tmp.y()].ourTile();
		//if (!m_squares[tmp.x()][tmp.y()].isLocked() && t && t->isBlank())
		{
			QAction *ret = m_menu->exec(event->screenPos());
			if (!ret) return;
			int which = ret->data().toInt();
			if (which == 0)
			{
				//connect(m_dlg, SIGNAL(tileChanged()), &m_scene, SLOT(updateSceneRect()));
				m_dlg->setData(t, m_tilesStringList);
				m_dlg->exec();
				//disconnect(m_dlg, SIGNAL(tileChanged()), &m_scene, SLOT(updateSceneRect()));
			}
		}
	}
}

void boardWidget::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
	QPoint tmp = Utils::calcMousePositionPoint(event->pos().toPoint(), m_tileSize);
	tile *t = m_squares[tmp.x()][tmp.y()].ourTile();
	if (event->button() == Qt::LeftButton && !m_squares[tmp.x()][tmp.y()].isLocked() && t)
	{
		QDrag *drag = Utils::createDragForTile(event->widget(), (QObject*)this, renderer::self()->tileBackground( t->isBlank() ),  t, m_tileSize);
		Qt::DropAction dropAction = drag->start(Qt::MoveAction);
		Q_UNUSED(dropAction);
		m_startdrag = event->pos().toPoint();
	}
	event->accept();
}

void boardWidget::dragEnterEvent(QGraphicsSceneDragDropEvent * event)
{
	// always accept the drop, as will be the dragMoveEvent to take care of checking where we are
	event->acceptProposedAction();
}

void boardWidget::dragMoveEvent(QGraphicsSceneDragDropEvent * event)
{
	m_activeSquare = QPoint(-1, -1);
	QPoint tmp = Utils::calcMousePositionPoint(event->pos().toPoint(), m_tileSize);

	if (event->mimeData()->hasText() && (m_squares[tmp.x()][tmp.y()].ourTile()==0 || !m_squares[tmp.x()][tmp.y()].isLocked()))
		event->acceptProposedAction();
	else
		event->ignore();
}

void boardWidget::dropEvent(QGraphicsSceneDragDropEvent * event)
{
	const Utils::sMimeData *myData =  qobject_cast<const Utils::sMimeData *>(event->mimeData());

	m_activeSquare = Utils::calcMousePositionPoint(event->pos().toPoint(), m_tileSize);

	tile *t = tileFactory::simpleXMLDeserialization(event->mimeData()->text());
	if (t)
	{
		QObject *source = myData->senderItem();
		if ((source == this) && Utils::isValidPoint(m_startdrag))
		{
			QPoint tmp = Utils::calcMousePositionPoint(m_startdrag, m_tileSize);
			//dropping on self
			if (tmp.x() == m_activeSquare.x() && tmp.y() == m_activeSquare.y())
			{
				m_startdrag = QPoint(-1, -1);
				event->acceptProposedAction();
				return;
			}

			removeTileFromNewlyAdded(tmp);
			m_squares[tmp.x()][tmp.y()].setTile(m_squares[m_activeSquare.x()][m_activeSquare.y()].ourTile());

			//dropping tile on tile
			if (m_squares[m_activeSquare.x()][m_activeSquare.y()].ourTile())
			{
				tilePosition ttp;
				ttp.setX(tmp.x());
				ttp.setY(tmp.y());
				ttp.setTile(m_squares[tmp.x()][tmp.y()].ourTile());
				m_addedtiles << ttp;
			}
		}
		else if (dynamic_cast<tileWidget*>(source))
		{
			if (m_squares[m_activeSquare.x()][m_activeSquare.y()].ourTile())
				removeTileFromNewlyAdded(QPoint(m_activeSquare.x(), m_activeSquare.y()));
			static_cast<tileWidget*>(source)->removeDroppedTile( m_squares[m_activeSquare.x()][m_activeSquare.y()].ourTile());
		}
		t->setKind(tile::HIGHLIGHT);

		removeTileFromNewlyAdded(m_activeSquare);
		m_squares[m_activeSquare.x()][m_activeSquare.y()].setTile(t);
		tilePosition tp;
		tp.setX(m_activeSquare.x());
		tp.setY(m_activeSquare.y());
		tp.setTile(t);
		m_addedtiles << tp;
	}
	m_startdrag = QPoint(-1, -1);


	event->acceptProposedAction();

	update();
}

void boardWidget::setTileStringList(const QList< QPair<QString, int> > &tileStrings)
{
	m_tilesStringList = tileStrings;
}

void boardWidget::removeTileFromNewlyAdded(const QPoint &p)
{
	int i = m_addedtiles.size() - 1;
	while (i >= 0)
	{
		if ((m_addedtiles.at(i).x() == p.x()) && (m_addedtiles.at(i).y() == p.y()))
		{
			m_addedtiles.removeAt(i);
		}
		--i;
	}	
}

bool boardWidget::wereTilesAdded() const
{
	return !m_addedtiles.empty();
}

tilePositionList boardWidget::tilesAdded() const
{
	return m_addedtiles;
}

void boardWidget::slotDeleteLastTiles()
{
	for (int i = 0; i < m_addedtiles.size(); i++)
	{
		if (m_squares[m_addedtiles.at(i).x()][m_addedtiles.at(i).y()].isLocked() == false)
		{
			//m_squares[m_addedtiles.at(i).x()][m_addedtiles.at(i).y()].setLocked(false);
			m_squares[m_addedtiles.at(i).x()][m_addedtiles.at(i).y()].setTile(0L);
		}
	}
	m_addedtiles.clear();
}

void boardWidget::reset()
{
	board::reset();
	update();
}
