/***************************************************************************
 *   Copyright (C) 2007 Dmitry Suzdalev <dimsuz@gmail.com>                 *
 *   Copyright (C) 2007 by Michal Sciubidlo <michal.sciubidlo@gmail.com>   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "renderer.h"

#include <QPainter>

#include <KSvgRenderer>
#include <KDE/KGameTheme>
#include <kpixmapcache.h>
#include <kdebug.h>
#include "settings.h"

renderer::renderer()
{
	m_renderer = new KSvgRenderer();
	m_cache = new KPixmapCache("kombination-pixmaps");
	m_cache->setCacheLimit(512);

	m_rackBackground = 0;
	m_blankBackground = 0;
	m_tileBackground = 0;
	m_boardBackground[0] = 0;
	m_boardBackground[1] = 0;
	m_boardBackground[2] = 0;
	m_boardBackground[3] = 0;
	m_boardBackground[4] = 0;
	m_boardBackground[5] = 0;

	m_size.setWidth(32);
	m_size.setHeight(32);
	if(!loadTheme( Settings::style() ))
	//if (!loadTheme( "graphics/default.desktop" ))
		kDebug() << "Failed to load any game theme!";
}

renderer::~renderer()
{
	delete m_renderer;
	delete m_cache;
}

renderer* renderer::self()
{
	static renderer instance;
	return &instance;
}

bool renderer::loadTheme(const QString &themeName)
{
	// variable saying whether to discard old cache upon successful new theme loading
	// we won't discard it if m_currentTheme is empty meaning that
	// this is the first time loadTheme() is called
	// (i.e. during startup) as we want to pick the cache from disc
	bool discardCache = !m_currentTheme.isEmpty();

	if( !m_currentTheme.isEmpty() && m_currentTheme == themeName )
	{
		kDebug() << "Notice: not loading the same theme";
		return true; // this is not an error
	}

	m_currentTheme = themeName;

	KGameTheme theme;
	if ( themeName.isEmpty() || !theme.load( themeName ) )
	{
		//kDebug()<< "Failed to load theme" << Settings::theme();
		kDebug() << "Trying to load default";
		if(!theme.loadDefault())
			return false;
		// in this case we need to discard any previously cached theme:
		// we loading default
		discardCache = true;
		// set it to something - not be empty - so we
		// do discard cache on next loadTheme
		// (see discardCache above)
		m_currentTheme = "default";
	}

	bool res = m_renderer->load( theme.graphics() );
	kDebug() << "loading" << theme.graphics();
	if ( !res )
		return false;

	if(discardCache)
	{
		kDebug() << "discarding cache";
		m_cache->discard();
	}

	loadTile(QString("rackBackground"), m_rackBackground);
	loadTile(QString("blankBackground"), m_blankBackground);
	loadTile(QString("tileBackground"), m_tileBackground);
	loadTile(QString("normal"), m_boardBackground[0]);
	loadTile(QString("center"), m_boardBackground[1]);
	loadTile(QString("doubleword"), m_boardBackground[2]);
	loadTile(QString("tripleword"), m_boardBackground[3]);
	loadTile(QString("doubleletter"), m_boardBackground[4]);
	loadTile(QString("tripleletter"), m_boardBackground[5]);

	return true;
}

void renderer::loadTile(const QString &elemName, QPixmap*& dest)
{
	QString cacheName = elemName + QString::number(m_size.width()) + 'x' + QString::number(m_size.height());
	if(!m_cache->find( cacheName, *dest ))
	{
		if (dest)
		{
			delete dest;
			dest = 0;
		}
		dest = new QPixmap(m_size.toSize());
		dest->fill(Qt::transparent);
		QPainter p(dest);
		m_renderer->render(&p, elemName);
		p.end();
		m_cache->insert(cacheName, *dest);
		kDebug() << "cache size:" << m_cache->size() << "kb";
	}
}

QPixmap renderer::tileBackground(bool isBlank)
{
	if (isBlank == 0)
		return *m_tileBackground;
	else
		return *m_blankBackground;
}

QPixmap renderer::boardElement(square::Kind elemType)
{
	return *(m_boardBackground[elemType]);
}

QPixmap renderer::rackBackground()
{
	return *m_rackBackground;
}

void renderer::resize( double size )
{
	setCellSize( QSizeF(size, size) );
}

void renderer::setCellSize( const QSizeF &size )
{
	m_size = size;
	loadTile(QString("rackBackground"), m_rackBackground);
	loadTile(QString("blankBackground"), m_blankBackground);
	loadTile(QString("tileBackground"), m_tileBackground);
	loadTile(QString("normal"), m_boardBackground[0]);
	loadTile(QString("center"), m_boardBackground[1]);
	loadTile(QString("doubleword"), m_boardBackground[2]);
	loadTile(QString("tripleword"), m_boardBackground[3]);
	loadTile(QString("doubleletter"), m_boardBackground[4]);
	loadTile(QString("tripleletter"), m_boardBackground[5]);
}
