/***************************************************************************
 *   Copyright (C) 2005-2007 by Albert Astals Cid <aacid@kde.org>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "spectator.h"

#include <QStringList>
#include <QVariant>
#include <QTcpSocket>

#include <kdebug.h>

#include "game.h"
#include "messagehandler.h"
#include "protocol.h"
#include "tile.h"

spectator::spectator(QTcpSocket *socket) : m_phase(WAITING_FOR_SPECTATOR), m_game(0), m_socket(socket)
{
	m_messageHandler = new messageHandler();
	m_messageHandler->setSocket(socket);
	
	connect(m_messageHandler, SIGNAL(newMessage(const message &)), this, SLOT(processNewMessage(const message &)));
	connect(m_socket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(slotSocketError(QAbstractSocket::SocketError)));
	message m;
	m.type = "welcome";
	m.data.insert("version", QString::number(VERSION));
	m_messageHandler->sendMessage(m);
}

spectator::spectator() : m_phase(WAITING_TO_START_GAME), m_game(0)
{
}

spectator::spectator(spectator *s)
{
	m_messageHandler = s->m_messageHandler;
	m_socket = s->m_socket;
	
	connect(m_messageHandler, SIGNAL(newMessage(const message &)), this, SLOT(processNewMessage(const message &)));
	connect(m_socket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(slotSocketError(QAbstractSocket::SocketError)));
	
	m_nick = s->m_nick;
	m_phase = s->m_phase;
	
	s->m_messageHandler = 0;
	s->m_socket = 0;
	
	delete s;
}

spectator::~spectator()
{
	delete m_messageHandler;
	deleteSocket();
}

void spectator::deleteSocket()
{
	if (m_socket)
	{
		m_socket->close();
		m_socket->deleteLater();
	}
}

bool spectator::isSpectator()
{
	return m_isSpectator;
}

void spectator::askForNick()
{
	m_phase = WAITING_FOR_NICK;
	
	message m;
	m.type = "sendNick";
	m_messageHandler->sendMessage(m);
}

void spectator::tooManyPlayers()
{
	message m;
	m.type = "tooManyPlayers";
	m_messageHandler->sendMessage(m);
}

const QString &spectator::nick() const
{
	return m_nick;
}

bool spectator::wasNickInUse() const
{
	return m_phase == NICK_ALREADY_IN_USE;
}

void spectator::setCurrentGame(game *g)
{
	m_game = g;
}

game *spectator::currentGame() const
{
	return m_game;
}

void spectator::acceptedInServer(const QStringList &players, QList<game *> games)
{
	int i;
	message m;
	m.type = "acceptedIntoServer";
	m.data.insert("players", QString::number(players.count()));
	i = 0;
	foreach(const QString &nick, players)
	{
		m.data.insert(QString("player%1").arg(i), nick);
	}
	m.data.insert("games", QString::number(games.count()));
	i = 0;
	foreach(const game *g, games)
	{
		m.data.unite(g->toMap(QString("game%1").arg(i)));
		++i;
	}
	m_messageHandler->sendMessage(m);
}

void spectator::addGame(game *g)
{
	if (!m_game)
	{
		// do not care about new games if I am on one
		message m;
		m.type = "newGame";
		m.data = g->toMap("game");
		m_messageHandler->sendMessage(m);
	}
}

void spectator::canNotCreateGame(const QString &reason)
{
	message m;
	m.type = "canNotCreateGame";
	m.data.insert("reason", reason);
	m_messageHandler->sendMessage(m);
}

void spectator::acceptedInGame(const QList< QPair<QString, int> > &players, const QString &language)
{
	message m;
	m.type = "acceptedIntoGame";
	m.data.insert("language", language);
	m.data.insert("count", QString::number(players.count()));
	for (int i = 0; i < players.count(); i++)
	{
		m.data.insert( QString("nick%1").arg(i), players[i].first );
		m.data.insert( QString("status%1").arg(i), QString::number(players[i].second) );
	}
	m_messageHandler->sendMessage(m);
}

void spectator::nickInUse()
{
	m_phase = NICK_ALREADY_IN_USE;
	
	message m;
	m.type = "nickInUse";
	m_messageHandler->sendMessage(m);
	
	//deleteLater();
}

void spectator::doesntAllowSpectators()
{
	message m;
	m.type = "serverDoesNotAllowSpectators";
	m_messageHandler->sendMessage(m);
}

void spectator::sendMessage(const QString &nick, const QString &text)
{
	message m;
	m.type = "message";
	m.data.insert("nick", nick);
	m.data.insert("message", text);
	m_messageHandler->sendMessage(m);
}

void spectator::startGame()
{
	m_phase = PLAYING_GAME;
	
	message m;
	m.type = "startingGame";
	m_messageHandler->sendMessage(m);

}

void spectator::sendNewTiles(const QList<tile*> &tiles)
{
	Q_UNUSED(tiles);
	
	message m;
	m.type = "newTiles";
	m_messageHandler->sendMessage(m);
}

void spectator::addPlayer(const QString &nick, bool isPlayer)
{
	message m;
	m.type = "newPlayer";
	m.data.insert("nick", nick);
	m.data.insert("status", isPlayer ? QString::number(STATUS_PLAYER_NOT_READY) : QString::number(STATUS_SPECTATOR));
	m_messageHandler->sendMessage(m);
}

void spectator::addPlayer(const QString &nick)
{
	if (!m_game)
	{
		// if I am playing I do not care for new players not in my game
		message m;
		m.type = "newPlayer";
		m.data.insert("nick", nick);
		m_messageHandler->sendMessage(m);
	}
}

void spectator::otherPlayerEnteredAGame(const QString &nick, int gameId)
{
	if (!m_game)
	{
		// if I am playing I do not care
		message m;
		m.type = "otherPlayerEnteredAGame";
		m.data.insert("nick", nick);
		m.data.insert("game", QString::number(gameId));
		m_messageHandler->sendMessage(m);
	}
}

void spectator::nextTurn(const QString &nick)
{
	m_phase = PLAYING_GAME;
	
	message m;
	m.type = "nextTurn";
	m.data.insert("nick", nick);
	m_messageHandler->sendMessage(m);
}

void spectator::sendTiles(const tilePositionList &tiles)
{
	Q_UNUSED(tiles);
	m_privTiles = tiles;
}

void spectator::correctWord(int score)
{
	m_phase = PLAYING_GAME;
	
	message m;
	m.type = "sendingTiles";
	m.data = m_privTiles.toMap();
	m_messageHandler->sendMessage(m);
	
	message m2;
	m2.type = "correctWords";
	m2.data.insert("score", QString::number(score));
	m_messageHandler->sendMessage(m2);
}

void spectator::unknownWords()
{
	m_phase = TRADING_PHASE;
}

void spectator::gameEnded(const QString &finisher, const QString &winner)
{
	m_phase = WAITING_TO_START_GAME;
	
	message m;
	m.type = "gameEnded";
	m.data.insert("winner", winner);
	m.data.insert("finisher", finisher);
	m_messageHandler->sendMessage(m);
}

void spectator::removePlayer(const QString &nick)
{
	message m;
	m.type = "removePlayer";
	m.data.insert("nick", nick);
	m_messageHandler->sendMessage(m);
}

void spectator::rejectVotedTiles(const QString &nick)
{
	Q_UNUSED(nick);
	kDebug() << "spectator " << m_nick << " sending reject tiles" << endl;
	m_phase = PLAYING_GAME;
	m_privTiles.clear();
}

void spectator::usesExclusiveTiles(const QString &nick)
{
	Q_UNUSED(nick);
}

void spectator::passedTurn(const QString &nick)
{
	message m;
	m.type = "passedTurn";
	m.data.insert("nick", nick);
	m_messageHandler->sendMessage(m);
}

void spectator::exchangedTiles(const QString &nick, int nTiles)
{
	message m;
	m.type = "exchangedTiles";
	m.data.insert("nick", nick);
	m.data.insert("number", QString::number(nTiles));
	m_messageHandler->sendMessage(m);
}

void spectator::sendStatus(const QString &nick, int which)
{
	message m;
	m.type = "status";
	m.data.insert("nick", nick);
	m.data.insert("status", QString::number(which));
	m_messageHandler->sendMessage(m);
}

void spectator::processNewMessage(const message &m)
{
	kDebug() << "spectator::processNewMessage" << " " << m.type << " " << m.data << endl;
	switch (m_phase)
	{
		case WAITING_FOR_SPECTATOR:
			if (m.type == "wantToJoin")
			{
				m_isSpectator = QVariant( m.data.value("playerType") ).toBool();
				emit wantToJoin(this);
				m_phase = WAITING_FOR_NICK;
			}
			// else wrongResponse()
		break;
		
		case WAITING_FOR_NICK:
			if (m.type == "sendNick")
			{
				m_nick = m.data.value("nick");
				m_phase = WAITING_TO_START_GAME;
				emit gotNick(this);
			}
			// else wrongResponse()
		break;

		case WAITING_TO_START_GAME:
			if (m.type == "message") handleMessage(m);
			else if (m.type == "join")
			{
				int gameId = m.data.value("gameId").toInt();
				bool asSpectator = m.data.value("asSpectator").toInt();
				emit joinGame(this, gameId, asSpectator);
			}
			else if (m.type == "createGame")
			{
				QString name = m.data.value("name");
				int maxPlayers = m.data.value("maxPlayers").toInt();
				QString comment = m.data.value("comment");
				QString languageCode = m.data.value("languageCode");
				bool allowSpectators = m.data.value("allowSpectators").toInt();
				emit createServerGame(this, name, maxPlayers, comment, languageCode, allowSpectators);
			}
			// else wrongResponse()
		break;
		
		case PLAYING_GAME:
		
		case NICK_ALREADY_IN_USE:
			// wrongResponse()
		break;

		default:
		break;
	}
}

void spectator::handleMessage(const message &m)
{
	emit gotMessage(this, m.data.value("message"));
}

void spectator::slotSocketError(QAbstractSocket::SocketError error)
{
	switch (error)
	{
		case QAbstractSocket::RemoteHostClosedError:
			emit remotelyDisconnected(this);
		break;

		default:
			// TODO tell the user
			kDebug() << "QAbstractSocket::RemoteHostClosedError number: " << error << endl;
		break;
	}
}

#include "spectator.moc"
