/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/

#include "server.h"
#include "gameManager.h"
#include "game.h"
#include "settings.h"

#include <qstring.h>
#include <qpixmap.h>
#include <qmap.h>
#include <qfile.h>
#include <qpixmapcache.h>

#include <kdebug.h>
#include <kstandarddirs.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kpassdlg.h>

GameManager::GameManager()
{
    m_launch = new Launch( this );
    parseXML( locate( "data", "kqf/data/games.xml" ) );
}

GameManager::~GameManager()
{
    delete m_launch;
}

bool GameManager::parseXML( const QString & xmlFile )
{
    if ( !KStandardDirs::exists( xmlFile ) )
    {
        KMessageBox::error( 0, i18n( "The required XML file with supported games' definitions does not exist:\n%1" )
                            .arg( xmlFile ) );
        return false;
    }


    QDomDocument doc;
    QFile file( xmlFile );
    if ( !file.open( IO_ReadOnly ) )
    {
        KMessageBox::error( 0, i18n( "The required XML file with supported games' definitions cannot be opened for reading:\n%1" )
                            .arg( xmlFile ) );
        return false;
    }
    QString msg;
    if ( !doc.setContent( &file, &msg ) )
    {
        KMessageBox::error( 0, i18n( "The required XML file with supported games' definitions cannot be parsed:\n%1\n%2" )
                            .arg( xmlFile ).arg( msg ) );
        file.close();
        return false;
    }
    file.close();

    QDomNodeList gameList = doc.documentElement().elementsByTagName( "game" );
    for ( uint i = 0; i < gameList.count(); i++ )
    {
        QDomElement gameElem = gameList.item( i ).toElement();
        if ( !gameElem.isNull() )
        {
            m_gameList.append( Game( gameElem ) );
            //kdDebug() << "Appending new game: " << m_gameList.last().name() <<  endl;
        }
    }
    qHeapSort( m_gameList );
    return true;
}

GameList GameManager::availableGames() const
{
    //kdDebug() << "Finding available games" << endl;

    GameList retlist;

    GameList list = m_gameList;
    GameList::iterator it;

    for ( it = list.begin(); it != list.end(); ++it )
    {
        //kdDebug() << "Checking game: " << ( *it ).name() << endl;
        QString exec = KStandardDirs::findExe( ( *it ).exec() );
        if ( !exec.isNull() )
        {
            //kdDebug() << "Found game: " << exec << endl;
            retlist.append( *it );
        }
    }

    qHeapSort( retlist );
    return retlist;
}

GameList GameManager::allGames() const
{
    return m_gameList;
}

QString GameManager::serverOption( const QString & gameID ) const
{
    Game * game = gameByID( gameID );
    if ( game )
        return game->serverOption();

    return QString::null;
}

QString GameManager::masterServerOption( const QString & gameID ) const
{
    Game * game = gameByID( gameID );
    if ( game )
        return game->masterServerOption();

    return QString::null;
}

QString GameManager::masterServer( const QString & gameID ) const
{
    Game * game = gameByID( gameID );
    if ( game )
        return game->masterServer();

    return QString::null;
}

QPixmap GameManager::gameIcon( const QString & gameID ) const
{
    Game * game = gameByID( gameID );
    if ( game )
        return game->gameIcon();

    return QPixmap();
}

QPixmap GameManager::lockedIcon()
{
    QPixmap pm;
    if ( !QPixmapCache::find( "locked", pm ) )
    {
        pm.load( locate( "data", "kqf/pics/status/locked.xpm" ), "XPM" );
        QPixmapCache::insert( "locked", pm );
    }

    return pm;
}

QPixmap GameManager::pbIcon()
{
    QPixmap pm;
    if ( !QPixmapCache::find( "punkbuster", pm ) )
    {
        pm.load( locate( "data", "kqf/pics/status/punkbuster.xpm" ), "XPM" );
        QPixmapCache::insert( "punkbuster", pm );
    }

    return pm;
}

QPixmap GameManager::pbLockedIcon()
{
    QPixmap pm;
    if ( !QPixmapCache::find( "locked_punkbuster", pm ) )
    {
        pm.load( locate( "data", "kqf/pics/status/locked_punkbuster.xpm" ), "XPM" );
        QPixmapCache::insert( "locked_punkbuster", pm );
    }

    return pm;
}

QPixmap GameManager::manIcon( const QString & color )
{
    QPixmap pm;
    QString name = "man-" + color;
    if ( !QPixmapCache::find( name, pm ) )
    {
        pm.load( locate( "data", "kqf/pics/status/" + name + ".xpm" ), "XPM" );
        QPixmapCache::insert( name, pm );
    }

    return pm;
}

QPixmap GameManager::flagIcon( const QString & country )
{
    QPixmap pm;
    if ( !QPixmapCache::find( country, pm ) )
    {
        pm.load( locate( "data", "kqf/pics/flags/" + country + ".png" ), "PNG" );
        QPixmapCache::insert( country, pm );
    }

    return pm;
}

Game * GameManager::gameByID( const QString & gameID ) const
{
    GameList::const_iterator it = m_gameList.begin();
    GameList::const_iterator itend = m_gameList.end();

    for ( ; it != itend; ++it )
    {
        if ( ( *it ).id() == gameID )
            return const_cast<Game *>( &( *it ) );
    }
    return 0;
}

Game * GameManager::baseGame( const QString & serverType ) const
{
    GameList::const_iterator it = m_gameList.begin();
    GameList::const_iterator itend = m_gameList.end();

    for ( ; it != itend; ++it )
    {
        if ( ( *it ).baseGame() == serverType )
            return const_cast<Game *>( &( *it ) );
    }
    return 0;
}

QString GameManager::passwordRule( const QString & gameID )
{
    if ( gameID == "DM3S" )           // Doom3
        return "si_usepass";
    else if ( gameID == "QWS" ||      // QuakeWorld
              gameID == "SNS" ||      // Sin
              gameID == "Q2S:KP" ||   // KingPin
              gameID == "SFS" ||      // Soldier of Fortune
              gameID == "Q2S:HR" ||   // Heretic 2
              gameID == "Q2S" ||      // Quake 2
              gameID == "HWS" )       // HexenWorld
        return "needpass";
    else if ( gameID == "HLS" || gameID == "A2S" )       // Half Life (2)
        return "sv_password";
    else if ( gameID == "BF1942" )    // Battlefield 1942
        return "_password";

    Game * game = gameByID( gameID );
    QString base;
    if ( game )
        base = game->getNode( "base" );

    if ( base == "Q3_SERVER" || gameID == "Q3S" )   // Quake 3 based servers (Q3, ET, Wolf)
        return "g_needpass";
    else if ( base == "UN_SERVER" ||  // Unreal based servers (AAO, Rune, Postal, etc.)
              gameID == "UNS" ||      // Unreal / UT
              gameID == "SMS" ||      // Serious Sam
              gameID == "T2S" )       // Tribes 2
        return "password";

    return QString::null;
}

QString GameManager::pbRule( const QString & gameID )
{
    if ( gameID == "HLS" || gameID == "A2S" )      // Half Life (2)
        return "secure";

    Game * game = gameByID( gameID );
    QString base;
    if ( game )
        base = game->getNode( "base" );

    if ( base == "Q3_SERVER" || gameID == "Q3S" )   // Quake 3 based servers (Q3, ET, Wolf)
        return "sv_punkbuster";

    return QString::null;
}

QStringList GameManager::gameNameRule( const QString & gameID )
{
    Game * game = gameByID( gameID );
    QString base;
    if ( game )
        base = game->getNode( "base" );

    if ( gameID == "UNS" ||          // Unreal
         gameID == "SMS" ||          // Serious Sam
         gameID == "D3G" ||          // Descent 3
         gameID == "SAS" ||          // Savage
         base == "UN_SERVER" )       // Unreal family
        return "gametype";
    else if ( gameID == "BF1942" )   // Battlefield 1942
        return "Game Id";
    else if ( gameID == "DM3S" || gameID == "NETP" )     // Doom 3, Netpanzer
        return "gamename";
    else if ( gameID == "SNS" ||      // Sin
              gameID == "Q2S:KP" ||   // KingPin
              gameID == "SFS" ||      // Soldier of Fortune
              gameID == "Q2S:HR" ||   // Heretic 2
              gameID == "Q2S" ||      // Quake 2
              gameID == "HLS" || gameID == "A2S" )       // Half Life (2)
        return "gamedir";
    else if ( gameID == "T2S" )       // Tribes 2
        return "game";
    else if ( gameID == "QWS" ||      // Quake World
              gameID == "HWS" )       // Hexen World
        return "*gamedir";
    else if ( base == "Q3_SERVER" || gameID == "Q3S" )   // Quake 3 based servers (Q3, ET, Wolf)
        return QStringList() << "fs_game" << "game" << "gamename";

    return QString::null;
}

QStringList GameManager::gameTypeRule( const QString & gameID )
{
    Game * game = gameByID( gameID );
    QString base;
    if ( game )
        base = game->getNode( "base" );

    if ( gameID == "UNS" ||          // Unreal
         gameID == "SMS" ||          // Serious Sam
         gameID == "D3G" ||          // Descent 3
         gameID == "SAS" ||          // Savage
         gameID == "NETP" ||         // Net Panzer
         base == "UN_SERVER" )       // Unreal family
        return "gamestyle";
    else if ( gameID == "BF1942" )   // Battlefield 1942
        return "gametype";
    else if ( gameID == "DM3S" )      // Doom 3
        return "si_gameType";
    else if ( gameID == "SNS" ||      // Sin
              gameID == "Q2S:KP" ||   // KingPin
              gameID == "SFS" ||      // Soldier of Fortune
              gameID == "Q2S:HR" ||   // Heretic 2
              gameID == "Q2S" ||      // Quake 2
              gameID == "HLS" || gameID == "A2S" )       // Half Life (2)
        return "gamename";
    else if ( gameID == "T2S" )       // Tribes 2
        return "mission";
    else if ( base == "Q3_SERVER" || gameID == "Q3S" )   // Quake 3 based servers (Q3, ET, Wolf)
        return QStringList() << "g_gametype" << "g_gametypestring";

    return QString::null;
}

QString GameManager::execClientType( const QString & gameID )
{
    Game * game = gameByID( gameID );
    if ( !game )
        return QString::null;

    QString type = game->execClientType();
    //kdDebug() << "Direct launch type is: " << type << endl;

    if ( !type.isNull() )
        return type;

    Game * base = baseGame( game->getNode( "base" ) );
    //kdDebug() << "Base game is: " << base->id() << endl;
    if ( base )
    {
        //kdDebug() << "Indirect launch type is: " << base->execClientType() << endl;
        return base->execClientType();
    }

    return QString::null;
}

bool GameManager::execClient( Server * server )
{
    // exec type
    QString type = execClientType( server->gameID() );
    //kdDebug() << "Trying to launch gametype: " << type << endl;

    Game * game = gameByID( server->gameID() );

    //kdDebug() << "We got game " << game->name() << endl;

    if ( !game )
    {
        KMessageBox::sorry( 0, i18n( "This game is not known to kqf." ) );
        return false;
    }

    // check the binary
    QString command = KStandardDirs::findExe( game->exec() );
    if ( command.isNull() )
    {
        KMessageBox::error( 0, i18n( "The game binary '%1' was not found in the system." ).arg( game->exec() ) );
        return false;
    }

    // check for password, if any
    if ( server->isLocked() )
    {
        QCString password;
        int result = KPasswordDialog::getPassword( password, i18n( "Connecting to this server requires a password" ) );
        if ( result == KDialogBase::Accepted )
            m_launch->setPassword( password );
        else
            return false;       // don't try to connect
    }
    else
        m_launch->clearPassword();   // clear the password

    if ( type == "q3_exec" ||            // Quake3 family
         type == "q1_exec_generic" ||    // Quake 1 family
         type == "q2_exec" ||            // Quake 2
         type == "q2_exec_generic" ||    // Quake 2 family
         type == "qw_exec" )             // Quake World
        return m_launch->quake_exec( server, command );
    else if ( type == "hl_exec" )        // HalfLife family
        return m_launch->quake_exec( server, command ); // not different now
    else if ( type == "ssam_exec" )      // Serious Sam
        return m_launch->quake_exec( server, command ); // not different now
    else if ( type == "ut_exec" )        // Unreal Tournament family
        return m_launch->ut_exec( server, command );
    else if ( type == "savage_exec" )    // Savage
        return m_launch->savage_exec( server, command );
    else if ( type == "netpanzer_exec" ) // Net Panzer
        return m_launch->netpanzer_exec( server, command );
    else if ( type == "bf1942_exec" )    // Battlefield 1942
        return m_launch->bf1942_exec( server, command );
    else if ( type == "descent3_exec" )  // Descent 3
        return m_launch->descent3_exec( server, command );
    else if ( type == "t2_exec" )        // Tribes 2
        return m_launch->t2_exec( server, command );
    else
        kdWarning() << "Unknown game launch type" << endl;

    return false;
}

bool GameManager::hasFlag( const QString & gameID, const QString & name ) const
{
    Game * game = gameByID( gameID );
    if ( !game )
        return false;

    return ( game->getNode( "flags" ).contains( name ) > 0 );
}

QString GameManager::gameName( const QString & gameID ) const
{
    Game * game = gameByID( gameID );
    if ( game )
        return game->name();

    return QString::null;
}

GameComboBox::GameComboBox( GameManager * manager, QWidget * parent, const char * name ):
    QComboBox( false, parent, name)
{
    GameList list;
    if ( Settings::displayAllGames() )
        list = manager->allGames();
    else
        list = manager->availableGames();

    QValueListIterator<Game> it;
    int i = 0;
    for ( it = list.begin(); it != list.end(); ++it )
    {
        Game game = ( *it );
        insertItem( game.gameIcon(), game.name() );
        m_gameMap.insert( i, game.id() );
        i++;
    }
}

QString GameComboBox::gameID() const
{
    return m_gameMap[currentItem()];
}

void GameComboBox::setGameID( const QString & gameID )
{
    for ( QMapIterator<int, QString> it = m_gameMap.begin(); it != m_gameMap.end(); ++it )
    {
        if ( ( *it ) == gameID )
        {
            setCurrentItem( it.key() );
            break;
        }
    }
}

#include "gameManager.moc"
