/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/


#include "kqfview.h"
#include "settings.h"
#include "player.h"
#include "game.h"
#include "dlgNewServer.h"
#include "propDialog.h"

#include <qlabel.h>
#include <qtoolbutton.h>
#include <qapplication.h>
#include <qsplitter.h>
#include <qpixmap.h>
#include <qmap.h>
#include <qclipboard.h>
#include <qfile.h>
#include <qdatastream.h>

#include <klocale.h>
#include <klistview.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <kiconloader.h>
#include <klistviewsearchline.h>
#include <kapplication.h>
#include <kstandarddirs.h>
#include <kglobal.h>
#include <kstdguiitem.h>
#include <kcursor.h>
#include <dcopref.h>
#include <kmacroexpander.h>

kqfView::kqfView( QWidget *parent )
    : kqfview_base( parent ), m_manager( 0 ), qstat( 0 ),
      m_lpid( 0 ), m_spid( 0 ), m_ppid( 0 ),
      m_updating( false ), m_currentGameID( FAVID )
{
    m_manager = new GameManager();
    qstat = new QStatIface( m_manager );
#ifdef HAVE_LIBGEOIP
    m_gi = GeoIP_new( GEOIP_STANDARD );
#endif

    // setup Source listview
    lvSource->setColumnWidthMode( 0, QListView::Manual );
    lvSource->hideColumn( 0 );
    lvSource->setSorting( 1 );

    // setup Servers listview
    lvServers->setColumnWidthMode( 0, QListView::Manual );
    lvServers->setColumnWidth( 0, 50 ); // prevent the name column from ever growing

    btnClearSearch->setIconSet( QApplication::reverseLayout() ?
                                SmallIconSet( "clear_left" ) : SmallIconSet( "locationbar_erase" ) );
    lvSearchLine->setListView( lvServers );

    //setup Players listview
    lvPlayers->setSorting( 1, false );

    slotLoadSettings();
    loadConfig();

    findGames();                // fill the left listview

    connect( lvSource, SIGNAL( selectionChanged( QListViewItem * ) ),
             this, SLOT( slotSourceGameChanged( QListViewItem * ) ) );

    connect( lvServers, SIGNAL( clicked( QListViewItem * ) ),
             this, SLOT( slotServerClicked( QListViewItem * ) ) );
    connect( lvServers, SIGNAL( returnPressed( QListViewItem * ) ),
             this, SLOT( slotServerClicked( QListViewItem * ) ) );
    connect( lvServers, SIGNAL( doubleClicked( QListViewItem * ) ),
             this, SLOT( slotConnect( QListViewItem * ) ) );

    connect( m_manager, SIGNAL( gameFinished() ),
             this, SLOT( setBack() ) );

    loadServerList( m_currentGameID ); // loadConfig() sets m_currentGameID
    enableServerActions();
}

kqfView::~kqfView()
{
    delete m_manager;
    delete qstat;
#ifdef HAVE_LIBGEOIP
    GeoIP_delete( m_gi );
#endif
}

void kqfView::loadConfig()
{
    KConfig * config = kapp->config();
    config->setGroup( "Splitters" );
    splitter1->setSizes( config->readIntListEntry( "Splitter1" ) );
    splitter2->setSizes( config->readIntListEntry( "Splitter2" ) );
    splitter3->setSizes( config->readIntListEntry( "Splitter3" ) );

    config->setGroup( "Game" );
    m_currentGameID = config->readEntry( "CurrentGame", FAVID );

    lvServers->restoreLayout( config, "ServerList" );
    lvPlayers->restoreLayout( config, "PlayerList" );
    lvRules->restoreLayout( config, "RuleList" );
}

void kqfView::saveConfig()
{
    KConfig * config = kapp->config();
    config->setGroup( "Splitters" );
    config->writeEntry( "Splitter1", splitter1->sizes() );
    config->writeEntry( "Splitter2", splitter2->sizes() );
    config->writeEntry( "Splitter3", splitter3->sizes() );

    config->setGroup( "Game" );
    config->writeEntry( "CurrentGame", m_currentGameID );

    lvServers->saveLayout( config, "ServerList" );
    lvPlayers->saveLayout( config, "PlayerList" );
    lvRules->saveLayout( config, "RuleList" );
}

void kqfView::slotLoadSettings()
{
    /*sillyLabel->setPaletteBackgroundColor( Settings::col_background() );
    sillyLabel->setPaletteForegroundColor( Settings::col_foreground() );

    // i18n : internationalization
    sillyLabel->setText( i18n( "This project is %1 days old" ).arg( Settings::val_time() ) );
    emit signalChangeStatusbar( i18n( "Settings changed" ) );*/
}

void kqfView::slotAddNewServer()
{
    NewServerDlg dlg( m_manager, this, "new_server_dlg" );
    dlg.setGameID( m_currentGameID );
    if ( dlg.exec() == KDialogBase::Accepted )
    {
        QString address = dlg.address();
        if ( address.endsWith( ":" ) ) // strip the colon if the port hasn't been entered
            address.remove( ":" );
        kdDebug() << "New server address: " << address << endl;
        kdDebug() << "New server game: " << dlg.gameID() << endl;
        Favorite fav( dlg.gameID(), "", address, "" ); // FIXME no comment input yet
        m_favoriteServers.append( fav );
        saveServerList( FAVID );
        if ( m_currentGameID == FAVID ) // reload in case we are inside the favorites category
        {
            loadServerList( FAVID );

            QListViewItem * server = lvServers->findItem( address, 1 );
            slotServerClicked( server );
        }
    }
}

void kqfView::slotAddToFavorites()
{
    Server * server = static_cast<Server *>( lvServers->selectedItem() );

    if ( server )
    {
        Favorite fav = Favorite( server->gameID(), server->name(), server->address(), server->comment() );

        if ( m_favoriteServers.contains( fav ) > 0 )
        {
            KMessageBox::information( this, i18n( "This server is already included in favorites." ) );
            return;
        }
        else
        {
            //kdDebug() << "Adding favorite: " << fav.name << endl;
            m_favoriteServers.append( fav );
            saveServerList( FAVID );
        }
    }
    else
        KMessageBox::information( this, i18n( "You have to select a server first." ) );
}

void kqfView::slotRemoveFromFavorites()
{
    Server * server = static_cast<Server *>( lvServers->selectedItem() );

    if ( !server )
    {
        KMessageBox::information( this, i18n( "You have to select a server first." ) );
        return;
    }

    if ( KMessageBox::warningContinueCancel( this, i18n( "Do you really want to remove this favorite server?\n%1" )
                                             .arg( server->strippedName() ), i18n( "Remove Favorite" ),
                                             KStdGuiItem::del() ) != KMessageBox::Continue )
        return;

    Favorite fav = Favorite( server->gameID(), server->name(), server->address(), server->comment() );
    m_favoriteServers.remove( fav );
    saveServerList( FAVID );
    delete server;
}

void kqfView::slotCopy()
{
    Server * server = static_cast<Server *>( lvServers->selectedItem() );
    if ( server )
    {
        kapp->clipboard()->setText( server->address(), QClipboard::Clipboard );
    }
}

void kqfView::slotFindPlayer()
{
    //
}

void kqfView::slotProperties()
{
    Server * server = static_cast<Server *>( lvServers->selectedItem() );
    if ( server )
    {
        PropDialog dlg( m_manager, server,
#ifdef HAVE_LIBGEOIP
                        m_gi,
#endif
                        this, "prop_dialog", true );
        dlg.exec();
    }
}

void kqfView::slotRefreshView()
{
    //
}

void kqfView::slotRefreshSelected()
{
    slotServerClicked( lvServers->selectedItem() );
}

void kqfView::slotUpdateView()
{
    QListViewItem * item = lvSource->selectedItem();
    if ( item )
    {
        updateServerList( item->text( 0 ) );
    }
}

void kqfView::slotConnect()
{
    slotConnect( lvServers->currentItem() );
}

void kqfView::slotConnect( QListViewItem * item )
{
    Server * server = static_cast<Server *>( item );
    if ( server )
    {
        if ( m_manager->execClient( server ) )
            setAway( server );
    }
    else
        KMessageBox::information( this, i18n( "You have to first select a server to connect to." ) );
}

void kqfView::slotDNSLookup()
{
    //
}

void kqfView::findGames()
{
    //kdDebug() << "Finding games" << endl;
    lvSource->clear();

    GameList list;
    if ( Settings::displayAllGames() )
        list = m_manager->allGames();
    else
        list = m_manager->availableGames();

    GameList::iterator it;

    //kdDebug() << "Gamelist has " << list.count() << " items" << endl;

    // add the root favorites item
    KListViewItem * fav = new KListViewItem( lvSource, FAVID, i18n( "Favorites" ) );
    fav->setPixmap( 1, SmallIcon( "bookmark" ) ); // KDE 4 icon name: bookmarks
    if ( FAVID == m_currentGameID )
        lvSource->setSelected( fav, true );

    for ( it = list.begin(); it != list.end(); ++it )
    {
        //kdDebug() << "Adding game source: " << ( *it ).name() << endl;
        QString gameID = ( *it ).id();
        KListViewItem * tmp = new KListViewItem( lvSource, gameID, ( *it ).name() );
        tmp->setPixmap( 1, m_manager->gameIcon( gameID ) );
        if ( gameID == m_currentGameID )
            lvSource->setSelected( tmp, true );
    }
}

void kqfView::slotSourceGameChanged( QListViewItem * item )
{
    if ( item )
    {
        if ( m_currentGameID != FAVID ) // favorites are saved immediately
        {
            saveServerList( m_currentGameID ); // save the old server list
        }
        m_currentGameID = item->text( 0 ); // change the game
        slotStopUpdating();     // stop the previous update job
        if ( !loadServerList( m_currentGameID ) ) // try to load from cache
            updateServerList( m_currentGameID ); // otherwise do a full update
        lvSearchLine->updateSearch();
    }
    enableServerActions();
}

void kqfView::updateServerList( const QString & gameID )
{
    if ( gameID == FAVID )      // no update for Favorites
        return;

    lvServers->clear();
    lvPlayers->clear();
    lvRules->clear();

    kdDebug() << "Updating server list for game ID: " << gameID << endl;

    m_updating = true;
    emit signalChangeStatusbar( i18n( "Updating server list from the master..." ) );
    emit signalEnableUpdateActions( false );
    m_lpid = qstat->getServerList( gameID );
    connect( qstat, SIGNAL( result( pid_t, const QDomElement & ) ),
             this, SLOT( slotGotServerList( pid_t, const QDomElement & ) ) );
}

void kqfView::slotGotServerList( pid_t pid, const QDomElement & elem )
{
    if ( m_lpid == pid )
    {
        kdDebug() << "Got server list: " << pid << endl;

        disconnect( qstat, 0, this, SLOT( slotGotServerList( pid_t, const QDomElement & ) ) );

        if ( elem.isNull() )
        {
            KMessageBox::detailedSorry( this, i18n( "Obtained an empty response from the master server." ),
                                        i18n( "The network or the master server may be down." ) );
            slotStopUpdating();
            return;
        }

        QDomNodeList serverList = elem.elementsByTagName( "server" );
        for ( uint i = 0; i < serverList.count(); i++ )
        {
            QDomElement selem = serverList.item( i ).toElement();

            if ( i == 0 && selem.hasAttribute( "servers" ) )       // first item is a summary
            {
                emit signalChangeStatusbar( i18n( "Obtained 1 server", "Obtained %n servers",
                                                  selem.attribute( "servers" ).toInt() ) );
                continue;
            }

            if ( !selem.isNull() && selem.hasChildNodes()
                 && ( selem.attribute( "status", "DOWN" ) == "UP" ) ) // some basic checks
                addServerNode( selem );
        }

        m_updating = false;
        emit signalEnableUpdateActions( true );
        slotServerClicked( lvServers->firstChild() );  // fetch the player list and update the favorites actions
    }
}

void kqfView::addServerNode( const QDomElement & elem )
{
    QDomNodeList childList = elem.childNodes();
    Server * server = new Server( lvServers );
    server->setGameManager( m_manager );
    server->setGameID( m_currentGameID );
    server->setAddress( elem.attribute( "address" ) );

    for ( uint i = 0; i < childList.count(); i++ )
    {
        QDomElement celem = childList.item( i ).toElement();
        if ( !celem.isNull() )
        {
            QString tagName = celem.tagName();
            QString value = celem.text();

            if ( tagName == "name" )
                server->setName( value );
            else if ( tagName == "map" )
                server->setMapName( value );
            else if ( tagName == "numplayers" )
                server->setNumPlayers( value.toInt() );
            else if ( tagName == "maxplayers" )
                server->setMaxPlayers( value.toInt() );
            else if ( tagName == "ping" )
                server->setPing( value.toInt() );
        }
    }

    if ( !elem.namedItem( "rules" ).isNull() )
    {
        server->setRules( elem.namedItem( "rules" ).toElement() );
    }

    // game icon
    server->setPixmap( 0, m_manager->gameIcon( server->gameID() ) );

    // priv status icon
    if ( server->isLocked() && server->hasPunkbuster() )
        server->setPixmap( 3, GameManager::pbLockedIcon() );
    else if ( server->isLocked() )
        server->setPixmap( 3, GameManager::lockedIcon() );
    else if ( server->hasPunkbuster() )
        server->setPixmap( 3, GameManager::pbIcon() );

    // slots icon
    if ( server->numPlayers() == server->maxPlayers() )
        server->setPixmap( 4, GameManager::manIcon( "red" ) );
    else if ( server->freeSlots() <= 0 )
        server->setPixmap( 4, GameManager::manIcon( "yellow" ) );
    else if ( server->numPlayers() > 0 )
        server->setPixmap( 4, GameManager::manIcon( "black" ) );

#ifdef HAVE_LIBGEOIP
    //kdDebug() << "Server " << server->address() << " is from: " << server->countryCode( m_gi ) << endl;
    server->setPixmap( 1, GameManager::flagIcon( server->countryCode( m_gi ) ) );
#endif
}

void kqfView::slotStopUpdating()
{
    if ( m_updating )
    {
        emit signalChangeStatusbar( i18n( "Aborted" ) );
        m_updating = false;
        emit signalEnableUpdateActions( true );
    }
}

void kqfView::slotServerClicked( QListViewItem * item )
{
    Server * server = static_cast<Server *>( item );

    if ( server )
    {
        refreshServerStats( server );
    }

    lvServers->triggerUpdate();

    enableServerActions();
}

void kqfView::addPlayerNode( const QDomElement & elem )
{
     QDomNodeList childList = elem.childNodes();

     if ( childList.length() == 0 )
         return;

     Player * player = new Player( lvPlayers );

     for ( uint i = 0; i < childList.count(); i++ )
     {
         QDomElement celem = childList.item( i ).toElement();
         if ( !celem.isNull() )
         {
             QString tagName = celem.tagName();
             QString value = celem.text();

             if ( tagName == "name" )
                 player->setName( value );
             else if ( tagName == "score" )
                 player->setFrags( value.toInt() );
             else if ( tagName == "ping" )
                 player->setPing( value.toInt() );
         }
     }
}

void kqfView::fillRules( const Server::RuleMap & rules )
{
    lvRules->clear();

    QMap<QString, QString> map = rules;
    QMap<QString, QString>::Iterator it;
    for ( it = map.begin(); it != map.end(); ++it )
    {
        new KListViewItem( lvRules, it.key(), it.data() );
    }
}

bool kqfView::saveServerList( const QString & gameID )
{
    QString filename = KGlobal::dirs()->saveLocation( "data", "kqf/" ) + gameID + ".dat";

    kdDebug() << "Saving server list to " << filename << endl;

    QFile file( filename );
    if ( !file.open( IO_WriteOnly ) )
        return false;

    QDataStream stream( &file );
    stream << ( Q_UINT32 ) 0xdeadbeef; // magic
    stream << ( Q_INT32 ) 1;    // version

    // write the data
    if ( gameID == FAVID )
    {
        stream << m_favoriteServers;
    }
    else
    {
        QListViewItemIterator it( lvServers );
        while ( it.current() )
        {
            Server * server = static_cast<Server *>( it.current() );
            if ( server )
                stream << *server;
            ++it;
        }
    }

    file.close();

    return true;
}

void kqfView::saveServerList()
{
    saveServerList( m_currentGameID );
}

bool kqfView::loadServerList( const QString & gameID )
{
    emit signalChangeStatusbar( i18n( "Loading servers from cache" ) );
    QApplication::setOverrideCursor( KCursor::waitCursor() );

    lvServers->clear();
    lvPlayers->clear();
    lvRules->clear();

    QString filename = KGlobal::dirs()->saveLocation( "data", "kqf/" ) + gameID + ".dat";
    kdDebug() << "Loading server list from " << filename << endl;

    if ( !QFile::exists( filename ) )
    {
        QApplication::restoreOverrideCursor();
        return false;
    }

    QFile file( filename );
    if ( !file.open( IO_ReadOnly ) )
    {
        QApplication::restoreOverrideCursor();
        return false;
    }

    QDataStream stream( &file );
    Q_UINT32 magic;
    Q_INT32 version;

    stream >> magic;
    stream >> version;
    if ( ( magic != 0xdeadbeef ) || ( version != 1 ) )
    {
        file.close();
        kdDebug() << "Stream magic: " << magic << endl;
        kdDebug() << "Stream ver: " << version << endl;
        kdDebug() << "Stream doesn't match" << endl;
        QApplication::restoreOverrideCursor();
        return false;
    }

    int i = 0;
    // load the data
    if ( gameID == FAVID )      // load and build the favorites
    {
        m_favoriteServers.clear();
        stream >> m_favoriteServers;
        QValueListIterator<Favorite> it;
        for ( it = m_favoriteServers.begin(); it != m_favoriteServers.end(); ++it )
        {
            Server * server = new Server( lvServers );
            server->setGameManager( m_manager );
            server->setGameID( ( *it ).gameID );
            server->setName( ( *it ).name );
            server->setAddress( ( *it ).address );
            server->setComment( ( *it ).comment );
            server->setPixmap( 0, m_manager->gameIcon( server->gameID() ) );
#ifdef HAVE_LIBGEOIP
            //kdDebug() << "Server " << server->address() << " is from: " << server->countryCode( m_gi ) << endl;
            server->setPixmap( 1, GameManager::flagIcon( server->countryCode( m_gi ) ) );
#endif

            i++;
        }
    }
    else                        // load and build a normal server list
    {
        while ( !stream.atEnd() )
        {
            Server * server = new Server( lvServers );
            server->setGameManager( m_manager );
            stream >> *server;

            server->setPixmap( 0, m_manager->gameIcon( server->gameID() ) );
#ifdef HAVE_LIBGEOIP
            //kdDebug() << "Server " << server->address() << " is from: " << server->countryCode( m_gi ) << endl;
            server->setPixmap( 1, GameManager::flagIcon( server->countryCode( m_gi ) ) );
#endif

            i++;
        }
    }

    file.close();

    emit signalChangeStatusbar( i18n( "Loaded 1 server from cache", "Loaded %n servers from cache", i ) );

    QApplication::restoreOverrideCursor();

    return true;
}

void kqfView::enableServerActions()
{
    bool add = false;
    bool remove = false;

    if ( lvServers->selectedItem() )
    {
        if ( m_currentGameID == FAVID )
        {
            add = false;
            remove = true;
        }
        else
        {
            add = true;
            remove = false;
        }
    }

    emit signalEnableServerActions( add, remove );
}

QPixmap kqfView::gameIcon( const QString & gameID )
{
    return m_manager->gameIcon( gameID );
}

void kqfView::refreshServerStats( Server * server )
{
    QString address = server->address();
    //kdDebug() << "Refreshing server " << address << endl;
    m_updating = true;
    emit signalChangeStatusbar( i18n( "Refreshing server %1" ).arg( address ) );
    emit signalEnableUpdateActions( false );
    m_spid = qstat->getServerInfo( address, server->gameID() );
    connect( qstat, SIGNAL( result( pid_t, const QDomElement & ) ),
             this, SLOT( slotGotServerStats( pid_t, const QDomElement & ) ) );
}

void kqfView::slotGotServerStats( pid_t pid, const QDomElement & elem )
{
    if ( m_spid == pid )
    {
        kdDebug() << "Got server stats: " << pid << endl;

        disconnect( qstat, 0, this, SLOT( slotGotServerStats( pid_t, const QDomElement & ) ) );

        QDomElement selem = elem.namedItem( "server" ).toElement(); // single <server> node

        if ( selem.isNull() )
        {
            KMessageBox::detailedSorry( this, i18n( "Obtained an empty response from the server." ),
                                        i18n( "The server may be down or unresponsive." ) );
            slotStopUpdating();
            return;
        }

        Server * server = static_cast<Server *>( lvServers->findItem( selem.attribute( "address" ), 1 ) );
        //Server * server = static_cast<Server *>( lvServers->selectedItem() );

        //kdDebug() << "Server should have address: " << selem.attribute( "address" ) << endl;

        if ( !server )
        {
            kdDebug() << "Server not found in listview" << endl;
            m_updating = false;
            emit signalEnableUpdateActions( true );
            return;
        }

        // TODO handle server timeout or being down

        QDomNodeList childList = selem.childNodes();
        for ( uint i = 0; i < childList.count(); i++ )
        {
            QDomElement celem = childList.item( i ).toElement();
            if ( !celem.isNull() )
            {
                QString tagName = celem.tagName();
                QString value = celem.text();

                //kdDebug() << "Got server pair: " << tagName << " " << value << endl;

                if ( tagName == "name" )
                {
                    server->setName( value );
                    updateFavName( server->gameID(), server->address(), value );
                }
                else if ( tagName == "map" )
                    server->setMapName( value );
                else if ( tagName == "numplayers" )
                    server->setNumPlayers( value.toInt() );
                else if ( tagName == "maxplayers" )
                    server->setMaxPlayers( value.toInt() );
                else if ( tagName == "ping" )
                    server->setPing( value.toInt() );
            }
        }

        // server rules
        if ( !selem.namedItem( "rules" ).isNull() )
        {
            server->setRules( selem.namedItem( "rules" ).toElement() );
            fillRules( server->rules() );
        }
        else
            lvRules->clear();

        // priv status pixmaps
        if ( server->isLocked() && server->hasPunkbuster() )
            server->setPixmap( 3, GameManager::pbLockedIcon() );
        else if ( server->isLocked() )
            server->setPixmap( 3, GameManager::lockedIcon() );
        else if ( server->hasPunkbuster() )
            server->setPixmap( 3, GameManager::pbIcon() );
        else
            server->setPixmap( 3, QPixmap() );

        // player count pixmaps
        if ( server->numPlayers() == server->maxPlayers() )
            server->setPixmap( 4, GameManager::manIcon( "red" ) );
        else if ( server->freeSlots() <= 0 )
            server->setPixmap( 4, GameManager::manIcon( "yellow" ) );
        else if ( server->numPlayers() > 0 )
            server->setPixmap( 4, GameManager::manIcon( "black" ) );

        // players
        if ( !selem.namedItem( "players" ).isNull() )
        {
            lvPlayers->clear();

            QDomElement playersElem = selem.namedItem( "players" ).toElement();

            QDomNodeList playerList = playersElem.elementsByTagName( "player" );
            emit signalChangeStatusbar( i18n( "Server has 1 player", "Server has %n players", playerList.count() ) );

            for ( uint i = 0; i < playerList.count(); i++ )
            {
                QDomElement pElem = playerList.item( i ).toElement();

                if ( !pElem.isNull() && pElem.hasChildNodes() ) // some basic checks
                    addPlayerNode( pElem );
            }
        }

        m_updating = false;
        emit signalEnableUpdateActions( true );
    }
}

void kqfView::updateFavName( const QString & gameID, const QString & address, const QString & name )
{
    QValueListIterator<Favorite> it;
    for ( it = m_favoriteServers.begin(); it != m_favoriteServers.end(); ++it )
    {
        if ( ( *it ).gameID == gameID && ( *it ).address == address )
            ( *it ).name = name;
    }
}

void kqfView::setAway( Server * server )
{
    if ( Settings::changePresence() )
    {
        QString msg = Settings::presenceMessage();

        QMap<QString, QString> map;
        map.insert( "game",  m_manager->gameName( server->gameID() ) ); // game name
        map.insert( "address", server->address() ); // server address
        map.insert( "server", server->name() );     // server name
        map.insert( "mod", server->gameName() );    // game mod name
        map.insert( "gametype", server->gameType() ); // game type

        msg = KMacroExpander::expandMacros( msg, map );

        DCOPRef kopete( "kopete", "KopeteIface" );
        if ( !kopete.send( "setAway", msg ) )
            kdDebug() << "Setting \"away\" for Kopete failed" << endl;

        DCOPRef konversation( "konversation", "Konversation" );
        if ( !konversation.send( "setAway", msg ) )
            kdDebug() << "Setting \"away\" for Konversation failed" << endl;
    }
}

void kqfView::setBack()
{
    if ( Settings::changePresence() )
    {
        DCOPRef kopete( "kopete", "KopeteIface" );
        if ( !kopete.send( "setAvailable" ) )
            kdDebug() << "Setting \"back\" for Kopete failed" << endl;

        DCOPRef konversation( "konversation", "Konversation" );
        if ( !konversation.send( "setBack" ) )
            kdDebug() << "Setting \"back\" for Konversation failed" << endl;
    }
}

#include "kqfview.moc"
