/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/

#ifndef _KQFVIEW_H_
#define _KQFVIEW_H_

#include "kqfview_base.h"
#include "server.h"
#include "gameManager.h"
#include "qstatIface.h"

#include <qwidget.h>
#include <qvaluelist.h>

#include <kconfig.h>

#include <config.h>
#ifdef HAVE_LIBGEOIP
#include <GeoIP.h>
#endif

class KURL;

#define FAVID "favorites"       // no i18n() here

/**
 * This is the main view class for kqf.  Most of the non-menu,
 * non-toolbar, and non-statusbar (e.g., non frame) GUI code should go
 * here.
 *
 * @short Main view
 * @author Lukas Tinkl <lukas@kde.org>
 */
class kqfView : public kqfview_base
{
    Q_OBJECT
public:
    /**
     * Default constructor
     */
    kqfView( QWidget *parent );

    /**
     * Destructor
     */
    virtual ~kqfView();

    /**
     * Load view configuration
     */
    void loadConfig();

    /**
     * Save view configuration
     */
    void saveConfig();

    /**
     * Utility method to return a gameIcon
     * @see GameManager::gameIcon
     */
    QPixmap gameIcon( const QString & gameID );

    /**
     * Save the currently selected game's server list
     */
    void saveServerList();

signals:
    /**
     * Use this signal to change the content of the statusbar
     */
    void signalChangeStatusbar( const QString & text );

    /**
     * Use this signal to change the content of the caption
     */
    void signalChangeCaption( const QString & text );

    /**
     * Emitted as a request to enable/disable actions pertaining to updating the server list
     */
    void signalEnableUpdateActions( bool enable );

    /**
     * Emitted as a request to enable/disable @p add or @p remove server to/from favorites
     */
    void signalEnableServerActions( bool add, bool remove );

public slots:
    /**
     * Popup a dialog to manually add a new server to the Favorites
     */
    void slotAddNewServer();

    /**
     * Add the currently selected server to Favorites
     */
    void slotAddToFavorites();

    /**
     * Remove the currently selected server from Favorites
     */
    void slotRemoveFromFavorites();

    /**
     * Copy the server address (IP:port) to the clipboard
     */
    void slotCopy();

    void slotFindPlayer();

    /**
     * Show the server properties dialog
     */
    void slotProperties();

    void slotRefreshView();

    /**
     * Refresh the currently selected server info
     */
    void slotRefreshSelected();

    /**
     * Rebuild the server list from the master
     */
    void slotUpdateView();

    /**
     * Connect to the currently selected server
     */
    void slotConnect();

    /**
     * Connect to the server given in @p item
     */
    void slotConnect( QListViewItem * item );

    void slotDNSLookup();

    /**
     * Stop any pending updates from the master
     */
    void slotStopUpdating();

private slots:
    /**
     * Invoked when the application settings changed
     * (thru the Configure dialog)
     */
    void slotLoadSettings();

    /**
     * Fired when the user changes the game in the left view
     * @param item the corresponding QListViewItem
     */
    void slotSourceGameChanged( QListViewItem * item );

    /**
     * Slot to handle the obtained list of servers
     */
    void slotGotServerList( pid_t pid, const QDomElement & elem );

    /**
     * Invoked when the user clicks on a server
     * @param item the corresponding listview item
     */
    void slotServerClicked( QListViewItem * item );

    /**
     * Slot to handle the obtained list of players
     */
    void slotGotServerStats( pid_t pid, const QDomElement & elem );

    /**
     * Invoked when the user joins a server,
     * changes the IM presence to away
     */
    void setAway( Server * server );

    /**
     * Invoked when the user leaves a server,
     * changes the IM presence to online/available
     */
    void setBack();

private:
    /**
     * Find supported games available in the system, fill the left view
     */
    void findGames();

    /**
     * Update the list of servers based on @p gameID
     */
    void updateServerList( const QString & gameID );

    /**
     * Updates the listview item @p server with current server attributes
     */
    void refreshServerStats( Server * server );

    /**
     * Parses <server> XML node and adds it to the servers' listview
     */
    void addServerNode( const QDomElement & elem );

    /**
     * Parses <player> XML node and adds it to the players' listview
     */
    void addPlayerNode( const QDomElement & elem );

    /**
     * Fill the rules' listview with @p rules (Name, Value pairs)
     */
    void fillRules( const Server::RuleMap & rules );

    /**
     * Save the server list into gameID.dat file
     */
    bool saveServerList( const QString & gameID );

    /**
     * Load the server list from gameID.dat file
     */
    bool loadServerList( const QString & gameID );

    /**
     * Decide if "favorites" actions should be enabled
     */
    void enableServerActions();

    /**
     * Update the @p name of a favorite server server with @p gameID and @p address
     */
    void updateFavName( const QString & gameID, const QString & address, const QString & name );

    /// favorite servers
    QValueList<Favorite> m_favoriteServers;

    GameManager * m_manager;
    QStatIface * qstat;
    pid_t m_lpid;               // server list
    pid_t m_spid;               // single server stats
    pid_t m_ppid;               // player list
#ifdef HAVE_LIBGEOIP
    GeoIP * m_gi;
#endif
    bool m_updating;            //< whether we are updating (server, player, rule list)
    QString m_currentGameID;    //< ID of the current game
};

#endif // _KQFVIEW_H_
