/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/

#include "launch.h"
#include "server.h"
#include "gameManager.h"

#include <string.h>

#include <qregexp.h>

#include <kdebug.h>
#include <kmessagebox.h>
#include <klocale.h>

Launch::Launch( GameManager * manager )
    : m_manager( manager )
{
    m_process = new KProcess( this );
    connect( m_process, SIGNAL( processExited( KProcess * ) ),
             this, SLOT( slotProcessExited( KProcess * ) ) );

    connect( m_process, SIGNAL( processExited( KProcess * ) ),
             m_manager, SIGNAL( gameFinished() ) );
}

Launch::~Launch()
{
    delete m_process;
}

void Launch::setPassword( const QCString & password )
{
    m_password = password;
}

void Launch::clearPassword()
{
    m_password = "";
}

bool Launch::quake_exec( Server * server, const QString & command )
{
    m_process->clearArguments();

    *m_process << command;           // binary

    *m_process << "+connect" << server->address(); // IP

    if ( server->hasPunkbuster() ) // PunkBuster
        *m_process << "+set" << "cl_punkbuster" << "1";

    if ( !m_password.isEmpty() )
        *m_process << "+password" << m_password;

    // TODO more params (custom config etc)

    kdDebug() << "Launching game: " << m_process->args() << endl;

    bool result = m_process->start();
    return result;
}

bool Launch::ut_exec( Server * server, const QString & command )
{
    m_process->clearArguments();

    *m_process << command;           // binary

    QString address = server->address(); // IP[:port]

    if ( m_manager->hasFlag( server->gameID(), "GAME_LAUNCH_HOSTPORT" ) )
    {
        QString hostport = server->rule( "hostport" );
        if ( !hostport.isNull() ) // the game needs its own port
        {
            address = address.replace( QRegExp( ":\\d*" ), ':' + hostport );
        }
    }

    if ( !m_password.isEmpty() )
        *m_process << QString( "%1?password=%2").arg( address ).arg( m_password );
    else
        *m_process << address;

    // TODO more params (custom config etc)

    kdDebug() << "Launching game: " << m_process->args() << endl;

    bool result = m_process->start();
    return result;
}

bool Launch::savage_exec( Server * server, const QString & command )
{
    m_process->clearArguments();

    *m_process << command;           // binary

    *m_process << "set" << "autoexec"; // settings

    if ( !m_password.isEmpty() )
        *m_process << QString( "connect %1; set cl_password %2" ).arg( server->address() )
            .arg( m_password ); // IP, password
    else
        *m_process << "connect" << server->address(); // IP

    // TODO more params (custom config, password, etc)

    kdDebug() << "Launching game: " << m_process->args() << endl;

    bool result = m_process->start();
    return result;
}

bool Launch::netpanzer_exec( Server * server, const QString & command )
{
    m_process->clearArguments();

    *m_process << command;           // binary

    *m_process << "-c" << server->address(); // IP

    kdDebug() << "Launching game: " << m_process->args() << endl;

    bool result = m_process->start();
    return result;
}

bool Launch::bf1942_exec( Server * server, const QString & command )
{
    m_process->clearArguments();

    *m_process << command;           // binary

    *m_process << "+restart" << "1"; // ???

    QString gameid = server->rule( "gameid" ); // game (mod) id
    if ( !gameid.isNull() )
    {
        *m_process << "+game" << gameid;
    }

    QString address = server->address(); // IP[:port]
    QString hostport = server->rule( "hostport" );
    if ( !hostport.isNull() ) // the game needs its own port
    {
        address = address.replace( QRegExp( ":\\d*" ), ':' + hostport );
    }

    *m_process << "+joinServer" << address;

    kdDebug() << "Launching game: " << m_process->args() << endl;

    bool result = m_process->start();
    return result;
}

bool Launch::descent3_exec( Server * server, const QString & command )
{
    m_process->clearArguments();

    *m_process << command;           // binary

    QString address = server->address(); // IP[:port]
    QString hostport = server->rule( "hostport" );
    if ( !hostport.isNull() ) // the game needs its own port
    {
        address = address.replace( QRegExp( ":\\d*" ), ':' + hostport );
    }

    *m_process << "--directip" << address;

    kdDebug() << "Launching game: " << m_process->args() << endl;

    bool result = m_process->start();
    return result;
}

bool Launch::t2_exec( Server * server, const QString & command )
{
    m_process->clearArguments();

    *m_process << command;           // binary

    // TODO more params (login)

    *m_process << "-online" << "-connect" << server->address();

    kdDebug() << "Launching game: " << m_process->args() << endl;

    bool result = m_process->start();
    return result;
}

void Launch::slotProcessExited( KProcess * proc )
{
    if ( !proc->normalExit() )
        KMessageBox::detailedError( 0, i18n( "The game exited abnormally." ),
                                    QString::fromLocal8Bit( strerror( proc->exitStatus() ) ) );
}

#include "launch.moc"
