/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/

#ifndef _SERVER_H_
#define _SERVER_H_

#include "gameManager.h"

#include <qmap.h>
#include <qdatetime.h>
#include <qdom.h>
#include <qdatastream.h>

#include <klistview.h>

#include <config.h>

#ifdef HAVE_LIBGEOIP
#include <GeoIP.h>
#endif

class QString;

/**
 * A helper class to serialize Favorite servers
 *
 * @short Favorite Server class
 */
class Favorite
{
public:
    Favorite() { }
    Favorite( const QString & id, const QString & n, const QString & a, const QString & c ):
        gameID( id ), name( n ), address( a ), comment( c ) {  }

    QString gameID;
    QString name;
    QString address;
    QString comment;
    bool operator==( const Favorite & fav );
    friend QDataStream & operator<<( QDataStream & s, const Favorite & fav );
    friend QDataStream & operator>>( QDataStream & s, Favorite & fav );
};

/**
 * Class combining data storage and a listview item
 * representing a game server
 *
 * @short Server class
 */
class Server: public KListViewItem
{
public:
    typedef QMap<QString, QString> RuleMap;

    Server( QListView * parent ):
        KListViewItem( parent ) { init(); }

    Server( QListViewItem * parent ):
        KListViewItem( parent ) { init(); }

    Server( QListView *parent, QString name ):
        KListViewItem( parent, name )
        {
            setName( name );
            init();
        }

    Server( QListViewItem *parent, QString name ):
        KListViewItem( parent, name )
        {
            setName( name );
            init();
        }

    /**
     * Common initialization (timestamps etc)
     */
    void init();

    /**
     * @return the ID of the game as identified by qstat
     * @see Game::id
     */
    QString gameID() const
        { return m_gameID; }

    /**
     * Set the ID of the game as identified by qstat
     */
    void setGameID( const QString & id )
        { m_gameID = id; }

    /**
     * @return the server name
     */
    QString name() const
        { return m_name; }

    /**
     * Set the server (visible) name
     */
    void setName( const QString & name );

    /**
     * @return the server name
     * (stripped of font markup)
     */
    QString strippedName() const
        { return m_strippedName; }

    /**
     * @return the server's address (IP:port)
     */
    QString address() const
        { return m_address; }

    /**
     * Set the server address
     */
    void setAddress( const QString & address )
        { m_address = address; }

    /**
     * @return when was the server last refreshed
     */
    QDateTime lastRefresh() const
        { return m_lastRefresh; }

    /**
     * @return the date of last answer from the server
     */
    QDateTime lastAnswer() const
        { return m_lastAnwser; }

    /**
     * @return the user's comment about the server
     */
    QString comment() const
        { return m_comment; }

    /**
     * Set the user's comment about the server
     */
    void setComment( const QString & comment )
        { m_comment = comment; }

    /**
     * @return the map being played (e.g. goldrush, battery, etc)
     */
    QString mapName() const
        { return m_mapName; }

    /**
     * Set the map being played (e.g. goldrush, battery, etc)
     */
    void setMapName( const QString & name )
        { m_mapName = name; }

    /**
     * @return the name of the game mod (etmain, etpro, ...)
     * (decoded from rules)
     */
    QString gameName() const;

    /**
     * @return the game type (SW, Campaign, etc.)
     * (decoded from rules)
     */
    QString gameType() const;

    int numPlayers() const
        { return m_players; }

    void setNumPlayers( int num )
        { m_players = num; }

    int maxPlayers() const
        { return m_maxPlayers; }

    void setMaxPlayers( int max )
        { m_maxPlayers = max; }

    /**
     * @return the number of private (reserved) player slots on the server
     */
    int privateSlots() const;

    /**
     * @return number of available (free) slots on the server
     */
    int freeSlots() const;

    /**
     * @return the server's connection ping
     */
    int ping() const
        { return m_ping; }

    /**
     * Set the server's connection ping
     */
    void setPing( int ping )
        { m_ping = ping; }

    /**
     * @return the rules of this server
     */
    RuleMap rules() const
        { return m_rules; }

    /**
     * @return value of the rule with @p name
     */
    QString rule( const QString & name ) const;

    /**
     * Set the rules based on contents of the <rules> tag
     */
    void setRules( const QDomElement & elem );

    /**
     * Copy rules from @p map
     */
    void setRules( const RuleMap & map )
        { m_rules = map; }

    /**
     * @return whether the server uses password
     */
    bool isLocked() const;

    /**
     * @return whether the server uses Punkbuster(tm)
     */
    bool hasPunkbuster() const;

    /**
     * @internal
     * Transfer the GameManager pointer
     */
    void setGameManager( GameManager * manager );

#ifdef HAVE_LIBGEOIP
    /**
     * @return the server's country of origin
     */
    QString countryCode( GeoIP * gi ) const;
#endif

    /**
     * Write the @p server definition to stream @p s
     */
    friend QDataStream & operator<<( QDataStream & s, const Server & server );

    /**
     * Read the @p server definition from stream @p s
     */
    friend QDataStream & operator>>( QDataStream & s, Server & server);

protected:
    virtual QString text( int col ) const;
    virtual int compare( QListViewItem * i, int col, bool ascending ) const;
    virtual void paintCell( QPainter * p, const QColorGroup & cg, int column, int width, int align );
    virtual int width( const QFontMetrics & fm, const QListView * lv, int c ) const;

private:
    QString decodeGametype( const QString & gameType ) const;

    QString m_gameID, m_name, m_strippedName;
    int m_ping;
    int m_players, m_maxPlayers;
    QString m_address;
    QDateTime m_lastRefresh;
    QDateTime m_lastAnwser;
    QString m_comment;
    QString m_mapName;
    RuleMap m_rules;
    GameManager * m_manager;      /// transfered pointer
};

#endif
