/*
    This file is part of KSpiral.
    Copyright 2009 Felix Lemke <hobbyblobby@arcor.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "gameEngine.h"
#include "ball.h"
#include <cmath>
#include "gameWindow.h"
#include <QGraphicsLineItem>
#include <QGraphicsScene>
#include <QPen>
#include <QTimer>
#include "svgreader.h"

KSpiral::GameEngine::GameEngine(SvgReader* path, KSpiral::GameWindow* window)
	: QObject()
	, m_path(path)
	, m_score(0)
	, m_maxBallcount(0)
	, m_ballCounter(0)
	, m_window(window)
	, m_firstBall(0)
	, m_lastBall(0)
	, m_isInChainStep(false)
	, m_timer(0)
{
}

KSpiral::GameEngine::~GameEngine()
{
	delete m_timer;
}

void KSpiral::GameEngine::startGame(int duration) {
	m_lastBall = m_firstBall = m_window->addBall(m_path->C(0));
	++m_ballCounter;
	m_lastBall->setPosition(0);
	m_lastBall->addState(KSpiral::ballInChain);
	m_lastBall->addState(KSpiral::ballIsStartball);
	m_lastBall->addState(KSpiral::ballIsEndball);
	m_ballDiameter = 2 * m_lastBall->radius();

	//start main chain animation
	const int framesPerSecond = 25;
	const qreal pointsPerSecond = m_path->length() / duration;
	m_distancePerStep = pointsPerSecond / framesPerSecond;
	m_timer = new QTimer;
	connect(m_timer, SIGNAL(timeout()), this, SLOT(chainStep()));
	m_timer->start(1000 / framesPerSecond);
}

void  KSpiral::GameEngine::cannonballStepped(KSpiral::Ball* cannonball) {
	if (!(cannonball->states() & KSpiral::ballInCollidingCheck)) {
		cannonball->addState(KSpiral::ballInCollidingCheck);
		//test for barriers
		if(!m_barrierItems.isEmpty()) {
			foreach(QPainterPath* path, m_barrierItems) {
				if(path->intersects(cannonball->mapToScene(cannonball->boundingRect()).boundingRect())) {
					cannonball->stopShoot();
					cannonball->removeState(KSpiral::ballInCollidingCheck);
					ballAtBorder(cannonball);
					return;
				}
			}
		}
		//test for balls
		foreach (QGraphicsItem* item, cannonball->collidingItems()) {
			if(!m_ignoredItems.contains(item)) {
				if(item->type() == KSpiral::Ball::BallItem) {
					cannonball->stopShoot();
					ballCollided(cannonball, (KSpiral::Ball*)item);
					cannonball->removeState(KSpiral::ballInCollidingCheck);
					return;
				}
			}
		}
		cannonball->removeState(KSpiral::ballInCollidingCheck);
	}
}

void KSpiral::GameEngine::ballAtBorder(KSpiral::Ball* cannonball) {
	delete cannonball;
	m_score = qMax(m_score - 1, 0);
}

void KSpiral::GameEngine::chainStep(qreal distance) {
	if(!m_isInChainStep) {
		m_isInChainStep = true;
		if(m_lastBall) {
			if(!m_cannonBalls.isEmpty()) {
				foreach(KSpiral::Ball* ball, m_cannonBalls) {
					arrangeStep(ball);
				}
			} else {
				if(distance == 0) distance = m_distancePerStep;
				executeChainStep(distance, m_lastBall, 1);
			}
		} else {
			gameEnds();
		}
		m_isInChainStep = false;
	}
}

#include <kdebug.h>
void KSpiral::GameEngine::executeChainStep(qreal distance, KSpiral::Ball* startBall, int direction) {
	KSpiral::Ball* tempBall = startBall;
	int index = 0;
	qreal startPosition = startBall->position() + distance;

	if(direction >= 0) {
		while(tempBall) {
			qreal nextPosition = startPosition + index * m_ballDiameter;
			if(nextPosition < 0 || nextPosition > m_path->length()) {
				if (tempBall->nextBall()) {
					tempBall = tempBall->nextBall();
					deleteBall(tempBall->prevBall());
					if(nextPosition < 0) m_ballCounter = qMax(m_ballCounter - 1, 0);
				} else {
					deleteBall(tempBall);
					break;
				}
			}
			tempBall->setPosition(nextPosition);
			tempBall->setPos(m_path->C(nextPosition));
			++index;
			if(tempBall->states() & KSpiral::ballIsStartball) {
				if(tempBall->nextBall()) tempBall->nextBall()->rewriteStates();
				tempBall->rewriteStates();
				break;
			} else {
				tempBall->rewriteStates();
				tempBall = tempBall->nextBall();
			}
		}
		if (!m_lastBall) {
			gameEnds();
			m_isInChainStep = false;
			return;
		}
	}
	tempBall = startBall;
	index = 0;
	startPosition = startBall->position() + distance;
	if(direction <= 0) {
		while(tempBall) {
			qreal nextPosition = startPosition - index * m_ballDiameter;
			if(nextPosition < 0 || nextPosition > m_path->length()) {
				if (tempBall->prevBall()) {
					tempBall = tempBall->prevBall();
					deleteBall(tempBall->nextBall());
				} else {
					deleteBall(tempBall);
					break;
				}
			}
			tempBall->setPosition(nextPosition);
			tempBall->setPos(m_path->C(nextPosition));
			++index;
			if(tempBall->states() & KSpiral::ballIsEndball) {
				if(tempBall->prevBall()) tempBall->prevBall()->rewriteStates();
				tempBall->rewriteStates();
				break;
			} else {
				tempBall->rewriteStates();
				tempBall = tempBall->prevBall();
			}
		}
		if (!m_lastBall) {
			gameEnds();
			m_isInChainStep = false;
			return;
		}
	}
	//create new last ball, if distance from start is big enough
	while (((m_maxBallcount == 0) xor (m_ballCounter < m_maxBallcount)) && m_lastBall->position() >= m_ballDiameter) {
		KSpiral::Ball* newBall = m_window->addBall(m_path->C(0));
        newBall->setPosition(0);
		++m_ballCounter;
		//arrange newBall in "chain"
		m_lastBall->setPrevBall(newBall);
		newBall->setNextBall(m_lastBall);
		newBall->addState(KSpiral::ballInChain);
        m_lastBall->rewriteStates(); //check distance from neighbours (set states like "isEndball";
        newBall->rewriteStates();
		m_lastBall = newBall;
	}
}

void KSpiral::GameEngine::deleteBall(KSpiral::Ball* ball) {
	if(ball->nextBall())
		ball->nextBall()->setPrevBall(ball->prevBall());
	else
		m_firstBall = ball->prevBall();
	if(ball->prevBall())
		ball->prevBall()->setNextBall(ball->nextBall());
	else
		m_lastBall = ball->nextBall();
	delete ball;
}

void KSpiral::GameEngine::deleteBalls(QList<KSpiral::Ball* > balls) {
	foreach(KSpiral::Ball* ball, balls)
		deleteBall(ball);
}

#include <kdebug.h>
void KSpiral::GameEngine::ballCollided(KSpiral::Ball* cannonball, KSpiral::Ball* chainBall) {
	if(chainBall->states() & KSpiral::ballInChain) {
		const qreal distanceFromNext = euklid(m_path->C(chainBall->position() + m_ballDiameter) - cannonball->pos());
		const qreal distanceFromPrev = euklid(m_path->C(chainBall->position() - m_ballDiameter) - cannonball->pos());
		if(distanceFromNext < distanceFromPrev) {
			cannonball->setNextBall(chainBall->nextBall());
			if(chainBall->nextBall()) chainBall->nextBall()->setPrevBall(cannonball);
			chainBall->setNextBall(cannonball);
			cannonball->setPrevBall(chainBall);
		} else {
			cannonball->setPrevBall(chainBall->prevBall());
			if(chainBall->prevBall()) chainBall->prevBall()->setNextBall(cannonball);
			chainBall->setPrevBall(cannonball);
			cannonball->setNextBall(chainBall);
		}
	} else {
		kDebug() << "collided with other ball";
	}
	if(cannonball->prevBall()) cannonball->prevBall()->addState(KSpiral::ballIsStartball);
	if(cannonball->nextBall()) cannonball->nextBall()->addState(KSpiral::ballIsEndball);
	m_cannonBalls.append(cannonball);
}

#include <kdebug.h>
void KSpiral::GameEngine::arrangeStep(KSpiral::Ball* cannonball) {
/*### after 5 steps the cannonball is in chain; positions have to be set ###*/
	if(cannonball->increaseArrangeStepCounter() > 5) {
		//cannonball has only one neighbour
		if(cannonball->prevBall() && !cannonball->nextBall()) {
			cannonball->setPosition(cannonball->prevBall()->position() + m_ballDiameter);
			cannonball->setPos(m_path->C(cannonball->position()));
			m_firstBall = cannonball;
		} else if(cannonball->nextBall() && !cannonball->prevBall()) {
			cannonball->setPosition(cannonball->nextBall()->position() - m_ballDiameter);
			cannonball->setPos(m_path->C(cannonball->position()));
			m_lastBall = cannonball;

		//cannonball has both neighbours
		} else if (cannonball->prevBall() && cannonball->nextBall()) {
			const qreal distFromNext = euklid(cannonball->nextBall()->pos() - cannonball->pos());
			const qreal distFromPrev = euklid(cannonball->prevBall()->pos() - cannonball->pos());
			if(distFromNext < distFromPrev) {
				cannonball->setPosition(cannonball->nextBall()->position() - m_ballDiameter);
				cannonball->setPos(m_path->C(cannonball->position()));
			} else {
			cannonball->setPosition(cannonball->prevBall()->position() + m_ballDiameter);
			cannonball->setPos(m_path->C(cannonball->position()));
			}

		//cannonball has no neighbours
		} else {
			kDebug() << "cannonball has no neighbours";
		}
		m_cannonBalls.removeAll(cannonball);
		cannonball->addState(KSpiral::ballInChain);
		removeTypeCombo(cannonball);
		return;
/*### prepare the chain dividing ###*/
	} else {
		KSpiral::Ball* referenceBall = cannonball;
		if(cannonball->prevBall() && !cannonball->nextBall())
			referenceBall = cannonball->prevBall();
		else if(!cannonball->prevBall() && cannonball->nextBall())
			referenceBall = cannonball->nextBall();
		else if(cannonball->prevBall() && cannonball->nextBall()) {
			const qreal distFromNext = euklid(cannonball->nextBall()->pos() - cannonball->pos());
			const qreal distFromPrev = euklid(cannonball->prevBall()->pos() - cannonball->pos());
			if(distFromNext < distFromPrev)
				referenceBall = cannonball->nextBall();
			else
				referenceBall = cannonball->prevBall();
		} else {
			kDebug() << "cannonball has no neighbours";
		}
/*### find insert point ###*/
		QPointF insertPoint;
		qreal checkDistance = 100 * m_ballDiameter; //some large value
		for (int position = qMax(referenceBall->position() - m_ballDiameter, 0.0); position < qMax(referenceBall->position() + m_ballDiameter, 0.0); position += m_ballDiameter / 10) {
			qreal dist = euklid(cannonball->pos() - m_path->C(position));
			if(checkDistance > dist) {
				insertPoint = m_path->C(position);
				checkDistance = dist;
			}
		}
/*### Move chain ###*/
		QPointF movingDifference((insertPoint.x() - cannonball->pos().x()) / 3, (insertPoint.y() - cannonball->pos().y()) / 3);
		cannonball->setPos(cannonball->pos() + movingDifference);
		if(cannonball->nextBall()) {
			qreal nextDiff = 0;
			const qreal nextPosition = cannonball->nextBall()->position();
			while(euklid(cannonball->pos() - m_path->C(nextPosition + nextDiff)) <= m_ballDiameter) {
				++nextDiff;
			}
			executeChainStep(nextDiff, cannonball->nextBall(), 1);
		}
		if(cannonball->prevBall()) {
			qreal prevDiff = 0;
			const qreal prevPosition = cannonball->prevBall()->position();
			while(euklid(cannonball->pos() - m_path->C(prevPosition - prevDiff)) <= m_ballDiameter) {
				++prevDiff;
			}
			executeChainStep(-prevDiff, cannonball->prevBall(), -1);
		}
		cannonball->rewriteStates();
	}
}

void KSpiral::GameEngine::removeTypeCombo(KSpiral::Ball* refBall) {
	QList<KSpiral::Ball* > comboBalls;
	const QString refType = refBall->ballType();
	int counterNext = 0;
	KSpiral::Ball* testBall = refBall;
	while(testBall->nextBall() && (testBall->nextBall()->ballType() == refType)) {
		testBall = testBall->nextBall();
		comboBalls << testBall;
		++counterNext;
	}
	if(testBall->nextBall()) testBall->nextBall()->addState(KSpiral::ballIsEndball);

	int counterPrev = 0;
	testBall = refBall;
	while(testBall->prevBall() && (testBall->prevBall()->ballType() == refType)) {
		testBall = testBall->prevBall();
		comboBalls << testBall;
		++counterPrev;
	}
	if(testBall->prevBall()) testBall->prevBall()->addState(KSpiral::ballIsStartball);

	if (counterNext + counterPrev + 1 >= 3) {
		deleteBalls(comboBalls);
		deleteBall(refBall);
		m_score += 3 + pow(2, counterNext + counterPrev - 2);
	}
	if(!m_lastBall) gameEnds();
}

void KSpiral::GameEngine::ballsLost(int count) {
	m_score = qMax(m_score - 2 * count, 0);
}

void KSpiral::GameEngine::gameEnds() {
	m_timer->stop();
	disconnect(m_timer, SIGNAL(timeout()), this, SLOT(chainStep()));
	emit gameIsOver(m_score);
}

KSpiral::Ball* KSpiral::GameEngine::ballAtIndex(int index) const {
	KSpiral::Ball* tempBall = m_firstBall;
	int i = 0;
	while (i < index && tempBall->prevBall()) {
		tempBall = tempBall->prevBall();
		++i;
	}
	return tempBall;
}

#include "gameEngine.moc"
