/*
    This file is part of KSpiral.
    Copyright 2009 Felix Lemke <hobbyblobby@arcor.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "gameWindow.h"

#include "ball.h"
#include "cannon.h"
#include <cmath>
#include "gameEngine.h"
#include "svgreader.h"

//Qt classes
#include <QBrush>
#include <QGraphicsScene>
#include <QGraphicsView>
#include <QMouseEvent>
#include <QPointer>
#include <QTimer>

//KDE classes
#include <KAction>
#include <KActionCollection>
#include <KActionMenu>
#include <KApplication>
#include <KConfig>
#include <KConfigGroup>
#include <KGameCanvas>
#include <KGameDifficulty>
#include <KGameRenderer>
#include <KGameRenderedItem>
#include <KGameRenderedObjectItem>
#include <KLocalizedString>
#include <KScoreDialog>
#include <KStandardDirs>
#include <KStandardAction>
#include <KStandardGameAction>


#include <QDebug>
//class MouseView needed to get the mouseMoveEvents and autoresize when window resize
namespace KSpiral {
	class MouseView : public QGraphicsView {
		public:
			MouseView(KSpiral::GameWindow* parent)
				: QGraphicsView() , m_fitItem(0) {m_parent = parent;}
			virtual ~MouseView() {}
			void setFitItem(QGraphicsItem* item) {
				m_fitItem = item;
				resizeEvent(0);
			}
		protected:
			void mouseMoveEvent(QMouseEvent* event) {
					m_parent->mouseMoveEvent(event);
			}
      void mousePressEvent(QMouseEvent* event) {
          m_parent->mousePressEvent(event);
      }
			virtual void resizeEvent(QResizeEvent* ) {
				if (m_fitItem) fitInView(m_fitItem, Qt::KeepAspectRatio);
			}
		private:
			KSpiral::GameWindow* m_parent;
			QGraphicsItem* m_fitItem;
	};
}

#include <kdebug.h>
KSpiral::GameWindow::GameWindow()
	: KXmlGuiWindow()
	, m_view(new KSpiral::MouseView(this))
	, m_scene(new QGraphicsScene)
  , m_background(0)
  , m_cannonTop(0)
  , m_backRect(0)
  , m_gameState(KSpiral::beforeFirstStart)
  , m_engine(0)
  , m_cannon(0)
{
	//create KGame elements
	KStandardGameAction::highscores(this, SLOT(showHighscores()), actionCollection());
	//add map names as difficulties in highscore dialog
	KStandardGameAction::gameNew(this, SLOT(startGame()), actionCollection());
	KStandardGameAction::quit(kapp, SLOT(quit()), actionCollection());

	//basic window construction
	setupGUI();
	setCentralWidget(m_view);
	m_view->setScene(m_scene);


	//load imageFile
	m_imageFile = KStandardDirs::locate("appdata", "images/egyptian.svg");
	QString configFile = KStandardDirs::locate("appdata", "images/egyptian.desktop");
  QString backgroundFile = KStandardDirs::locate("appdata", "images/Time-For-Lunch-2.jpg");

  m_view->setBackgroundBrush(QBrush(QImage(backgroundFile)));

// 	m_renderer = new QSvgRenderer(m_imageFile);
  m_renderer = new KGameRenderer("images/egyptian.desktop");
  m_renderer->setDefaultPrimaryView(m_view);

  //parse config:
	KConfig config(configFile, KConfig::SimpleConfig);
	KConfigGroup configGroup(&config, "KGameTheme");
	m_levelName = configGroup.readEntry("Name", QString());
	m_unlocalizedLevelName = configGroup.readEntryUntranslated("Name", QString());
	qDebug() << "Map from" << configGroup.readEntry("Author", QString("unknown"));
	if (!configGroup.readEntry("Description", QString()).isEmpty()) qDebug() << configGroup.readEntry("Description", QString());
	m_duration = configGroup.readEntry("Duration", QList<int >());
	m_maxBallcount = configGroup.readEntry("BallCount", QList<int >());
	m_cannonSpeed = configGroup.readEntry("CannonSpeed", 200);
	m_ballTypes = configGroup.readEntry("Balls", QStringList());
	m_barriers = configGroup.readEntry("Barriers", QStringList());

	//setup difficulties
	KGameDifficulty::init(this, this, SLOT(difficultyChanged()));
	KGameDifficulty::addStandardLevel(KGameDifficulty::VeryEasy);
	KGameDifficulty::addStandardLevel(KGameDifficulty::Easy);
	KGameDifficulty::addStandardLevel(KGameDifficulty::Medium);
	KGameDifficulty::addStandardLevel(KGameDifficulty::Hard);
	KGameDifficulty::addStandardLevel(KGameDifficulty::VeryHard);

	//insert config possibility and restore old difficulty
	KGameDifficulty::setLevel(KGameDifficulty::Easy);

	//load path
	m_path = new SvgReader;
	m_path->openPathFromSvgFile(new QFile(m_imageFile), QString("Way"));
}

KSpiral::GameWindow::~GameWindow() {
	delete m_engine;
	delete m_cannon;
	delete m_path;
	delete m_view;
	delete m_scene;
}

void KSpiral::GameWindow::resizeEvent(QResizeEvent* event) {
    if(m_gameState == KSpiral::beforeFirstStart) {
        if(!m_backRect) {
            kDebug() << "vor dem spiel";
            m_backRect = new QGraphicsRectItem(0, 0, event->size().width() * 0.6, event->size().height() * 0.6);
            m_backRect->setPen(QPen(Qt::transparent));
            m_backRect->setBrush(QBrush(Qt::black));
            m_backRect->setOpacity(0.3);
            m_scene->addItem(m_backRect);
            m_view->centerOn(m_backRect);

            QGraphicsTextItem* text = new QGraphicsTextItem(QString(i18n("Click to start game")), m_backRect);
            text->setFont(font());
            qreal scale = m_backRect->rect().width() / text->boundingRect().width() * 0.75;
            text->scale(scale, scale);
            text->setTransform(text->transform().translate(-text->boundingRect().center().x(), -text->boundingRect().center().y()));
            text->setPos(m_backRect->rect().center());
        } else {
            qreal scale = qMin(event->size().width() / m_backRect->rect().width(), event->size().height() / m_backRect->rect().height()) * 0.6;
            m_backRect->setTransform(QTransform().scale(scale, scale));
            m_view->centerOn(m_backRect);
        }
    }
}

void KSpiral::GameWindow::drawGameElements() {
	//setup background
	m_background = new KGameRenderedObjectItem(m_renderer, "Background");
	const QRectF bgRect = m_renderer->boundsOnSprite("Background");
  m_background->setFixedSize(bgRect.size());
  m_background->setPos(bgRect.topLeft());
	m_background->setZValue(0); //background level
	m_scene->addItem(m_background);
	m_view->setFitItem(m_background);

	//setup cannon
	m_cannon = new KSpiral::Cannon(m_renderer, "Cannon", m_cannonSpeed);
	m_cannon->setPos(m_renderer->boundsOnSprite("Cannon").topLeft());
	m_cannon->setBallList(m_ballTypes);
	m_cannon->setZValue(1);
	m_scene->addItem(m_cannon);
	m_cannon->loadCannon();

	m_cannonTop = new KGameRenderedObjectItem(m_renderer, "CannonTop");
  QSize cannonTopSize = m_renderer->boundsOnSprite("CannonTop").size().toSize();
  m_cannonTop->scale(cannonTopSize.width(), cannonTopSize.height());
  m_cannonTop->setOffset(-0.5, -0.5);

  m_cannonTop->setPos(m_renderer->boundsOnSprite("CannonTop").topLeft());
	m_cannonTop->setZValue(20); //level over balls
	m_cannon->setTop(m_cannonTop);
	m_scene->addItem(m_cannonTop);

	m_barrierItems = QList<QPainterPath* >();
	foreach(QString element, m_barriers) {
		if (m_renderer->spriteExists(element)) {
			KGameRenderedItem* barrier = new KGameRenderedItem(m_renderer, element);
      barrier->setRenderSize(m_renderer->boundsOnSprite(element).size().toSize());
			barrier->setZValue(20); //over ball level
			barrier->setPos(m_renderer->boundsOnSprite(element).topLeft());
			m_scene->addItem(barrier);
		}
		if(m_renderer->spriteExists(element + "Outline")) {
			SvgReader* barrierPath = new SvgReader;
			barrierPath->openPathFromSvgFile(new QFile(m_imageFile), element + "Outline");
			m_barrierItems << barrierPath;
		}
	}
	m_ignoredItems << m_background << m_cannon << m_cannonTop;
}

void KSpiral::GameWindow::startGame() {
    m_gameState = KSpiral::inGame;
    if(m_backRect)
    {
        delete m_backRect;
        m_backRect = 0;
    }

	if(m_engine) {
		delete m_engine;
		delete m_scene;
		m_scene = new QGraphicsScene;
		m_view->setScene(m_scene);
	}
	drawGameElements();

	int count = m_maxBallcount.value(0, 0); //it is KGameDifficulty::VeryEasy
	if (KGameDifficulty::level()  == KGameDifficulty::Easy) count = m_maxBallcount.value(1, 0);
	else if (KGameDifficulty::level()  == KGameDifficulty::Medium) count = m_maxBallcount.value(2, 0);
	else if (KGameDifficulty::level()  == KGameDifficulty::Hard) count = m_maxBallcount.value(3, 0);
	else if (KGameDifficulty::level()  == KGameDifficulty::VeryHard) count = m_maxBallcount.value(4, 0);

	int duration = m_duration.value(0, 0); //it is KGameDifficulty::VeryEasy
	if (KGameDifficulty::level()  == KGameDifficulty::Easy) duration = m_duration.value(1, 0);
	else if (KGameDifficulty::level()  == KGameDifficulty::Medium) duration = m_duration.value(2, 0);
	else if (KGameDifficulty::level()  == KGameDifficulty::Hard) duration = m_duration.value(3, 0);
	else if (KGameDifficulty::level()  == KGameDifficulty::VeryHard) duration = m_duration.value(4, 0);

	m_engine = new KSpiral::GameEngine(m_path, this);
	m_engine->setBarriers(m_barrierItems);
	m_engine->setIgnoredItems(m_ignoredItems);
	m_engine->setMaxBallcount(count);
	connect(m_engine, SIGNAL(gameIsOver(int )), this, SLOT(gameIsOver(int )));
	m_engine->startGame(duration);
}

void KSpiral::GameWindow::difficultyChanged() {
	if(m_engine) //game runs
		startGame();
}

void KSpiral::GameWindow::mouseMoveEvent(QMouseEvent* event) {
	if(m_cannon) {
		const QPointF mousePos = m_view->mapToScene(event->pos());
		const QPointF cannonCenter = m_cannon->mapToScene(m_cannon->boundingRect().center());
		qreal angle = atan2((mousePos.y() - cannonCenter.y()), (mousePos.x() - cannonCenter.x()));
		m_cannon->setAngle(angle); //rotate() also moves the cannon
	}
}

#include <kdebug.h>
void KSpiral::GameWindow::mousePressEvent(QMouseEvent* event) {
  event->accept();
  kDebug() << "Mouse clicked";
	if(m_cannon) {
		KSpiral::Ball* cannonball = m_cannon->shoot();
		connect(cannonball, SIGNAL(ballStepped(KSpiral::Ball*)), m_engine, SLOT(cannonballStepped(KSpiral::Ball*)));
	} else {
		startGame();
	}
}

void KSpiral::GameWindow::showHighscores() {
	KScoreDialog* dialog = new KScoreDialog(KScoreDialog::Name | KScoreDialog::Score, this);
	dialog->addLocalizedConfigGroupNames(KGameDifficulty::localizedLevelStrings());
	dialog->setConfigGroupWeights(KGameDifficulty::levelWeights());

    //### following code works only with patched kscoredialog
// 	QMap<QByteArray, QString> levelMap;
// 	levelMap.insert(m_unlocalizedLevelName.toUtf8(), m_levelName);
// 	levelMap.insert(QByteArray("NotDoneYetMap"), QString("NotDoneMap"));
// 	dialog->addLocalizedLevelNames(levelMap);
	dialog->exec();
	delete dialog;
}

void KSpiral::GameWindow::gameIsOver(int score) {
	kDebug() << "game is over";
	KScoreDialog::FieldInfo scoreInfo;
	scoreInfo[KScoreDialog::Score].setNum(score);
	KScoreDialog* dialog = new KScoreDialog(KScoreDialog::Name | KScoreDialog::Score, this);
	dialog->addLocalizedConfigGroupNames(KGameDifficulty::localizedLevelStrings());
	dialog->setConfigGroupWeights(KGameDifficulty::levelWeights());
	dialog->setConfigGroup(KGameDifficulty::localizedLevelString());

//     kDebug() << m_levelName << m_unlocalizedLevelName.toUtf8();
//     dialog->setLevelName(qMakePair(m_unlocalizedLevelName.toUtf8(), m_levelName));
	dialog->addScore(scoreInfo);
	dialog->exec();
	delete dialog;
	delete m_engine;
	m_engine = 0;
}

KSpiral::Ball* KSpiral::GameWindow::addBall(KSpiral::Ball* ball) {
	ball->setZValue(10);
	m_scene->addItem(ball);
	return ball;
}

KSpiral::Ball* KSpiral::GameWindow::addBall(const QPointF& point, QString ballType) {
	if (ballType.isEmpty())
		ballType = randomBalltype();
	return addBall(new KSpiral::Ball(point, m_renderer, ballType));
}

#include "gameWindow.moc"
