//Magazynier, a warehouse assistant game.
//Copyright (C) 2008, 2009 Tadeusz Andrzej Kadlubowski yess@hell.org.pl
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License along
//with this program; if not, write to the Free Software Foundation, Inc.,
//51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

#include "levelModel.h"

#include <KDebug>
#include <QGraphicsItem>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsView>
#include <KLocale>
#include <QStringList>
#include <QVector>

#include "astar.h"
#include "commands.h"
#include "item.h"
#include "renderer.h"

QPoint *LevelModel::m_moveDirUp = 0;
QPoint *LevelModel::m_moveDirRight = 0;
QPoint *LevelModel::m_moveDirDown = 0;
QPoint *LevelModel::m_moveDirLeft = 0;

LevelModel::LevelModel(QObject *parent) : //{{{
	QGraphicsScene(parent),
	m_goalsLeft(0),
	m_shiftCount(0),
	m_levelCompleted(false),
	m_moveCount(0)

{
	if (m_moveDirUp == 0) {
		m_moveDirUp = new QPoint(0, -1);
		m_moveDirLeft = new QPoint(-1, 0);
		m_moveDirDown = new QPoint(0, 1);
		m_moveDirRight = new QPoint(1, 0);
	}
}//}}}

LevelModel::~LevelModel() //{{{
{
	for (int col = 0; col < m_columnCount; ++col) {
		for (int row = 0; row < m_rowCount; ++row) {
			delete m_levelItems->at(col)->at(row);
		}
		delete m_levelItems->at(col);
	}
	delete m_levelItems;

	delete m_playerPos;

}//}}}

void LevelModel::initLevelData(QString &data) //{{{
{
	QStringList lines;
       	lines = data.split(QChar('\n'));
	m_columnCount = 0;
	m_rowCount = lines.size();
	for (int row = 0; row < m_rowCount; ++row) {
		if (lines[row].size() > m_columnCount) {
			m_columnCount = lines[row].size();
		}

		//Remove empty lines
		if (lines[row].isEmpty()) {
			lines.removeAt(row);
			--m_rowCount;
			--row;
		}
	}

	Renderer::instance()->setLevelSize(m_rowCount, m_columnCount);

	m_levelItems = new QVector<QVector<MagazynierItem *>*>(m_columnCount);
	for (int col = 0; col < m_columnCount; ++col) {
		m_levelItems->replace(col, new QVector<MagazynierItem *>(m_rowCount));
		for (int row = 0; row < m_rowCount; ++row) {
			MagazynierItem *newItem = new MagazynierItem(Renderer::instance()->gameRenderer());
			newItem->setItemType(Empty);
			m_levelItems->at(col)->replace(row, newItem);
		}
	}
	
	for (int row = 0; row < m_rowCount; ++row) {
		QString line = lines[row];
		kDebug() << line;
		for (int col = 0; col < line.size(); ++col) {
			MagazynierItem *newItem = getItem(QPoint(col, row));
			addItem(newItem);
			setItem(QPoint(col, row), newItem);
			if (line[col] == ' ') {
				newItem->setItemType(Empty);
				continue;
			}
			if (line[col] == '#') {
				newItem->setItemType(Wall);
				continue;
			}
			if (line[col] == '@') {
				newItem->setItemType(Player);
				newItem->setOrientation(*m_moveDirUp);
				m_playerPos = new QPoint(col, row);
				m_playerOrientation = *m_moveDirUp;
				continue;
			}
			if (line[col] == '$') {
				newItem->setItemType(Box);
				continue;
			}
			if (line[col] == '*') {
				newItem->setItemType(Box|Goal);
				continue;
			}
			if (line[col] == '+') {
				newItem->setItemType(Player|Goal);
				newItem->setOrientation(*m_moveDirUp);
				m_playerOrientation = *m_moveDirUp;
				m_playerPos = new QPoint(col, row);
				++m_goalsLeft;
				continue;
			}
			if (line[col] == '.') {
				newItem->setItemType(Goal);
				++m_goalsLeft;
				continue;
			}
			kDebug() << "Unknown level data:" << line[col];
		}
	}
} //}}}

int LevelModel::getColumnCount(void) const //{{{
{
	return m_columnCount;
} //}}}

int LevelModel::getRowCount(void) const //{{{
{
	return m_rowCount;
} //}}}

MagazynierItem *LevelModel::getItem(QPoint pos) const //{{{
{
	return m_levelItems->at(pos.x())->at(pos.y());
} //}}}

void LevelModel::setItem(QPoint pos, MagazynierItem *item) //{{{
{
	m_levelItems->at(pos.x())->replace(pos.y(), item);
	item->setLevelPos(pos);
	
	return;
} //}}}

void LevelModel::populateScene(void) //{{{
{
	int row, col, itemSize;
	itemSize = Renderer::instance()->getItemSize();
	for (col = 0; col < getColumnCount(); ++col) {
		for (row = 0; row < getRowCount(); ++row) {
			MagazynierItem *item = getItem(QPoint(col,row));
			kDebug() << "Item at" << col << row << ":" << item->itemType();
			item->setRenderSize(QSize(itemSize, itemSize));
			item->updatePixmap();
			item->setLevelPos(QPoint(col, row));
		}
	}
} //}}}

void LevelModel::drawBackground(QPainter *painter, const QRectF &rect) //{{{
{
	QGraphicsView* view = views().first();
	QPointF topLeft = view->mapToScene(0, 0);
	painter->drawPixmap(topLeft, Renderer::instance()->getBackground());
} //}}}

void LevelModel::moveFromEmptyToEmpty(QPoint pos, QPoint delta) //{{{
{
	kDebug() << "Move from empty" << pos << "to empty by" << delta;
	
	MagazynierItem *sourceItem = getItem(pos);
	MagazynierItem *targetItem = getItem(pos + delta);

	setItem(pos, targetItem);
	setItem(pos + delta, sourceItem);

	if (sourceItem->itemType() == Player) {
		*m_playerPos += delta;
		updatePlayerOrientation(delta);
		kDebug() << "player position updated to" << *m_playerPos;
	}

	return;
} //}}}

void LevelModel::moveFromEmptyToGoal(QPoint pos, QPoint delta) //{{{
{
	kDebug() << "Move from empty" << pos << "to goal by" << delta;
	
	MagazynierItem *sourceItem = getItem(pos);
	MagazynierItem *targetItem = getItem(pos + delta);

	int movingItemType = sourceItem->itemType();

	targetItem->setItemType(Goal | movingItemType);
	sourceItem->setItemType(Empty);

	if (movingItemType == Player) {
		*m_playerPos += delta;
		updatePlayerOrientation(delta);
		kDebug() << "player position updated to" << *m_playerPos;
	} else if (movingItemType == Box) {
		--m_goalsLeft;
		kDebug() << m_goalsLeft << "goal(s) left";
	}

	return;
} //}}}

void LevelModel::moveFromGoalToEmpty(QPoint pos, QPoint delta) //{{{
{
	kDebug() << "Move from goal" << pos << "to empty by" << delta;

	MagazynierItem *sourceItem = getItem(pos);
	MagazynierItem *targetItem = getItem(pos + delta);

	int movingItemType = sourceItem->itemType() & (~Goal);

	targetItem->setItemType(movingItemType);
	sourceItem->setItemType(Goal);
	
	if (movingItemType == Player) {
		*m_playerPos += delta;
		updatePlayerOrientation(delta);
		kDebug() << "player position updated to" << *m_playerPos;
	} else if (movingItemType == Box) {
		++m_goalsLeft;
	}

	return;
} //}}}

void LevelModel::moveFromGoalToGoal(QPoint pos, QPoint delta) //{{{
{
	kDebug() << "Move from goal" << pos << "to goal by" << delta;

	MagazynierItem *sourceItem = getItem(pos);
	MagazynierItem *targetItem = getItem(pos + delta);

	setItem(pos, targetItem);
	setItem(pos + delta, sourceItem);
	
	if (sourceItem->itemType() == (Player | Goal)) {
		*m_playerPos += delta;
		updatePlayerOrientation(delta);
		kDebug() << "player position updated to" << *m_playerPos;
	}	

	return;
} //}}}

void LevelModel::updatePlayerOrientation(QPoint orientation) //{{{
{
	m_playerOrientation = orientation;

	MagazynierItem *playerItem = getItem(*m_playerPos);
	kDebug() << "player item type:" <<playerItem->itemType();
	playerItem->setOrientation(orientation);
} //}}}

bool LevelModel::canMove(QPoint pos, QPoint delta) //{{{
{
	MagazynierItem *target = getItem(pos + delta);

	if (target->itemType() == Empty || target->itemType() == Goal) {
		return true;
	}

	if (target->itemType() == Box || target->itemType() == (Goal | Box)) {
		MagazynierItem *boxTarget = getItem(pos + delta + delta);
		if (boxTarget->itemType() == Empty || boxTarget->itemType() == Goal) {
			return true;
		}
	}

	kDebug() << "Can't move from" << pos << target->itemType() << "in the way";
	return false;
} //}}}

MagazynierCommand *LevelModel::move(QPoint pos, QPoint delta, QUndoCommand *parentCommand) //{{{
{
	if (!canMove(pos, delta)) {
		return NULL;
	}
	
	kDebug() << "Move by" << delta;
	MagazynierCommand *aggregateCommand = new MagazynierCommand(this, parentCommand);

	MagazynierItem *here = getItem(pos);
	MagazynierItem *target = getItem(pos + delta);
		
	if (here->itemType() == Player || here->itemType() == 0) {
		kDebug() << "player moves to" << target->itemType();
		if (target->itemType() == Empty) {
			aggregateCommand->setCounts(1, 0);
			new MoveFromEmptyToEmptyCommand(this, pos, delta, aggregateCommand);
		} else if (target->itemType() == Goal) {
			aggregateCommand->setCounts(1, 0);
			new MoveFromEmptyToGoalCommand(this, pos, delta, aggregateCommand);
		} else if (target->itemType() == Box) {
			MagazynierItem *boxTarget = getItem(pos + delta + delta);
			aggregateCommand->setCounts(1, 1);
			if (boxTarget->itemType() == Empty) {
				new MoveFromEmptyToEmptyCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromEmptyToEmptyCommand(this, pos, delta, aggregateCommand);
			} else if (boxTarget->itemType() == Goal) {
				new MoveFromEmptyToGoalCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromEmptyToEmptyCommand(this, pos, delta, aggregateCommand);
			}
		} else if (target->itemType() == (Box | Goal)) {
			MagazynierItem *boxTarget = getItem(pos + delta + delta);
			aggregateCommand->setCounts(1, 1);
			if (boxTarget->itemType() == Empty) {
				new MoveFromGoalToEmptyCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromEmptyToGoalCommand(this, pos, delta, aggregateCommand);
			} else if (boxTarget->itemType() == Goal) {
				new MoveFromGoalToGoalCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromEmptyToGoalCommand(this, pos, delta, aggregateCommand);
			}
		}
	} else if (here->itemType() == (Player | Goal) || here->itemType() == Goal) {
		if (target->itemType() == Empty) {
			aggregateCommand->setCounts(1, 0);
			new MoveFromGoalToEmptyCommand(this, pos, delta, aggregateCommand);
		} else if (target->itemType() == Goal) {
			aggregateCommand->setCounts(1, 0);
			new MoveFromGoalToGoalCommand(this, pos, delta, aggregateCommand);
		} else if (target->itemType() == Box) {
			MagazynierItem *boxTarget = getItem(pos + delta + delta);
			aggregateCommand->setCounts(1, 1);
			if (boxTarget->itemType() == Empty) {
				new MoveFromEmptyToEmptyCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromGoalToEmptyCommand(this, pos, delta, aggregateCommand);
			} else if (boxTarget->itemType() == Goal) {
				new MoveFromEmptyToGoalCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromGoalToEmptyCommand(this, pos, delta, aggregateCommand);
			}
		} else if (target->itemType() == (Box | Goal)) {
			MagazynierItem *boxTarget = getItem(pos + delta + delta);
			aggregateCommand->setCounts(1, 1);
			if (boxTarget->itemType() == Empty) {
				new MoveFromGoalToEmptyCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromGoalToGoalCommand(this, pos, delta, aggregateCommand);
			} else if (boxTarget->itemType() == Goal) {
				new MoveFromGoalToGoalCommand(this, pos + delta, delta, aggregateCommand);
				new MoveFromGoalToGoalCommand(this, pos, delta, aggregateCommand);
			}
		}
	}
	return aggregateCommand;
} //}}}

void LevelModel::moveUp(void)
{
	MagazynierCommand *command = move(*m_playerPos, *m_moveDirUp); 
	if (command != NULL) {
		emit addNewUndoCommand(command);
	}
}

void LevelModel::moveLeft(void)
{
	MagazynierCommand *command = move(*m_playerPos, *m_moveDirLeft);
	if (command != NULL) {
		emit addNewUndoCommand(command);
	}
}

void LevelModel::moveRight(void)
{
	MagazynierCommand *command = move(*m_playerPos, *m_moveDirRight);
	if (command != NULL) {
		emit addNewUndoCommand(command);
	}
}

void LevelModel::moveDown(void)
{
	MagazynierCommand *command = move(*m_playerPos, *m_moveDirDown);
	if (command != NULL) {
		emit addNewUndoCommand(command);
	}
}

void LevelModel::moveDone(int moveCount, int shiftCount) //{{{
{
	if (moveCount != 0) {
		m_moveCount += moveCount;
		emit moveCountChanged(m_moveCount);
	}

	if (shiftCount != 0) {
		m_shiftCount += shiftCount;
		emit shiftCountChanged(m_shiftCount);
	}

	if (!m_levelCompleted && m_goalsLeft == 0) {
		kDebug() << "Level completed";
		m_levelCompleted = true;
		emit levelCompleted();
	}

	return;
} //}}}

void LevelModel::moveUndone(int moveCount, int shiftCount) //{{{
{
	if (moveCount != 0) {
		m_moveCount -= moveCount;
		emit moveCountChanged(m_moveCount);
	}

	if (shiftCount != 0) {
		m_shiftCount -= shiftCount;
		emit shiftCountChanged(m_shiftCount);
	}
	return;
} //}}}

int LevelModel::getMoveCount() //{{{
{
	return m_moveCount;
} //}}}

int LevelModel::getShiftCount() //{{{
{
	return m_shiftCount;
} //}}}

void LevelModel::mouseReleaseEvent(QGraphicsSceneMouseEvent * mouseEvent) //{{{
{
	if(mouseEvent->button() != Qt::LeftButton || m_levelCompleted)
		return;

	int x = (int)(mouseEvent->scenePos().x() / Renderer::instance()->getItemSize());
	int y = (int)(mouseEvent->scenePos().y() / Renderer::instance()->getItemSize());
	QPoint pos = QPoint(x, y);

	if(x<0 || y<0 || x>=getColumnCount() || y>=getRowCount())
		return;

	MagazynierCommand *command = NULL;
	if(getItem(pos)->itemType()==Empty || getItem(pos)->itemType()==Goal) {
		kDebug() << x << y << "Cell empty!";
		QList<QPoint> *path = AStar::findPath(this, *m_playerPos, pos);
		if(path!=NULL) {
			command = new MagazynierCommand(this);
			QPoint prevPos = *m_playerPos;
			foreach(const QPoint pos, *path) {
				kDebug() << "Moving from" << prevPos << "by" << pos-prevPos;
				move(prevPos, pos-prevPos, command);
				prevPos = pos;
			}
		}
		delete path;
	}else if((pos-(*m_playerPos)).manhattanLength()==1 && getItem(pos)->itemType() & Box) {
		// Only move a box when the player is standing next to it (manhattanLength==1)
		command = move(*m_playerPos, pos-(*m_playerPos));
	}

	if (command != NULL) {
		emit addNewUndoCommand(command);
	}
} //}}}

// vim: foldmethod=marker
