/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/


#ifndef SOLVER_H
#define SOLVER_H

#include "ds/data.h"

#include <QObject>
#include <QList>
#include <QHash>
#include <QPoint>
#include <KDebug>
#include <ThreadWeaver/Job>

class Interval
{
public:
    Interval(const int beginAt,
        const int endAt,
        const int column,
        Data::HintEntry * const block,
        const bool isColumn);

    void setBeginIndex(int);
    int beginIndex() const;
    void setEndIndex(int);
    int endIndex() const;
    int column() const;
    int length() const;
    Data::HintEntry* block() const;
    bool isColumn() const;
    void setBlockSum(int);
    int blockSum();

    void setDirty();
    void setNotDirty();
    bool dirty()const;


private:
    int _beginIndex;
    int _endIndex;
    int const _column;
    Data::HintEntry * const _block;
    bool const _isColumn;
    bool _dirty;
    int _blockSum;
};

inline QDebug operator<<(QDebug dbg, const Interval &i)
{
    if (i.block() == NULL) {
        dbg.nospace() << "BLOCK IS NULL - THIS WON'T WORK!";
    }  else {
         dbg.nospace() << "(" << i.beginIndex() << ".." << i.endIndex() << ":" << i.block() << (i.isColumn()?"column":"row") <<i.column() <<"/"<<(i.dirty()?"[*]":"[ ]")<< ")";
    }

 return dbg;
}


inline QDebug operator<<(QDebug dbg, const QList<Interval*>* l)
{
    dbg.nospace() << "(";
    for(int i = 0; i < l->size(); i++) {
        Interval * interval = l->at(i);
        if (interval != NULL) dbg.nospace() << *interval;
        else dbg.nospace() << "NULL";
        if (i<l->size()-1) dbg.nospace() << "\n";
    }
    dbg.nospace() << ")";

 return dbg;
}

class DPoint: public QPoint
{
public:
    DPoint(int x, int y, bool direction);
    DPoint(QPoint p, bool direction);
    bool getDirection() const;

private:
    bool const direction;


};

inline uint qHash ( const DPoint & key ) {

    uint h1 = qHash(key.x());
    uint h2 = qHash(key.y());
    uint r1 = ((h1 << 16) | (h1 >> 16)) ^ h2;

    uint h3 = qHash(key.getDirection());

    uint retval = ((h3 << 16) | (h3 >> 16)) ^ r1;

    return retval;
}


inline bool operator==(const DPoint &p1, const DPoint &p2)
 {
     bool retval = p1.x() == p2.x()
            && p1.y() == p2.y()
            && p1.getDirection() == p2.getDirection();

    return retval;
 }

inline QDebug operator<<(QDebug dbg, const DPoint &p)
{
 dbg.nospace() << "(" << p.x() << "," << p.y() << ":" << p.getDirection() << ")";

 return dbg;
}

class Solver: public ThreadWeaver::Job
{
    Q_OBJECT
public:
    Solver();


public slots:
    void setData(Data * data);
    void findSolution();
protected:
    virtual void run ();

signals:
    void finished();

private:
    void initializeIntervals(
        QList<QList<Data::HintEntry*>*>* hintListList,
        int verticalDirectionCount,
        bool markInColumns);

    void makeIntervalSmaller(Interval * interval);
    void paintBoxesInInterval(Interval * interval);
    void isIntervalFinished(Interval * interval);

    void setToState(Interval* interval, int index, bool on);
    void markDirty(QPoint p, Interval* interval, bool direction);
    DPoint toPoint(Interval* interval, int index, bool direction);

    void adjustInterval(Interval * interval, int border, bool adjustTheEnd);


    Data * m_data;
    QList<Interval*> * dirtyIntervals;

    QHash<DPoint, QList<Interval*>*> * entryToInterval;
};



#endif // SOLVER_H
