/* mainwindow.cpp

    Parsek - free client for playing 4X space strategy games
    Copyright (C) 2006-2007  Jure Repinc <jlp@holodeck1.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#include "mainwindow.h"

MainWindow::MainWindow(QWidget *parent) : KMainWindow(parent)
{
    setupGame();
    setupActions();
    setupMenus(false);
    setupStatusBar();
    setupGameWindows();

    connect(statuslistener, SIGNAL(signalConnected()),this,SLOT(connected()));
    connect(statuslistener, SIGNAL(signalDisconnected()),this,SLOT(disconnected()));
    connect(statuslistener, SIGNAL(signalAccountCreated(bool)) ,this, SLOT(accountCreated(bool)));
    connect(statuslistener, SIGNAL(signalLoggedIn(bool)),this,SLOT(loggedIn(bool)));
    connect(statuslistener, SIGNAL(signalEotStarted()),this,SLOT(eotStarted()));
    connect(statuslistener, SIGNAL(signalEotEnded()),this,SLOT(eotEnded()));
    connect(statuslistener, SIGNAL(signalTimeToEot(uint32_t)),this,SLOT(timeToEot(uint32_t)));
}

void MainWindow::setupGame()
{
    eventloop = new ParsekEventLoop();
    statuslistener = new ParsekStatusListener();
    game = new GameLayer();
    game->setClientString(string("parsek/") + PARSEK_VERSION);
    game->setEventLoop(eventloop);
    game->setGameStatusListener(statuslistener);
}

void MainWindow::setupActions()
{
    connectAction = new KAction(this);
    connectAction->setIcon(KIcon("network-connect"));
    connectAction->setText(i18n("&Connect"));
    connectAction->setStatusTip(i18n("Connect to a game server"));
    connect(connectAction, SIGNAL(triggered()),this, SLOT(connectToServer()));
    quitAction = KStandardAction::quit(this, SLOT(quitGame()), this);
    quitAction->setStatusTip(i18n("Quit the game"));
    disconnectAction = new KAction(this);
    disconnectAction->setIcon(KIcon("network-disconnect"));
    disconnectAction->setText(i18n("&Disconnect"));
    disconnectAction->setStatusTip(i18n("Disconnect from current server"));
    connect(disconnectAction, SIGNAL(triggered()),this, SLOT(disconnect()));
}

void MainWindow::setupMenus(bool connected)
{
    if (!connected)
    {
        gameMenu = new KMenu(i18n("&Game"));
        gameMenu->addAction(connectAction);
        gameMenu->addSeparator();
        gameMenu->addAction(quitAction);
        help = helpMenu();
        menuBar()->addMenu(gameMenu);
        menuBar()->addMenu(help);
    }
    else
    {
        gameMenu->clear();
        gameMenu->addAction(connectAction);
        gameMenu->addAction(disconnectAction);
        gameMenu->addSeparator();
        gameMenu->addAction(quitAction);

    }

}

void MainWindow::setupStatusBar()
{
    clockLabel = new QLabel();
    clockLabel->setText("EOT Clock:");
    clockDisplay = new QLabel();
    clockDisplay->setContentsMargins(0,0,6,0);
    clockDisplay->setNum(000);
    clockLabel->hide();
    clockDisplay->hide();
    connectionLed = new KLed(Qt::red);

    statusBar()->showMessage("Welcome to Parsek!",50000);
    statusBar()->addPermanentWidget(clockLabel);
    statusBar()->addPermanentWidget(clockDisplay);
    statusBar()->addPermanentWidget(connectionLed);
}

void MainWindow::setupGameWindows()
{
    logger = new LoggerWidget();
    game->setLogger(logger);
    QDockWidget *loggerDock = new QDockWidget(i18n("Log"), this);
    loggerDock->setAllowedAreas(Qt::BottomDockWidgetArea | Qt::TopDockWidgetArea);
    loggerDock->setWidget(logger);
    addDockWidget(Qt::BottomDockWidgetArea, loggerDock);

    status = new StatusWidget();
    QDockWidget *statusDock = new QDockWidget(i18n("Status"), this);
    statusDock->setAllowedAreas(Qt::BottomDockWidgetArea | Qt::TopDockWidgetArea);
    statusDock->setWidget(status);
    addDockWidget(Qt::BottomDockWidgetArea, statusDock);

    objectview = new ObjectView();
    QDockWidget *objectviewdock = new QDockWidget("System View",this);
    objectviewdock->setWidget(objectview);
    objectviewdock->setFeatures(QDockWidget::NoDockWidgetFeatures);
    objectviewdock->setTitleBarWidget(0);
    addDockWidget(Qt::RightDockWidgetArea,objectviewdock);
}

void MainWindow::connectToServer()
{
    ConnectToServerDialog connectionDialog(this);
    if (connectionDialog.exec())
    {
        qApp->processEvents();
        serveraddress = connectionDialog.serversHistoryCombo->currentText().toUtf8().constData();
        clientuser = connectionDialog.usernameLine->text().toUtf8().constData();
        clientpass = connectionDialog.passwordLine->text().toUtf8().constData();
        if (game->connect(serveraddress)) {
            connectionLed->setColor(Qt::yellow);
        }
    }
}
void MainWindow::quitGame()
{
    if (game->getStatus() != gsDisconnected)
    {
        game->disconnect();
    }
    close();
}
void MainWindow::disconnect()
{
    if (game->getStatus() != gsDisconnected)
    {
        game->disconnect();
    }
}
void MainWindow::connected()
{
    setupMenus(true);
    status->statusConnected();
    statusBar()->showMessage("Connected!");
    connectionLed->setColor(Qt::green);

    if (setupGameList())
    {
        game->login(clientuser,clientpass,servergame);
    }
    else
    {
        game->disconnect();
    }

}
bool MainWindow::setupGameList()
{
    std::set<GameInfo*> games = game->getAvailableGames();

    if (games.empty())
    {
        KMessageBox::error(this,
                           i18n("Sorry, no games are available on this server. Disconnecting."),
                           i18n("No Available Games"));
        return false;
    }
    else
    {
        ConnectToGameDialog gameSelectionDialog(this);
        for(std::set<GameInfo*>::iterator itcurr = games.begin(); itcurr != games.end(); ++itcurr)
        {
            std::string temp = (*itcurr)->getGameName();
            QString gamename = QString::fromUtf8(temp.c_str());
            gameSelectionDialog.addGame(gamename);
        }

        if (gameSelectionDialog.exec()) //accepted
        {
            servergame = gameSelectionDialog.getGame().toUtf8().constData();
            return true;
        }
        else //rejected
        {
            return false;
        }
    }
}
void MainWindow::accountCreated(bool)
{
     statusBar()->showMessage("Account Created!",1000);
     status->statusAccountCreated();
}
void MainWindow::loggedIn(bool)
{
    statusBar()->showMessage("Logged In!",5000);
    status->statusLoggedIn();
    objectview->setGame(game);

    clockUpdate = new QTimer(this);
    connect(clockUpdate, SIGNAL(timeout()), this, SLOT(updateClock()));
    clockUpdate->start(1000);

    game->getTimeRemaining();

    status->setServeraddress(serveraddress.c_str());
    status->setClientusername(clientuser.c_str());
    status->setServername(game->getActiveGame()->getServerName().c_str());
    status->setServerversion(game->getActiveGame()->getServerVersion().c_str());
    status->setGamename(servergame.c_str());
    status->setRulesetname(game->getActiveGame()->getRulesetName().c_str());
    status->setRulesetversion(game->getActiveGame()->getRulesetVersion().c_str());
}
void MainWindow::disconnected()
{
    clockDisplay->hide();
    clockLabel->hide();
    statusBar()->showMessage("Disconnected!");
    status->statusDisconnected();
    connectionLed->setColor(Qt::red);
    objectview->clear();
}
void MainWindow::timeToEot(uint32_t timetoeot)
{
    clock = timetoeot;
    clockLabel->show();
    clockDisplay->show();
    logger->parsek("Clock update received");
}
void MainWindow::eotEnded()
{
    //update game data
    //clock is automatically updated
}
void MainWindow::eotStarted()
{
    //display dialog box
}
void MainWindow::updateClock()
{
    if (clock != 0)
    {
        clock--;
    }
    if (clock <= 60)
    {
        flashClock();
    }
    clockDisplay->setNum(static_cast<int>(clock));
}
void MainWindow::flashClock()
{
   if (clock % 2 == 0 && clock != 0)
    {
          clockDisplay->setStyleSheet("color: rgb(255, 0, 0)");
          clockLabel->setStyleSheet("color: rgb(255, 0, 0)");
    }
    else
    {
           clockDisplay->setStyleSheet("color: rgb(0, 0, 0)");
           clockLabel->setStyleSheet("color: rgb(0, 0, 0)");
   }
}



