/* ObjectModel.cpp

    Parsek - free client for playing 4X space strategy games
    Copyright (C) 2006-2007  Jure Repinc <jlp@holodeck1.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <QIcon>
#include <klocale.h>
#include <stdint.h>
#include "objectimport.h"
#include "objectitem.h"
#include "objectmodel.h"

using namespace TPProto;

ObjectModel::ObjectModel(QObject *parent) : QAbstractItemModel(parent)
{
    m_root = new ObjectItem();
    m_universe = new ObjectItem();
    m_game = 0;
}

ObjectModel::~ObjectModel()
{
    delete m_root;
    delete m_universe;
    delete m_game;
}

void ObjectModel::setGame(TPProto::GameLayer *game)
{
    m_game = game;
}

void ObjectModel::getUniverse()
{
    if (m_game)
    {
        m_game->getObjectCache()->requestObject(0, boost::bind(&ObjectModel::receiveUniverse, this, _1));
    }
}
void ObjectModel::receiveUniverse(boost::shared_ptr<Object> object){
    if (object){
        delete m_universe;
        ObjectItem* m_new = new ObjectItem(m_root);
        importObject(object,m_new);
        m_root->appendChild(m_new);
        Q_FOREACH (uint32_t id, object->getContainedObjectIds())
        {
            getObject(id,m_new);
        }
    }
    reset();
}

void ObjectModel::getObject(unsigned int objectid, ObjectItem *parent)
{
    m_game->getObjectCache()->requestObject(objectid, boost::bind(&ObjectModel::receiveObject, this, _1, parent));
}

void ObjectModel::receiveObject(boost::shared_ptr<Object> object, ObjectItem *parent)
{
    ObjectItem *addedObject = new ObjectItem(parent);
    importObject(object, addedObject);
    parent->appendChild(addedObject);
    Q_FOREACH (unsigned int id, object->getContainedObjectIds())
    {
        getObject(id,addedObject);
    }
    reset();
}

QModelIndex ObjectModel::index(int row, int column,
                                const QModelIndex &parent) const
{
    ObjectItem *parentItem;

    if (!parent.isValid())
    {
        parentItem = m_root;
    }
    else
    {
        parentItem = static_cast<ObjectItem*>(parent.internalPointer());
    }

    if ( row >= 0 && row < parentItem->childrenCount() )
        return createIndex( row, column, parentItem->child(row) );
    else
        return QModelIndex();
}

QModelIndex ObjectModel::parent(const QModelIndex &index) const
{
    ObjectItem *item = m_ObjectItemFromIndex(index);
    if (!item)
        return QModelIndex();

    ObjectItem *parentItem = item->parent();
    if (!parentItem || parentItem == m_root)
        return QModelIndex();

    ObjectItem *grandParentItem = parentItem->parent();
    if (!grandParentItem)
        return QModelIndex();

    int row = grandParentItem->children.indexOf(parentItem);
    return createIndex(row, 0, parentItem);
}

int ObjectModel::rowCount(const QModelIndex &parent) const
{
    ObjectItem *parentItem;

    if (!parent.isValid() )
        parentItem = m_root;
    else
        parentItem = static_cast<ObjectItem*>( parent.internalPointer() );

    return parentItem->childrenCount();
}

int ObjectModel::columnCount (const QModelIndex &parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant ObjectModel::data(const QModelIndex &index, int role) const
{
    ObjectItem *object = m_ObjectItemFromIndex(index);
    if (!object)
    {
        return QVariant();
    }

    if ( role == Qt::DecorationRole )
    {
        //return QIcon("");
    }
    if ( role == Qt::DisplayRole)
    {
        switch(index.column())
        {
            case 0: return object->id();
            case 1: return object->name();
            default: return QVariant();
        }
    }
    else
        return QVariant();
}

QVariant ObjectModel::headerData(int section, Qt::Orientation orientation,
                                  int role) const
{
    if (role != Qt::DisplayRole) return QVariant();
    if (section == 0 && orientation == Qt::Horizontal) return i18n("ID");
    if (section == 1 && orientation == Qt::Horizontal) return i18n("Name");
    if (orientation == Qt::Vertical) return section;
    else return QVariant();
}

ObjectItem *ObjectModel::m_ObjectItemFromIndex(const QModelIndex &index) const
{
    if (index.isValid()) {
        return static_cast<ObjectItem *>(index.internalPointer());
    } else {
        return m_root;
    }
}

void ObjectModel::refresh(){
    reset();
}
Qt::ItemFlags ObjectModel::flags(const QModelIndex &index) const
{
    if (!index.isValid())
        return 0;

    return Qt::ItemIsSelectable | Qt::ItemIsEnabled;
}
